/* 
   utils/cubes.c 
   Berechnung kubischer Gitter aus Vektoren die Elementarzelle aufspannen 
*/
#include <stdlib.h>
#include <stdio.h>

#if 1
#define N 14
int e_cell[3][3]={{2,2,2},{2,-1,1},{1,2,-1}};	/* elementary cell (a,b,c) */
#else
#define N 8
int e_cell[3][3]={{2,0,0},{0,2,0},{0,0,2}};   /* elementary cell (a,b,c) */
#endif
int vertex[N][3];               /* position of vertizees */         

/* +++TeX+++

 Beispiel in zwei Dimensionen. $a=(a_x,a_y)$ und $b=(b_x,b_y)$ sind die 
 zwei Vektoren, die die Elementarzelle aufspannen. 
 Sie sind im allgemeinen nicht senkrecht zueinander. 
 Der Vektor $r=(r_x,r_y)$ liegt innerhalb der Elementarzelle, wenn seine
 Komponenten $(r_a,r_b)$ bezueglich der Elementarvektoren zwischen $0$ und $1$
 liegen.
 \begin{eqnarray*}
      r &=& r_a a   + r_b b \\
 %
    a r &=& r_a a^2 + r_b a b \\
    b r &=& r_a a b + r_b b^2 \\
 %
    r_a &=& \frac{(b^2a-(a,b)b)r}{a^2b^2-(a,b)^2} \\
    r_b &=& \frac{(a^2b-(a,b)a)r}{a^2b^2-(a,b)^2} \\
 \end{eqnarray*}
 Grenzfall $(a,b)=0$ liefert bekannte Beziehungen. 
 Im 3 dimensionalen muss entsprechend ein Gleichungssystem 3.\ Ordnung 
 geloest werden.
 \begin{verbatim}
  maple> solve( { ar=ra*aa+rb*ab+rc*ac,\ 
  maple>          br=ra*ab+rb*bb+rc*bc,\
  maple>          cr=ra*ac+rb*bc+rc*cc},{ra,rb,rc});
  {ra=( + (bb cc - bc bc) ar + (ac bc - cc ab) br + (bc ab - ac bb) cr)/%1,
   rb=-( + (ab cc - bc ac) ar + (ac ac - cc aa) br + (bc aa - ac ab) cr)/%1,
   rc=( + (bc ab - bb ac) ar + (ab ac - aa bc) br + (aa bb - ab ab) cr)/%1}
  %1 := + ac bc ab + ac bc ab + bb cc aa - bc bc aa - ac ac bb - ab ab cc
  \end{verbatim}
  Auf die Division durch die Determinante kann evl. verzichtet werden, in dem
  die Einheit auf Det(a,b,c) gesetzt wird.
  Hilfreich koennte die Speicherung der Determinante der Einheitszelle und
  der inversen Elementarzelle (normiert mit der Determinante) sein.
   ---TeX--- */

/* vertex -> (x,y,z) */
int get_koord(const int i,int s[3]){
  if(i<0 || i>=N) { printf("error: get_koord\n"); return -1; }
  s[0]=vertex[i][0];
  s[1]=vertex[i][1];
  s[2]=vertex[i][2];
  return 0;}
  
/* (x,y,z) -> vertex */  
int get_vertex(int xx[3]){
  int i,x[3],(*e)[3]=e_cell,aa,ab,ac,bb,bc,cc,ar,br,cr,ra,rb,rc,det,(*s)[3]=vertex;
  x[0]=xx[0];x[1]=xx[1];x[2]=xx[2];
  aa=e[0][0]*e[0][0]+e[0][1]*e[0][1]+e[0][2]*e[0][2];
  ab=e[0][0]*e[1][0]+e[0][1]*e[1][1]+e[0][2]*e[1][2];
  ac=e[0][0]*e[2][0]+e[0][1]*e[2][1]+e[0][2]*e[2][2];
  bb=e[1][0]*e[1][0]+e[1][1]*e[1][1]+e[1][2]*e[1][2];
  bc=e[1][0]*e[2][0]+e[1][1]*e[2][1]+e[1][2]*e[2][2];
  cc=e[2][0]*e[2][0]+e[2][1]*e[2][1]+e[2][2]*e[2][2];
  det= + ac*bc*ab + ac*bc*ab + bb*cc*aa - bc*bc*aa - ac*ac*bb - ab*ab*cc;
  ar=e[0][0]*x[0]   +e[0][1]*x[1]   +e[0][2]*x[2];
  br=e[1][0]*x[0]   +e[1][1]*x[1]   +e[1][2]*x[2];
  cr=e[2][0]*x[0]   +e[2][1]*x[1]   +e[2][2]*x[2];
  ra=( + (bb*cc - bc*bc)*ar + (ac*bc - cc*ab)*br + (bc*ab - ac*bb)*cr);
  rb=-( + (ab*cc - bc*ac)*ar + (ac*ac - cc*aa)*br + (bc*aa - ac*ab)*cr);
  rc=( + (bc*ab - bb*ac)*ar + (ab*ac - aa*bc)*br + (aa*bb - ab*ab)*cr);
  if(ra<0)ra=-((-ra-1)/det)-1;else ra/=det;
  if(rb<0)rb=-((-rb-1)/det)-1;else rb/=det;
  if(rc<0)rc=-((-rc-1)/det)-1;else rc/=det;
  x[0]-=ra*e[0][0]+rb*e[1][0]+rc*e[2][0];
  x[1]-=ra*e[0][1]+rb*e[1][1]+rc*e[2][1];
  x[2]-=ra*e[0][2]+rb*e[1][2]+rc*e[2][2];
  for(i=0;i<N;i++) if(x[0]==s[i][0] && x[1]==s[i][1] && x[2]==s[i][2]) return i;
  printf("error get_vertex: %2d %2d %2d -(%2d %2d %2d)-> %2d %2d %2d not found\n",
    xx[0],xx[1],xx[2],ra,rb,rc,x[0],x[1],x[2]);
  return -1;}

main(){ /* vertex_init */
   int i,j,k,r[3],ra,rb,rc,aa,ab,ac,bb,bc,cc,ar,br,cr,det,emin[3],emax[3],
     (*s)[3]=vertex; const int (*e)[3]=e_cell /* ,*a=e[0],*b=e[1],*c=e[2] */ ;
   /* plaetze finden! */
   
   for(i=0;i<3;i++)printf("%c=(%2d %2d %2d)\n",'a'+i,e[i][0],e[i][1],e[i][2]);
   printf("N=Volumen=Spatprodukt=Det=%d\n",
     e[0][0]*(e[1][1]*e[2][2]-e[1][2]*e[2][1])   /* ax(by cz - bz cy) */ 
    +e[0][1]*(e[1][2]*e[2][0]-e[1][0]*e[2][2])   /* ay(bz ca - bx cz) */
    +e[0][2]*(e[1][0]*e[2][1]-e[1][1]*e[2][0])); /* az(bx cb - by cx) */

   /* einhuellenden Kubus ermitteln (alle eck-koords checken) */
   for(i=0;i<3;i++){ 
     emin[i]=emax[i]=e[0][i]+e[1][i]+e[2][i]; 
     if(emin[i]>0)emin[i]=0;
     if(emax[i]<0)emax[i]=0;
     for(j=0;j<3;j++){
       if(e[j][i]<emin[i])emin[i]=e[j][i];
       if(e[j][i]>emax[i])emax[i]=e[j][i];
       for(k=j+1;k<3;k++){
         if(e[j][i]+e[k][i]<emin[i])emin[i]=e[j][i]+e[k][i];
         if(e[j][i]+e[k][i]>emax[i])emax[i]=e[j][i]+e[k][i];
   }}}
   printf("emin=(%d %d %d) emax=(%d %d %d)\n",
   emin[0],emin[1],emin[2],emax[0],emax[1],emax[2]);
   
   i=0;
   aa=e[0][0]*e[0][0]+e[0][1]*e[0][1]+e[0][2]*e[0][2];
   ab=e[0][0]*e[1][0]+e[0][1]*e[1][1]+e[0][2]*e[1][2];
   ac=e[0][0]*e[2][0]+e[0][1]*e[2][1]+e[0][2]*e[2][2];
   bb=e[1][0]*e[1][0]+e[1][1]*e[1][1]+e[1][2]*e[1][2];
   bc=e[1][0]*e[2][0]+e[1][1]*e[2][1]+e[1][2]*e[2][2];
   cc=e[2][0]*e[2][0]+e[2][1]*e[2][1]+e[2][2]*e[2][2];
   det= + ac*bc*ab + ac*bc*ab + bb*cc*aa - bc*bc*aa - ac*ac*bb - ab*ab*cc;
   for(r[0]=emin[0];r[0]<=emax[0];r[0]++)
   for(r[1]=emin[1];r[1]<=emax[1];r[1]++)  /* ineffektiv! */
   for(r[2]=emin[2];r[2]<=emax[2];r[2]++){
    ar=e[0][0]*r[0]   +e[0][1]*r[1]   +e[0][2]*r[2];
    br=e[1][0]*r[0]   +e[1][1]*r[1]   +e[1][2]*r[2];
    cr=e[2][0]*r[0]   +e[2][1]*r[1]   +e[2][2]*r[2];
    ra=( + (bb*cc - bc*bc)*ar + (ac*bc - cc*ab)*br + (bc*ab - ac*bb)*cr);
    rb=-( + (ab*cc - bc*ac)*ar + (ac*ac - cc*aa)*br + (bc*aa - ac*ab)*cr);
    rc=( + (bc*ab - bb*ac)*ar + (ab*ac - aa*bc)*br + (aa*bb - ab*ab)*cr);
    printf(" r=(%2d %2d %2d) det^2(ra rb rc)=(%4d %4d %4d)",
      r[0],r[1],r[2],ra,rb,rc);
    printf(" ra*a+rb*b+rc*c=(%2d %2d %2d)",
        (ra*e[0][0]+rb*e[1][0]+rc*e[2][0])/det,
        (ra*e[0][1]+rb*e[1][1]+rc*e[2][1])/det,
        (ra*e[0][2]+rb*e[1][2]+rc*e[2][2])/det );
    if(ra>=0 && rb>=0 && rc>=0 && ra<det && rb<det && rc<det){
      printf(" i=%3d\n", i);
      s[i][0]=r[0];
      s[i][1]=r[1];
      s[i][2]=r[2]; i++; }
    else printf("\n");
   }
   if(i!=N)printf("error vertex_init i!=N i=%d N=%d\n",i,N);

   /* next neighbours */
   j=0;
   for(i=0;i<N;i++){
     get_koord(i,r);
     r[0]++;k=get_vertex(r);r[0]--; printf("%2d: %2d - %2d\n",j,i,k); j++;
     r[1]++;k=get_vertex(r);r[1]--; printf("%2d: %2d - %2d\n",j,i,k); j++;
     r[2]++;k=get_vertex(r);r[2]--; printf("%2d: %2d - %2d\n",j,i,k); j++;
   }
   return 0;
}
