/*
   This is the main program. see README
   exact diagonalization of Heisenberg/Hubbard/tJ-systems
   
   ToDo: Simplify the program to make the code better readable!!!!

   startvec = \sum_j c^+_{i_j,s} ... |0> = 4u 2d 3u |0>    
   
   complex? phase factor later added during computation of H(i,j)
    multiply phase to base in a way that Hamiltonian and Vector is reell!?

   set parameters
   <OP> as pointer of OP-function
   static int (*a)(int b); o.ae.

   tU -> ...u..u. * ..d..d divide into up and down part (2 tables)   
   tJ -> ...uu... ...--d.d => (N over u)*(N-u over d)
   JJ -> ...
   
   Warn: operation x<<32 leads to errors for 32bit (no shift)
  
   result of (~0)>>1==~0 != (~0u)>>1  depends on compiler?
   
*/

#include "block.h"
#include "spins.h"
#include "tbase.h"
#include "ansatz.h"
#include "pipe.h"
#include "vector.h"
#ifdef CONFIG_PTHREAD   /* ToDo: movto mthreads.h */
#include <pthread.h>
pthread_attr_t pt_attr;  /* SunOS has SCOPE_PROCESS as default, we need SCOPE_SYSTEM */
pthread_cond_t pt_cond1;  /* conditions master -> childs */
pthread_cond_t pt_cond2;  /* conditions childs -> master */
pthread_mutex_t pt_mutex1; /* mutex */
pthread_mutex_t pt_mutex2; /* mutex */
#else  /* dummy definitions to eleminate ifdef constructions */
#define pthread_mutex_lock(x)
#define pthread_mutex_unlock(x)
#endif
#ifdef CONFIG_MPI
#include <mpi.h>
#endif
#include "vdate.h"
#include "version.h"

#define Korrelations 1  /* 1=with 0=without spin-correlations */

#ifndef Ham1
#define Ham1	1	/* SiSj - ninj/4 + (ham1)?ninj/4:0 */          
#endif
#ifndef Ham2            /* results differ in U*(N/4-(nu+nd)/2) */
#define Ham2	0	/* 0 = U nu nd ,  1 = U(nu-.5)(nd-.5) */
#endif
#ifndef Ham3
#define Ham3	0	/* 1 = with JH-Oles-term cc and cpcp, better sym */
#endif 


#include "spinstri.h"	/* spinstri.c::outtri() */
#include "mem.h"	/* mymalloc();myfree() */
#include "symmetry.h"	/* lattice symmetries */
#include "diag.h"
#include "myprint.h"	/* maple-Ausgabe */
#include "xval.h"	/* expectation values */
#include "h_file.h"	/* read/write Hamilton matrix on Htmp */
#include "error.h"	/* err_fatal */
#include "hilbert.h"
#include "modeldef.h"

#ifndef Add34
#define Add34 0.0	/* SiSj+Add34*ninj shifting energy, default is 0.0 */
			/* may be +0.75 or -0.25 (or +0.25) is a good idea */
#endif

/* against egcs-2.90.29 980515 (egcs-1.0.3 release)
 * warning: initialization of non-const reference "class a_cplx &" from rvalue "a_cplx" */
double  h_xypara=1.0;
double  h_zpara=1.0;	/* H=h_zpara*SizSjz+h_xypara*(SixSjx+SiySjy),
				 z-anisotropy Feb99, xy Jun99 */
int     algo;		/* 0=lanzcos,1=richter,2=wenzel2x2 +8=newns,+16=ns-only */
int     save_mode=0; 	/* 1=setup after crash possible, 2=continue */
int     checkpoint=0;   /* checkpoint for crash status */
int	pew;		/* number eigenvalue to check */
int	max_ea;		/* max. zu messender Entartungsgrad */
int     x_p;		/* Darstellungsform fuer |x> <=3 */
ULong	n1;		/* Matrixsize , n2=Linsize obsolete */
int     n1_bits;        /* bits needed to store n1, used for MPI packing */
ULong   min_b_sn1;      /* pseudo minimum stored n1 of all blocks (MPI) */
zahl ew[NEW];		/* Eigenwerte     (max 4)	*/
ULong  *l2b;		/* Hilfs-Tabellen storeh2 */

/* ToDo: norm2 remove it, not necessary (tiny speedup) but complex code */
// ATnorm2 *norm2=NULL;	/* vector idx->sym-idx, norm(sym-idx)	*/
// minfo_t m_n2 = {NULL,0,0,"tmp/n2.map",NULL}; /* norm2 companion */
// int base2ok=0;			/* flag: norm2 not yet inititialized */


ULong hfmax;			/* max. num of hxy[]-elements on disk	*/
#if 0
FILE *hfile,*fhxy1,*fhxy2;	/* File enthaelt Hamiltonien		*/
#endif
ULong usemem,maxmem;		/* maximaler speicher -> no paging!	*/
int  startvec=4;		/* 0:rand(iv), 1:v[iv]=1, 2:v=file! 
				   3:v=intxtfile 4:rand */

int cfg_sisj;	/* option for correlations, bit0=SiSj bit1=SS bit2=ZiZj(?) */

tbase1 p_mask[NN][NN];		/* bit masks for parity()  */

/* max/min/sum hnz_blocks per thread (balance in storeh/geth_block) */
ULong bsnz_max, bhnz_max, bhnz_min, bhnz_sum;
ULong mpi_traffic=0,     /* MPI statistics, data volume (/band_width) */
      mpi_syncs=0,       /* number of MPI transactions (*latency) */
      mpi_maxlen=0;      /* maximum len data block in units mcplx/ulong */

/* local variables for pthreads */
typedef struct {
  int num;     /* to identify itself (ro) */
  /* volatile */ void *data;  /* to send data-block (set by thread 0, read by workers) */
  /* volatile */ int  task;   /* thread 0 sets 1 to run worker, worker sets 0 if ready */
  /* volatile */ void (*worker)(void *data); /* parallel working function */
  /* mutex for each thread? */
} main_pts;
volatile main_pts main_pt[B_NUM]; /* [un]locked by mutex1 */

#ifdef CONFIG_PTHREAD
/* MPI code is in not thread safe in general, we have to care about it */
pthread_mutex_t token_ring[B_NUM*4]; /* of double(quad) length */

/* send jobtype (task) to threads and restart them */
void parallel_start( int thread_i, void (*worker)(void *data), void *data ){
    pthread_mutex_lock(&pt_mutex1);
    main_pt[thread_i].task=1; /* run worker */
    main_pt[thread_i].data=data;
    main_pt[thread_i].worker=(/*volatile*/ void (*)(void *))worker;
    /* ToDo: do next only once after storing all tasks
             or send signal only to the job */
    pthread_cond_broadcast(&pt_cond1); /* data send to threads */
    pthread_mutex_unlock(&pt_mutex1);
}

/* waiting for threads */
void parallel_wait( int thread_i ){
    pthread_mutex_lock(&pt_mutex1);
    while ( main_pt[thread_i].task )  
      pthread_cond_wait(&pt_cond2,&pt_mutex1); /* wait for child */
    pthread_mutex_unlock(&pt_mutex1);
}
#else /* dummy parallel functions (serialized pthreads) */
void parallel_start( int thread_i, void (*worker)(void *data), void *data ){
  worker(data); }
void parallel_wait( int thread_i ){}
#endif

/* spins.h: inline zahl norm(zahl r){return(r*r);} */
char *sclock(long t){  /* 0..960s 16m..960m 16h... => 00:00:00 */
   static char s[64]; if (t==~0l) t=itime();
#if 0
   /* this format is human readable but bad to show in gnuplot */
   sprintf(s,"%02ld:%02ld:%02ld",t/3600,(t/60)%60,t%60);
#elif 1
   char s2[24];
   /* this format is a mixture for simple parse and human read */
   /* this format is human readable but not simply enough to parse */
   if(t>4*60*60l)sprintf(s2,"%ldh",(t+1799)/3600); else
   if(t>   4*60l)sprintf(s2,"%ldm",(t+29)  /  60); else sprintf(s2,"%lds",t); 
   /* 0.1-9999.9min is best compromize, 5s..150h=6d */
   /* a recursive call like sprintf(s,"%s",s) is dangerous */
   sprintf(s,"%6.2f =%s",t/60.,s2);
#endif
   return(s);}

char *outc(cplx r){  /* output [complex] float number */    
    static char s1[100];
#if VecType & 4
    double phase=arg(r)/M_PI; if (phase<0) phase+=2;
    /* sprintf(s1,"(%g,%g)",real(r),imag(r)); -> |r|*exp(phase*I*Pi) */
    if (fabs(r)<0.0001 || fabs(r)>9.99999)
         sprintf(s1,"%12.5e*phase(%5.3f)",fabs(r), phase);
    else sprintf(s1,"%12.9f*phase(%5.3f)",fabs(r), phase);
#else /* % 12.9f=-0.123456789 vs. % 12.5e=-1.12345e-09 */
    if (fabs(r)<0.0001 || fabs(r)>9.99999)
         sprintf(s1,"% 12.5e",D(r));
    else sprintf(s1,"% 12.9f",D(r)); /* 12.9f or 12.5e */
#endif
    return(s1);}

char *outphase(tphase r){  /* complexe zahl oder integer */    
    static char s1[100];
#if VecType & 4
    double phase=arg(r)/M_PI; if (phase<0) phase+=2;
    /* sprintf(s1,"+(% g,% g)",real(r),imag(r)); */
    sprintf(s1,"(1%+5.0e)*phase(%5.3f)", 1-fabs(r), phase);
#else
    sprintf(s1,"%+d",r); 
#endif
    return(s1);}

void nl(FILE *f){ fprintf(f,"\n");return;}
/* output eigenvalues */
void outew(FILE *f){
   int k;
   for (k=0;k<NEW;k++) fprintf(f," %12.8f",(double)ew[k]);
   return;
}


/* ---------------------------------------------------- *
 *      statistics about hamiltonien  (sym,num_nz,...)	*
 *	-	-	-	-	-	-	
 *  double counts to 2^53=9e15 (54bit+10bit) exp e+307(9bit)
 *  
 */
struct tstatH {    
      ULong nzxmin,nzxmax; /* max. number nz elements of a line */
      cplx sum12, sum21;  /* sum-upperright,lowerleft  r    to detect errors */
      double hh12, hh21;  /* sum-upperright,lowerleft <r|r> to detect errors */
      double traceH;      /* Trace of H for manual error detection */
      /* statistics for data transfers within and between nodes */
      /* MPI_Allreduce dont allow LONG_LONG */
      double hnz_nd,	  /* number of non-diagonal elements */
             hnz_d,	  /* number of diagonal elements     */
             hnz_local_thread,
             hnz_inter_thread,
             hnz_inter_node; /* ToDo: [nodes] for min/max/mean%??? */
      } 
      statH[B_NUM]; /* ToDo: -> statH[B_NUM] distinct cache lines */
            
void statH_ini(){ /* reset statistic counters */
 int i;
 for(i=0;i<pt_n;i++){ /* thread safe */
   statH[i].nzxmax=0l;
   statH[i].nzxmin=999999l;
   statH[i].sum12=C(0,0);
   statH[i].sum21=C(0,0);
   statH[i].hh12=0.0;
   statH[i].hh21=0.0;
   statH[i].traceH=0.0;
   statH[i].hnz_d=0;
   statH[i].hnz_nd=0;
   statH[i].hnz_local_thread=0;
   statH[i].hnz_inter_thread=0;
   statH[i].hnz_inter_node  =0;
 }
}
/* called by hamilton_puth_line() on each line for statistic analysis
 * also check for potential errors
 * ToDo: replace by tbx=bx%pt_n (thread block) and gbx=tbx+mpi_i*pt_n (global block)
 * called per num_ahead_lines
 */
void add_statH(thbuf *hbuf){
  double rr;
  ULong xy, x, n_old, y;
  int bx, by, ii;
  cplx r;
  static char buf[80];
  bx=hbuf->block;
  if (bx>=pt_n)                F1("bx>=pt_n bx=%d", bx);
  x=hbuf->i - hbuf->i%hbuf->num_ahead_lines;
  // ToDo: check x > n1+num_ahead_lines !?
  if (x>=n1 || x>=node_n1) return; // happens if i+ahead > n1
// if (x>=n1) F1("x>=n1 x=%.0f",1.*x)
// if (x>=node_n1) F1("x>=node_n1 x=%.0f",1.*x)
  if (x>=b_len[bx+mpi_i*pt_n]) {
     static int reported=0;
     if (!reported)
       printf("\n %d: x>=b_len   x=%.0f (may happen if i+ahead>n1)",
          mpi_i*pt_n+bx, 1.*x);
     reported=1;
     return; // happens on extra calls for balancing???
  }
  x+=b_ofs[bx+mpi_i*pt_n]; /* 0..n1, ToDo: remove? */
  n_old=0;
  if (vvv&64 && mpi_i==0)
    printf("\nvvv&64, output_H x_max=%ld lines:",x_max);

  /* synchronize mpi output */
  fflush(stdout);  /* just for the case ... */
  setvbuf(stdout,buf,_IOFBF,80);

  for (ii=0;ii<hbuf->n;ii++) {  /* for each hnz element */
    xy=hbuf->el[ii].jj;          /* 0..b_len[by]-1 */
    r =hbuf->el[ii].rr;
    y =xy & ~XY_FLAG;
    /* by=b_idx2blk(y); */ /* get destination block number, H[bx,by] */
    // by=scfg2blk(hbuf->el[ii].bj); /* 0..mpi_n*pt_n-1 */
    by=hbuf->el[ii].blkj; /* 0..mpi_n*pt_n-1 */
    if (by>=mpi_n*pt_n || y>=n1 || y>=b_len[by]) {
      printf("\n%d.%d Error statH: n1=%ld n=%d ii=%d bx.x,by.y,rr= %d.%lu %d.%lu %.3e",
        mpi_i, hbuf->block, n1, hbuf->n, ii,
        bx+mpi_i*pt_n, x-b_ofs[bx+mpi_i*pt_n], by, y, D(r));
      fflush(stdout);
    }
    y+=b_ofs[by]; /* 0..n1 */
    if (vvv&64 && x<x_max){	/* show 1..x_max H-lines */
      
      /* printf("\n x=%8lu y=%8lu h_xy=%s",x,y,outc(r)); --- old --- */
      { 
        printf("\n addh(%5lu,%5lu, ",x,y);
        outmpl(stdout,D(r));			/* +(a/b)^(1/2) */
#if VecType & 4
        printf(","); outmpl(stdout,Im(r));	/* +(a/b)^(1/2) */
#endif
        printf("):"); }				/* for maple */
        fflush(stdout);
    }
    if (by     ==bx%pt_n+mpi_i*pt_n)
                          statH[bx%pt_n].hnz_local_thread++; else
    if (by/pt_n==mpi_i)   statH[bx%pt_n].hnz_inter_thread++; else
                          statH[bx%pt_n].hnz_inter_node++;
    if (x>y) statH[bx%pt_n].sum12+=r;		/* summe of half H's */
    if (x<y) statH[bx%pt_n].sum21+=r;		/*   should be conjugated value */
    if (x>y) statH[bx%pt_n].hh12+=Norm(r);	/* summe of half H's */
    if (x<y) statH[bx%pt_n].hh21+=Norm(r);	/*   should be same */
    if (x==y) statH[bx%pt_n].traceH+=Re(r);	/* should be base invariant */
    rr=fabs(r);				/* cmplx => real */
    if(rr){
      if (x!=y) statH[bx%pt_n].hnz_nd++;
      else      statH[bx%pt_n].hnz_d++;
    }
    if (xy&XY_FLAG) { /* EOL, start next matrix line */
//printf("\nDBG nzx=%lu n_old=%lu ii=%d x=%lu",(ULong)ii+1-n_old,n_old,ii,x);
      if (statH[bx%pt_n].nzxmax<ii+1-n_old)
          statH[bx%pt_n].nzxmax=ii+1-n_old;
      if (statH[bx%pt_n].nzxmin>ii+1-n_old)
          statH[bx%pt_n].nzxmin=ii+1-n_old;
      x++; n_old=ii+1;
    }
  }
  fflush(stdout);
  setvbuf(stdout,NULL,_IONBF,0);
  return;
}

void statH_out(){		/* output statistics vvv&2 */
  int i;
  for(i=1;i<pt_n;i++){      /* sum all threads to statH[0] */
    statH[0].hnz_local_thread += statH[i].hnz_local_thread;
    statH[0].hnz_inter_thread += statH[i].hnz_inter_thread;
    statH[0].hnz_inter_node   += statH[i].hnz_inter_node;
    statH[0].sum12  += statH[i].sum12;
    statH[0].sum21  += statH[i].sum21;
    statH[0].hh12   += statH[i].hh12;
    statH[0].hh21   += statH[i].hh21;
    statH[0].traceH += statH[i].traceH;
    statH[0].hnz_nd   += statH[i].hnz_nd;
    statH[0].hnz_d    += statH[i].hnz_d;
    if (statH[0].nzxmax < statH[i].nzxmax)
        statH[0].nzxmax = statH[i].nzxmax;
    if (statH[0].nzxmin > statH[i].nzxmin)
        statH[0].nzxmin = statH[i].nzxmin;
  }
#ifdef CONFIG_MPI
  { double add[7], sum[7];
    add[0]=statH[0].hh12; add[1]=statH[0].hh21;
    add[2]=statH[0].traceH;
    add[3]=Re(statH[0].sum12); add[4]=Re(statH[0].sum21);
    add[5]=Im(statH[0].sum12); add[6]=Im(statH[0].sum21);
    MPI_Allreduce(add, sum, 7, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    statH[0].hh12  =sum[0];
    statH[0].hh21  =sum[1];
    statH[0].traceH=sum[2];
    statH[0].sum12 =C(sum[3],sum[5]);
    statH[0].sum21 =C(sum[4],sum[6]);
  }
  { double add[7], sum[7];
    add[0]=statH[0].hnz_local_thread;
    add[1]=statH[0].hnz_inter_thread;
    add[2]=statH[0].hnz_inter_node;
    add[3]=statH[0].hnz_nd;
    add[4]=statH[0].hnz_d;
    MPI_Allreduce(add, sum, 5, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    statH[0].hnz_local_thread=sum[0];
    statH[0].hnz_inter_thread=sum[1];
    statH[0].hnz_inter_node  =sum[2];
    statH[0].hnz_nd=sum[3];
    statH[0].hnz_d =sum[4];
  }  
  { ULong add[7], sum[7];
    add[0]=statH[0].nzxmax;
    MPI_Allreduce(add, sum, 1, MPI_UNSIGNED_LONG, MPI_MAX, MPI_COMM_WORLD);
    statH[0].nzxmax=sum[0];
    add[0]=statH[0].nzxmin;
    MPI_Allreduce(add, sum, 1, MPI_UNSIGNED_LONG, MPI_MIN, MPI_COMM_WORLD);
    statH[0].nzxmin=sum[0];
  }
#endif
  /* MPI_ToDo: sum over all nodes, may fail for multithreads */
  if (!mpi_i) {
    double sum_hnz=statH[0].hnz_local_thread
                  +statH[0].hnz_inter_thread
                  +statH[0].hnz_inter_node  ;
    printf("\n vvv&2 statistics for sparse matrix H:");
    printf("\n h_nz transfer local,pthread,mpi= %3.0f%% %3.0f%% %3.0f%%",
      statH[0].hnz_local_thread*100./sum_hnz,
      statH[0].hnz_inter_thread*100./sum_hnz, 
      statH[0].hnz_inter_node  *100./sum_hnz);
    printf("\n h_nz_ndiag>eps : %10.0f",statH[0].hnz_nd);
    printf("\n h_nz_diag      : %10.0f",statH[0].hnz_d);
    printf("\n h_nz_x_min     : %10lu",statH[0].nzxmin); /* v2.33 */
    printf("\n h_nz_x_max     : %10lu",statH[0].nzxmax); /* v2.27 */
    printf("\n h_nz/n1        : %13.2f", 1+statH[0].hnz_nd*1./n1);
    printf("\n trace of H     : %12.5e", statH[0].traceH);
    printf("\n sum_h12   hh12 : %s %15.8e", outc(statH[0].sum12), statH[0].hh12);
    printf("\n sum_h21   hh21 : %s %15.8e", outc(statH[0].sum21), statH[0].hh21);
    printf("\n sum h21-h12^*  : %s %15.8e", outc(statH[0].sum21-Conj(statH[0].sum12)), statH[0].hh21-statH[0].hh12);
    if(    fabs(statH[0].sum12 -Conj(statH[0].sum21))
    >1e-8*(fabs(statH[0].sum12)+fabs(statH[0].sum21)) ) E0("H not hermitian");
  }
}
/* --------------------------------------------------------------- */  




#ifndef NUM_AHEAD_LINES
#define NUM_AHEAD_LINES 1  /* minimum, bigger is better, nhm_buffer_lines */
#endif
/* see spins.h */
/*  thbuf *buf1; buf1=alloc_hbuf(); ... free_hbuf(buf1); */
thbuf *alloc_hbuf(){ 
  thbuf *buf;
  /* ToDo: dynamic mem alloc for hbuf->el if needed */
  buf=(thbuf *)mymalloc( sizeof(thbuf), NULL); /* no msg,  */
  if (!buf) {
    printf("\nERROR alloc_hbuf: malloc(thbuf=%.2e) failed, mpi=%d",
            1.*sizeof(thbuf), mpi_i);
    return 0;
  }
  buf->n=0;	/* ini num_elements_in_buf, returned/used by nhv */
  buf->num_ahead_lines=NUM_AHEAD_LINES; /* set before nhv called */ 
  buf->block=-1;  /* line belongs to block-nr, -1 means unused, before nhv */
  buf->v0=(tvec *)NULL; /* store/stat or nHn, set before nhv called */
  buf->v1=(tvec *)NULL; /* v1+=H*v0 if !0, set by the nhv caller */
  buf->bi.o=0;   /* option: 0=fullbase,1=smallest,2=ifsmallest */
  buf->bj.o=0;
  buf->bi.num=0;
  buf->bj.num=0;
#if 0
  buf->bi.ib= (tbase *)NULL;  /* force SIGSEGFAULT if [] accessed */
  buf->bi.is=(tphase *)NULL;
  buf->bj.ib= (tbase *)NULL;  /* force SIGSEGFAULT if [] accessed */
  buf->bj.is=(tphase *)NULL;
  /* really needed? */
  buf->bj.ib= (tbase *)malloc(sizeof(tbase) *MaxSBase);
  buf->bj.is=(tphase *)malloc(sizeof(tphase)*MaxSBase);
#endif
  buf->nhn=0; /* sum diagonal element */
  buf->rx=0;  /* <nHv> */
  return buf;
}

void *free_hbuf( thbuf *hbuf ){
#if 0
  if (hbuf->bi.ib) free(hbuf->bi.ib); /* avoid memory leak */
  if (hbuf->bi.is) free(hbuf->bi.is);
#endif
  return myfree( (void *) hbuf, NULL );
}


/* store elements of one H-line to the HBlock files on disk 
 * input: hbuf.i(b_ofs[blk]...,>node_ofs) + hbuf.n + hbuf.el[0..n-1]
 * output: H-file + Statistics
 * called by _nhv if m1&2 = storeh, 
 */
void hamilton_puth_line(thbuf *hbuf){ /* renamed from nhmline v2.26 */
  int ii, blk;
  ULong xy; int by; cplx r; HR_TYPE rr; /* h_put-variables */
  blk=hbuf->block;

  /* vvv&2 makes SH factor 2 slower on V490-sparcIV+ cache? v2.25? */
  if (vvv&2) add_statH(hbuf); /* H-statistics for debugging */

  /* b_hnz - number of hamilton-elements per compute block */
  /* hfmax - maximum file space for H per block */
  if (b_hnz[blk]+hbuf->n < hfmax) {	/* store Hij, ToDo: per line */
    for(ii=0;ii<hbuf->n;ii++) {
      /* store a matrix element (xy=y_idx,r=Hij) into the file buffer b_hbl
       * and write to a file if buffer is full,
       * buffer and file size is expensive, try to spare
       * keywords: hput puth put_h h_put
       */
      xy=hbuf->el[ii].jj; /* 0..(b_len[by]-1)[+XY_FLAG] checked in statH */
      by=hbuf->el[ii].blkj;  /* 0..(pt_n*mpi_n-1) */
#if defined(DEBUG)
#warning "DEBUGGING puth_line on"
      if ((xy&~XY_FLAG)>=b_len[by]) { printf("\nDBG blk %2d %2d %8ld",blk,by,xy); ERR(9999); }
#endif
      r =hbuf->el[ii].rr;
#if HRMAX>0                 /* use a index table to spare disk place */
      { /* get hr_index or store as new hr_index else */
        int j;
        /* Note: Norm(hr-r)<1e-13 can fail for some model systems, */
        /*       but hr==r is probably unecessary strong, no speed down */
        for(j=0;j<hr_len[blk];j++) if(Norm(hr[blk][j]-r)<=1e-13) break;
        /* here we should store max(hr[][]-r)<1e-13, but this takes time */
        if (j==hr_len[blk]) { /* not found */
           hr[blk][j]=r; hr_len[blk]++;
           if(hr_len[blk]>=HRMAX) err_fulltable(); /* OVL */
        }
        rr=(HR_TYPE)j;		/* store index to value */
      }
#else    
      rr=M_d2f(r);		/* store value direct */
#endif    
      b_hbuf[blk].rr[ b_hbl[blk] ]=rr;
      b_hbuf[blk].by[ b_hbl[blk] ]=(BY_TYPE)by;
      b_hbuf[blk].xy[ b_hbl[blk] ]=(XY_TYPE)xy;
// printf("\nDBG puth by,y,rr,flag= %d %lu %g %lu", by, xy&~XY_FLAG, r, xy&XY_FLAG);
      /* store in Hblock-buffer, ToDo: store only complete lines? */
      b_hbl[blk]++; /* count buffered elements, store at buffer too? */
      if(b_hbl[blk]==HBLen){   /* flush buffer to disk-file */
        if (my_fwrite((void *)&b_hbuf[blk],Hsize,1,blk%F_NUM)!=1) {
          hfmax=b_hnz[blk]; break; /* hard write error, stop writing */ 
        }
        b_hbl[blk]=0; /* reset buffer counter */
      }
    }
    b_snz[blk]+=hbuf->n; /* count stored elements, full ahead lines ToDo: full line */
    if (hbuf->i+hbuf->num_ahead_lines<=b_len[hbuf->block+mpi_i*pt_n])
         b_sn1[blk]+=hbuf->num_ahead_lines; /* count lines stored */
    else b_sn1[blk]= b_len[hbuf->block+mpi_i*pt_n];
  }
  if (b_hnz[blk]        <=hfmax
   && b_hnz[blk]+hbuf->n>=hfmax
   && hfmax
   && mpi_i*pt_n+blk<4) // assuming that this error will appear on all nodes
     ERR(630); /* v2.33, maxfile reached, stop writing */
  if (b_hnz[blk]+hbuf->n<b_hnz[blk]) ERR(610); /* counter overflow, warn */
  b_hnz[blk] += hbuf->n;

  /* check if all elements of this matrix line have been stored */
  return;
}


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 * do b2i = scfg2idx with buffered cfgs,
 *   b_smallest is already called within nhm (ToDo: do it to here)
 * called by hamilton_nhv
 * if num_ahead_lines=1 it is per line, but since v2.30 its for
 *    num_ahead_lines lines.
 * must be called by each mpi-process and thread to avoid deadlock v2.33
 * careful use MPI within threaded part here!
 * ToDo: call b_smallest/sq(norm) from here cached/streamed
 */
void nhm_line(thbuf *hbuf) { /* cfg-to-idx via MPI for a H-line */ 
  int ii, i2=0, blk, blk2;
  cplx sum=0;
  static mcplx r0b[B_NUM][NUM_AHEAD_LINES+1]; /* buffer for <nHv>[] */
  mcplx    *r0=r0b[hbuf->block];

  blk = hbuf->block + mpi_i*pt_n;

  /* cfg-to-idx dont work for storeh2, no l1 */
  for (ii=0;ii<hbuf->n;ii++) {  /* hbuf->n <= NZXMAX*NUM_AHEAD_LINES */
    hbuf->el[ii].blkj=scfg2blk(hbuf->el[ii].bj);
    if (mpi_i==hbuf->el[ii].blkj/pt_n) {
      blk2=    hbuf->el[ii].blkj/*%pt_n*/;
      hbuf->el[ii].jj |= b2i(hbuf->el[ii].bj)  /* b2i=0..(node_n1-1) */
                        -b_ofs[blk2]+node_ofs; /* jj=0..(b_len-1) */
    }
    else hbuf->el[ii].jj &= XY_FLAG; /* clear for better debugging */
  }
  /* ToDo: better do it after _nhv call when needed??? */
  if(hbuf->v0){ /* only needed for <nHv> */
    for (ii=0;ii<hbuf->n;ii++) {
      if (mpi_i==hbuf->el[ii].blkj/pt_n)
          sum += hbuf->el[ii].rr * M_f2d(hbuf->v0->v[
                 hbuf->el[ii].blkj % pt_n][
                 hbuf->el[ii].jj & ~XY_FLAG]);
      if (hbuf->el[ii].jj & XY_FLAG) {
          r0[i2]=sum; i2++; sum=0; /* store diag-block <i2|H|v0> */
      }
    }
  }
#ifdef CONFIG_MPI  /* if mpirun fails, try lamclean+ipcs+ipcrm */
  { /* ToDo: send only the relevant parts, not full hbuf */
    /* ToDo: pt_n>1 COMM_WORLD -> pt_n*Worlds? against deadlock MPI_Comm_Create */ 
    /* ToDo: we send hbuf including bi,bj which we dont need (lot of size) */
    int i1, i2, i3, id, node2, node1; MPI_Status status;
    /* ToDo: simplify to array[2] idx 0 for send, 1 for recv */
    /* nzxmax per block? reduce buffer space!? */
    /* !!! ToDo: [NZXMAX*NUM_AHEAD_LINES] => malloc[max+1024k] : size/MPIs
     *     but receive-buffer is to small
     *  OR  error-cont and restart with bigger buffer if its to small 
     */
#ifndef HNZ_MPI_BUFSIZE
/* ca. (NZXMAX/mpi_n+2)=NUM_AHEAD_LINES ... NZXMAX*NUM_AHEAD_LINES */
#define HNZ_MPI_BUFSIZE (NZXMAX*NUM_AHEAD_LINES)
#endif
    static mcplx r1b[B_NUM][HNZ_MPI_BUFSIZE+1], /* buffer for <nHv>[] */
                 r2b[B_NUM][HNZ_MPI_BUFSIZE+1];
    mcplx    *r1=r1b[hbuf->block],
             *r2=r2b[hbuf->block];
    static ULong jj1b[B_NUM][HNZ_MPI_BUFSIZE],
                 jj2b[B_NUM][HNZ_MPI_BUFSIZE];
    ULong *jj1=jj1b[hbuf->block], /* pointer to local buffer */
          *jj2=jj2b[hbuf->block];
    static struct tsend {  /* send buffer */
      tbase  bj;
      short  blkj; /* only needed for pt_n>1, ToDo: remove?  memory! */
//      ULong  jj;   /* only for v0!=0 XY_FLAG, ToDo: remove!? memory! */
    } bsend[B_NUM][HNZ_MPI_BUFSIZE],
      brecv[B_NUM][HNZ_MPI_BUFSIZE];
    /* ToDo: bsend[B_NUM*NZXMAX*AHEAD] + startidx[block]=MAX/pt,len[]  */
    if (mpi_n*pt_n > 1l<<(8*sizeof(short))) ERR(9999);

    /* ensure balanced number of MPI calls outside, using hbuf->n=0 */
    /* v2.33 n1=3 mpi*pt=4*1,2*2, must also be called for b_len[]=0 */
    for (id=1;id<mpi_n;id++) {
      node1=(mpi_i      +id)%mpi_n;  /* dest node+ii */
      node2=(mpi_i+mpi_n-id)%mpi_n;  /* src  node-ii */
      /* ring transfer is faster but scales worse in later versions */
      /* pairwise transfer is faster for intra-node but dont work on odd n */
      /* copy the part of data which is relevant for the destination */
      for (i1=ii=0;ii<hbuf->n;ii++) {
        blk2=    hbuf->el[ii].blkj;
        if (blk2/pt_n != node1) continue; /* check for destination node */
        if (i1==HNZ_MPI_BUFSIZE) ERR(9999); /* HNZ_MPI_BUFSIZE is to small */
        bsend[hbuf->block][i1].bj   = hbuf->el[ii].bj;   /* sym. spin cfg */
        bsend[hbuf->block][i1].blkj = hbuf->el[ii].blkj; /* block-no. */
        /* next is only needed for unused storeh, only for XY_FLAG */
//        bsend[hbuf->block][i1].jj   = hbuf->el[ii].jj; /* idx of cfg??? */
        i1++;
      }
      pthread_mutex_lock(&token_ring[hbuf->block]);
      /* ToDo: wait for other threads with tokenring and send all data
       *   only from thread 0, after that start all threads with tokenring 
       */
      mpi_syncs++; mpi_traffic += i1*sizeof(struct tsend); /* shared! */
      mpi_syncs++; mpi_traffic += i1*Lsize; /* count send only */
      if (i1>(int)mpi_maxlen) mpi_maxlen=i1;
if(vvv&4 && mpi_i<8 && mpi_syncs<64)
      printf("\nMPI_a %3d.%d S= %3d R= %3d send=%9d",mpi_i,hbuf->block,node1,node2,i1);
      /* MPI_STATUS_IGNORE causes SEGFAULT on Tru64-5.1B DMPI-1.1 */
      i2=HNZ_MPI_BUFSIZE;
      /* ToDo: check with n1=1..3 and 2*2 or 4*1 mpi-threads */
      /* ToDo: MPI_Sendrecv_replace but no reduced send-Data-length? */
      /* ToDo: for dyn. buffer send oldlen as flag, oldlen recv = size*2 */
      MPI_Sendrecv(bsend[hbuf->block],i1*sizeof(struct tsend),MPI_BYTE,node1,0,
                   brecv[hbuf->block],i2*sizeof(struct tsend),MPI_BYTE,node2,0,MPI_COMM_WORLD,&status);
      MPI_Get_count(&status,MPI_BYTE,&i2); /* less data received? */ 
      i2/=sizeof(struct tsend);            /* get num data recvd */
      pthread_mutex_unlock(&token_ring[hbuf->block+1]);

      /* we need to send .blkj .bj [.rr] */
      /* ToDo: b2i(bj,&blk) should return idx + blkj to reduce traffic??? */
      for (ii=0;ii<i2;ii++)  /* count blk_hnz (NZX*AHEAD) */
        jj1[ii] = b2i(brecv[hbuf->block][ii].bj) /* b2i=0..(node_n1-1) */
               -b_ofs[brecv[hbuf->block][ii].blkj]
               +node_ofs;                        /* jj=0..(b_len-1) */
      if (hbuf->v0)
      for (ii=0;ii<i2;ii++)  /* i2 <= HNZ_MPI_BUFSIZE, simple! */
        r1[ii]  = hbuf->v0->v[brecv[hbuf->block][ii].blkj%pt_n][jj1[ii]];

      pthread_mutex_lock(&token_ring[hbuf->block+pt_n]);
if(vvv&4 && mpi_i<8 && mpi_syncs<64)
      printf("\nMPI_b %3d.%d S= %3d R= %3d send=%9d",mpi_i,hbuf->block,node2,node1,i2);
      /* reduce MPI traffic to a minimum */
      if (!hbuf->v0) MPI_Sendrecv(jj1,i2*Lsize,MPI_BYTE,node2,hbuf->block,
                                  jj2,i1*Lsize,MPI_BYTE,node1,MPI_ANY_TAG,MPI_COMM_WORLD,&status);
      else           MPI_Sendrecv(r1,i2*sizeof(mcplx),MPI_BYTE,node2,hbuf->block,
                                  r2,i1*sizeof(mcplx),MPI_BYTE,node1,MPI_ANY_TAG,MPI_COMM_WORLD,&status);
      pthread_mutex_unlock(&token_ring[(hbuf->block+pt_n+1)%(2*pt_n)]);
//      pthread_mutex_unlock(&token_ring[(hbuf->block+1)%(pt_n)]);
      /* copy results from other node */
      if (!hbuf->v0)
        for (i3=ii=0;ii<hbuf->n;ii++) { /* idx_cfg | XY_FLAG */
          if (hbuf->el[ii].blkj/pt_n==node1) hbuf->el[ii].jj |= jj2[i3++];
      }
      if (hbuf->v0) /* decompress r2 and add to r0[]=<i2|H|v0> */
      for (i1=i3=ii=0;ii<hbuf->n;ii++) { /* <i2|H|v0> ??? */
        if (hbuf->el[ii].blkj/pt_n==node1)
                           r0[i1] += hbuf->el[ii].rr * r2[i3++];
        if (hbuf->el[ii].jj & XY_FLAG) i1++; /* next HLine <= AHEAD_LINES */
      }
    } /* id=0..mpi_n-1 */
  } /* MPI */
#endif  /* CONFIG_MPI */
  if (hbuf->v0) /* store <i2|H|v0> */
    for (ii=0;ii<hbuf->num_ahead_lines;ii++) hbuf->el[ii].rr = r0[ii];
  /* if (v0) return hbuf[i].rr = <i|H|v0> */
}

/* -----------------------------------
 * r*(j|j><i|i), Hij=r*|j><i|, i=n, j=m, hbuf->bi,bj ??? 
 *   v2.28: do cfg-to-idx later in nhm_line
 * m1=mode (removed, same action for all modes, v2.34)
 *  hbuf->i is the cfg index on the node (0..b_len-1) # less bits
 *  cfgj is the cfg modified by operator H, r*cfgj=H(i,j).bi
 * ToDo: store para_index instead of r
 * ToDo: store to hbuf only, call b_smallest later! (better caching, +FPGA) 
 * -----------------------------------------------------------
 */
void nhm(tbase cfgj, cplx r, thbuf *hbuf) {
  ULong j=0;
  int blk, blkj;
  struct sbase *bi=&(hbuf->bi); /* symcfg(bi) = input */
  struct sbase *bj=&(hbuf->bj); /* buffer for symcfg(bj=Hij*bi) */
  if (r==0) return;    /* 32+8: true=0 (only if para==0) */
#ifdef CONFIG_ANSATZ
  if (!ansatz_test(cfgj)) return;  /* ToDo: buffered call? */
#endif
  bj->scfg=cfgj; /* default return if bj==bi */
  blkj=blk=hbuf->block; /* needed??? */
  /* get j and multiply r with the quotient of symmetrie cofficients (i|j) */
  j=hbuf->i;    /* index of configuration, where H is acting on (b_len) */
  // ToDo: cache 2..AHEAD cfgs and call b_smallest blockwise
  // ToDo: seperate diag-buffer for num_nondiag,rr  + n2(?)
  if (bi->scfg!=cfgj)
  { /* false: 34+6=5% 32+8=4% speedup=4-5% Hnondiagonal 1/nzx? */
    bj->o=0;
    /* get minimum symmetric config bj from cfgj, speedup by buffering? */
//    bj->norm2=1; bj->scfg=cfgj;
    b_smallest(cfgj,bj); /* 98% cpu-time */
    /* if(phase==0): 32+8 true=0..3% sym_k-depending */
    if (phase_eq(bj->sign,0)) return; /* b2i fails else */
    r *= bj->sign;
    /* 20+20: true=(.2e6/431e6)^2<1%, see n2stats */
    if (bi->norm2!=bj->norm2) r*=sqrt(bi->norm2/(double)bj->norm2);
  }
  if (bi->scfg==bj->scfg) { hbuf->nhn+=D(r); } /* Hii can not be cplx */
  /* old: storeh+storeh2+fulldiag only, new v2.28: all */
  { /* use n<0 as flag? for storing to hbuf? storing hnz/n1 elements */
    /* obsolete: subblock blk = b_idx2blk(j); */
    /* ToDo: what happens for m1&4 (storeh2) ??? put all data in first block! */
    hbuf->el[hbuf->n].rr=r;    /* fill H-buffer of this thread */
    hbuf->el[hbuf->n].bj=bj->scfg; /* v2.27, rename to scfg */
    hbuf->el[hbuf->n].jj=0; /* v2.30 reset to carry XY_FLAG (set in _nhv) */
#if 0
  { int ii;
    /* data compression after b_smallest! */
    /* search for scfg in stored H line */
    for (ii=hbuf->next_line;ii<hbuf->n;ii++)
      if (hbuf->el[ii].bj==bj->scfg) break;
    if (ii==hbuf->n){ hbuf->n++;if(hbuf->n>=NZXMAX*NUM_AHEAD_LINES)ERR(9200);
    }                            /* not stored */
    else { hbuf->el[ii].rr+=r; } /* already stored */
  }
#else
    hbuf->n++;  if (hbuf->n>=NZXMAX*NUM_AHEAD_LINES) ERR(9200);
#endif
    /* sq: 32+8  hnz/n1= 27.1663  -8.22686824 162s 164s       */
    /* sq: 32+8  hnz/n1= 27.2569  -8.22686824 159s 164s +0.3% */
    /* sq: 15+15 hnz/n1= 31.582 31.79  (3..51)   compressed 3m20s     */
    /* sq: 15+15 hnz/n1= 31.834 32.04 (15..61) uncompressed 3m15s +1% */
    /* spin1/s1chain.gpl: N=20 20+20 19.30 vs. 19.35 +0.25% */
    /* s1chain+AddSS: N=16 8+8 n1=19 nzx=11.00 vs. 1927.95 Bug! SS*Nw */
    /* sq+AddSS: 4+36 n1=632 nzx=115.54 vs. 159.02 */
  } 
  return;
} 

/* ----------------------------------------------------------------
 * m1=mode: 0=<n|H*v0> (return nhn too)
 *          1=<n|H*n>   obsolete! used for 2x2-algo
 *            until v2.25 for uududd==uuddud <nHn>=0, but its != 0
 *            result correct? differences? 3+3 diverged (n1=3)
 *  uuuddd/q6, uududd/q12, ududud/q2
 *  [1/2, q(1/2), 0] [q(1/2), 1/2, q(3/2)] [0, q(3/2), -3/2] trace=-1/2
 *  nhn: 1/2  1/2 -3/2  8It e0=-2.118033988749896
 * m1=mode: 2=storeh,
 *          4:buf_..=H|n> without_sym+idx ... (called by storeh2)
 *          8:buf_..=H|n> full
 * ToDo: always return nhn (by pointer arg?) do m1=0,1 in one step! m1==1 obsolete
 *           i + nhn in buffer[]?
 * sample: 6sites
 *   |i>=|sym_ududud>=(ududud+dududu)/q2                    - different norm2
 *   |i>=|sym_ududud>=(ududud+dududu+ududud+dududu+...)/q2/6  # not q12!
 * norm2: 1*2 
 * ToDo: MPI+FPGA:
 *       if i%num_ahead_lines return from ahead_hbuf to hbuf
 *       else  fill aheadhbuf=hbuf[NZXMAX*num_ahead] + call scfg + b2i, (wop analog)
 *       return 1st line from aheadhbuff to hbuf
 *       on overflow reduce num_ahead_lines or go to num_ahead_lines before overflow
 *       aheadhbuf=hbuf+.startel pointing to el line i
 * ToDo: return <nHn> so that nhv(m1=0);nhv(m1=1) can be replaced by nhv(m1=0) 
 * ToDo: put i into hbuf (reduce arguments)
 * ToDo: split nhv to hbuf=H*|n> (Helements) and v0*hbuf
 * input:
 *    l1[0..(node_n1-1)]
 *    hbuf->blk= 0..(pt_n-1) // add mpi*pt_n if needed (b_ofs+b_len)
 *    i  = line index (better, use range to reduce calls) b_ofs[blk]...
 *             move to hbuf->i  = 0..(node_n1-1)
 *    hbuf->v0 = NULL(vector.c,spins.c) or v0(xval.c_rh,diag.c) (can be moved to hbuf?)
 *    hbuf = work buffer (something set?)
 *    m1 = mode (see above)
 * modified: hbuf.{}
 * calling: nhm+puth_buf to store stream of elements
 * output: hbuf.{(rx,nhn if v0 set),n,el[0..n-1].{bj,jj,rr}}
 * return: nhv (if v given) ToDo: return in hbuf!?
 * we need to work on range of i=i1..i2 to have blocking and less MPI-overhead
 * must be called by each mpi-process and thread to avoid deadlock (also n1=0)
 * called by:
 *   diag.c:diag2_thread                  < n|H|v0> <n|H|n> m1=0 v1=0?
 *   xval.c:wh                            <v0|H|v0>         m1=0 v1=0?
 *   xval.c:whq                           |H*v0-rh*v0|      m1=0 v1=0?
 *   spins.c:hamilton_geth_block=hamilton i<len[blk] v1+=H*v0 <v0Hv0> m1=0
 *   vector.c:randvec                     i<len[blk] <nHn> v0=0 m1=0
 *   spins.c:storeh_block=storeh          i<len[blk]       v0=0 m1=2
 *   spins.c:storeh2                      i<n1             v0=0 m1=4 (no valid l1)
 *   spins.c:fulldiag         hbuf[nHm..] i<len[blk]       v0=0 m1=8
 * nahead: num H lines to store, nahead=1..NUMAHEAD (1=old version)
 *         0 means no but call MPI block for balance v2.35
 *         the idea of nahead>1 is to reduce hamilton_nhv calls
 * ---------------------------------------------------------------- */
cplx hamilton_nhv(ULong i, ULong nahead, thbuf *hbuf, int m1){
  tbase ib1, ib2;     /* target and workconfig */
  int ip1,ip2,ii,i1,ia,ib; /* worksites, workindizes */
  ULong iahead;
  int next_line; // pointer to el[] start of ne line
  cplx r1;   /* ToDo: old: sum = nhv = <i|H|v0> moved to [2*i+1].rr ??? */
  double rp,
         sun; /* sun=diagonal sum, onlyZ? */
#if CM(0,1,1)
  int j, is;
#endif
  hbuf->bi.o=0;                        /* b_smallest option */

{
 hbuf->n=0; /* number of stored elements in ahead buffer */
 next_line=0;

 /* get H lines ahead to group nonlocal memory access */
 for (iahead=0;iahead<nahead;iahead++) {

  /* dont do calculations on empty MPI blocks (for balance) */
  if (i+iahead>=b_len[mpi_i*pt_n+hbuf->block%pt_n]) break;

  hbuf->bi.scfg=ib1=getl1(i+iahead+b_ofs[mpi_i*pt_n+hbuf->block%pt_n]
                            -node_ofs); /* bi.scfg ?? */
  hbuf->i=i+iahead;
  sun=0;

// #warning "Debugging code"
// if (i>=b_len[hbuf->block%pt_n+pt_n*mpi_i]) ERR(9999); /* */

  /* calling b_smallest here, adds only n1/hnz computing time, check: ToDo */
  if(!(m1&4)){  /* ToDo: remove if()! call b_smallest in every case!? */
    b_smallest(ib1,&(hbuf->bi)); /* norm2 needed (+n1/hnz cpu-power) */
    /* ToDo: shift to nhm_line() */
  }
#ifdef AddSS	/* only ok for JJ-model !???, AddSS must be a float number */	
    /* add AddSS(SS-ZZ)=sum_(i,j).5(SipSjm+SimSjp) to Ham, max. nu*nd terms? */
    for(ia=0;ia<nn;ia++){    ip1=geti(ib1,ia);
     if(ip1==1 || ip1==2) sun+=.5*AddSS;  /* diagonal part Z^2 */
     for(ib=ia+1;ib<nn;ib++){ ip2=geti(ib1,ib);
       if((ip1^ip2)==3) nhm(Flip(ib1,ia,ib),AddSS,hbuf);
    } }
#endif
/* ??? should we use symmetry at this point for speed up ??? i-j => pi-pj 
 * with smallest fitting  resulting configuration? */
  for(i1=0;i1<nw;i1++){
    ia=iw[i1][0];ip1=geti(ib1,ia);
    ib=iw[i1][1];ip2=geti(ib1,ib);
    rp=para[ww.wj[i1]];  /* exchange parameter J */
#if defined(CONFIG_S1SYM) && 1 /* N=8 s=5/2 7*faster */
    /* for composed spins take only lowest up(down) (ex: s=5,m=1: uuuudd)
     * and correct by a factor num_up(down) to speed up of (2s)^2/4
     *  ToDo: emulate by S(+/-,idx_composed)
     *  S^+|j,m>=((j-m)*(j+m+1))^(1/2)|j,m+1>
     *  S^-|j,m>=((j+m)*(j-m+1))^(1/2)|j,m-1>
     * Why C does not support break(2) to leave 2 loops?
     * subspins of composed spins must be ordered: uuu...ddd
     * ToDo: test mixed spin-lengths
     */
    if (sym.num_lm!=nn && rp && ip1!=ip2) { /* ud - du flip? */
     int nm[4],i,node;
     if (sym.ls[ia]>1 && sym.ls[ib]>1 && sym.li[ia]==sym.li[ib]) rp=0;
     for (node=0;node<2 && rp;node++){  ii=iw[i1][node];
       if (sym.ls[ii]>1) {  /* composed spin detected */
        if (sym.li[ii]!=0 && sym.li[ii]!=sym.ls[ii]-1) rp=0;
        else { /* member 0 or s-1 */
          nm[0]=nm[1]=nm[2]=nm[3]=0; /* later add 3. - .3 flip */
          for (i=0;i<nn;i++) if (sym.lm[i]==sym.lm[ii])
            for (;i<nn;i=sym.ln[i]) nm[geti(ib1,i)]++;
          if (sym.li[ii]==0)            rp*=nm[2];  /* S^+|....d> */
          if (sym.li[ii]==sym.ls[ii]-1) rp*=nm[1];  /* S^-|u....> */
        }
       }
     }
    }
#endif
    ii=ip1^ip2;
#if   CM(1,1,0)		/* SiSj+ADham1*ninj  shift -3/4,1/4 to -1,0 etc. */ 
# if   IFz		/* +ninj/4 */
#  if   Ham1
    CM({},if(ip2 && ip1),{})sun+=(.25*h_zpara+Add34)*para[ww.wj[i1]];
#  endif
# endif
    if(ii==3){
# if     IFz
      sun-=.5*h_zpara*para[ww.wj[i1]];		/* SiSj-ninj/4 */          
# endif
      if(rp!=0){ /* nondiagonal terms (sym: partly diagonal) */ 
        r1=h_xypara*.5*rp;
        /* signum of imag depends on whether ud or du */
        IFww( r1*=C(ww.co[i1],ww.si[i1]*((ip2==1)?-1:1)); )
        nhm(Flip(ib1,ia,ib),r1,hbuf);
      }
    }
# ifdef	DEBUG	
    if(err_fatal){ E0("nhv SxSx");printf(" nw=%d i=%d j=%d",i1,ia,ib); }
# endif
#endif  /* CM(1,1,0) */
#if   CM(0,1,1)		/* t_{ij} cp_i c_j     t=t_{ij} t^*=t_{ji} */
    if( CM( ii , ii>0 && ii<3 , ii ) ) {
      /* valid: .u  ,only tU: ud u3 ToDo: faster with tU only ii>0 ??? */
      is=parity(ib1 & p_mask[ia][ib]); /* fermionic signum without sites */
      for(j=0;j<2;j++)if(ii & (j+1)){ /* up or down */
        r1=is*para[ww.wt[i1]];        /* signum*t_{ij} */
        if( (j==0) ? ip1&2 : ip2&1 ) r1=-r1; /* fermionic sign correction */
        IFww( r1*=C(ww.co[i1],ww.si[i1]*((ip2 & (j+1))?-1:1)); )
        nhm(tbase_xor(tbase_xor(ib1,ia,j),ib,j),r1,hbuf);
      }}
#endif
  }
#if CM(0,0,1)
#if Nv             /* four-fermion-term, inter site repulsion+exchange */
    for(i1=0;i1<Nw;i1++){          /* wv: uu,ud,du,dd,u3,d3,3u,3d,33 */
      ia=iw[i1][0];ip1=geti(ib1,ia);
      ib=iw[i1][1];ip2=geti(ib1,ib);
#if Nv & 4    /* code activated if Nv is 4 for performance reasons */
      /* hole-hole repulsion, hole = empty site = (1-niu*nid) */
      /* this Hamiltonian is used by Fatiha Ouchni, Jul2004 */
      if (ip1==0 && ip2==0) sun+=para[ww.wh[i1]];
#endif
#if Ham3
#error "Code not verified. Is the sign correct?"
      if(ip1*ip2==0)if((ip1^ip2)==3) /* .3,3. -JH Oles */
        nhm(Flip(ib1,ia,ib),-para[ww.wj[i1]],hbuf);
        /* ensures symmetry */
#endif  
      /* ToDo: more general would be a 4x4 matrix ww.repulsion with
       *       sun+=para[ww.wv[i1]]*ww.repulsion[ip1][ip2]; (less if's)
       * transform ip1,ip2 to symmetric quantum numbers
       *           nij_sum=ni+nj-2=-2..2, nij_diff=ni-nj  z=nu-nd=-2..2 */
      if((ip1>0)&&(ip2>0)){
        if(ip1&1)if(ip2&1)sun+=para[ww.wv[i1]]+para[ww.wj[i1]];/* uu  U+JH */
        if(ip1&2)if(ip2&2)sun+=para[ww.wv[i1]]+para[ww.wj[i1]];/* dd       */
        if(ip1&1)if(ip2&2)sun+=para[ww.wv[i1]];		/* ud   U    */
        if(ip1&2)if(ip2&1)sun+=para[ww.wv[i1]];		/* du        */
        if((ip1^ip2)==3)					/* ud,du  JH */
         nhm(Flip(ib1,ia,ib),para[ww.wj[i1]],hbuf);}
    }
#endif
    for(i1=0;i1<nn;i1++){ ip1=geti(ib1,i1);
#if Ne
      if(ip1&1)sun+=para[ww.we[i1]]; /* E nu */
      if(ip1&2)sun+=para[ww.we[i1]]; /* E nd */
#endif
#if Ham2
      sun+=((ip1%3)?-.25:.25)*para[ww.wu[i1]];	/* U(nu-.5)(nd-.5) */
#else
      if(ip1==3)sun+=para[ww.wu[i1]];		/* U nu nd */
#endif
    }
#endif
#ifdef HFIELD
    for(i1=0;i1<nn;i1++){ ip1=geti(ib1,i1);
      sun-=((ip1&1)-(ip1>>1&1))*para[ww.wf[i1]]/2;
    }
#endif

#define S_DEBUG(a) a
#ifdef SSANISO
#ifndef CONFIG_S1SYM
# error "defining SSANISO without CONFIG_S1SYM makes no sence"
#endif
   { /* this code was introduced by Reimar Schmidt, Jan 2004 */
     int k,i2; double oZi; static unsigned int dbg=0;
     /* oZi - eigenvalue of S^z_i of a composed spin
      *  Di - ww.wd index to float parameter (defined in daten.def)
      *  code revised by Joerg v2.20 */
     /* num_lm is N for a s=1 chain of length N build by NN=2*N spins s=1/2 */
     for(k=0;k<sym.num_lm;k++){  /* k = index of a composed spin-s object */
       for(oZi=0.0,i2=NN,i1=0;i1<nn;i1++)  /* sum S^z(i) over a composed spin */
         if(sym.lm[i1]==k) { /* check if spin i is member of composed spin */
           oZi+=(3-2*geti(ib1,i1))/2.0;
           if (i1<i2) i2=i1; /* get index of 1st member of composed spin */
         }
       sun+=oZi*oZi*para[ww.wd[i2]]; /* add to the diagonal element */
       S_DEBUG( if (dbg++<10) printf("\n dbgD k=%2d oZi=%4.1f d[i]=%.2f %s", k, oZi, para[ww.wd[i2]],ccfg(ib1));)
     }
   }
#endif
#if CM(1,0,0)
   if (nq) {  /* four-site ring term, see op_rot4() too */
     int i2;
   /* sample:
    *             1      J12=J23=J34=J41, J13, J24
    *           / | \    Q1234 = four-site term (cyclic exchange)
    *          /  |  \   ferri state: S_total=0, S13=S24=1 EF=-2J12+J13/2+Q1234
    *         4---+---2      J24 term? (Nedko-Solution)
    *          \  |  /
    *           \ | /
    *             3
    */
     for (i1=0;i1<nq;i1++) {
       ib2 = ib1;
       for (i2=0; i2<4; i2++) {
         ip1 = geti(ib1, iq[i1][i2]);             /* take site i */
         ib2 = puti(ib2, iq[i1][(i2+1)%4], ip1);  /* put to site i+1 */
       }
       /* rotation */
       nhm(ib2,para[ww.wq[i1]],hbuf);

       ib2 = ib1;
       for (i2=0; i2<4; i2++) {
         ip1 = geti(ib1, iq[i1][i2]);             /* take site i */
         ib2 = puti(ib2, iq[i1][(i2+3)%4], ip1);  /* put to site i-1 */
       }
       /* inverse rotation */
       nhm(ib2,para[ww.wq[i1]],hbuf);
     }
   }
#endif

#ifdef EDIAG
    sun+=EDIAG;  /* add 1000.0 if energy is -1000.00365 for better results */
#endif
    /* diag element is not always the last because sym(H*v) could be diag */
    nhm(ib1,sun,hbuf);    /* store/add diagonal value <nHn> */

  /* ToDo: put it as first element and add diagonalterm later? n2?
   *   use extra table for num nz per line? or first entry with num nzx */
  if (hbuf->n - next_line==0) { /* we need minimum 1entry for EOL flag */
    hbuf->el[hbuf->n].jj=0;
    hbuf->el[hbuf->n].blkj=hbuf->block+mpi_i*pt_n; /* this node! */
    hbuf->el[hbuf->n].rr=0; /* dummy element */
    hbuf->el[hbuf->n].bj=ib1;
    hbuf->n++;
  }
  next_line=hbuf->n; /* store EOL position */
  hbuf->el[hbuf->n-1].jj |= XY_FLAG; /* mark EOL */
//  printf("\nDBG EOL @ %d", hbuf->n);
} /* fill buffer ahead */
    /* m1&4: storeh2;
     * store EOL to hbuf for line detection;
     * do scfg2idx(remote) and if v0 sum r*v0[blkj][j](remote) */
    if (!(m1&4)) nhm_line(hbuf);
    if   (m1&2)  hamilton_puth_line(hbuf); /* write to H-file */
} /* buffer filled */

  hbuf->i=i+nahead;   /* index of configuration 0..b_len[blk]-1, set outside!? */
#if 0
    for (ii=0; ii<hbuf->n; ii++) {
      if (hbuf->el[ii].jj==i && hbuf->el[ii].blkj==hbuf->block+mpi_i*pt_n)
          hbuf->nhn += D(hbuf->el[ii].rr);    /* <nhn> */
      if (hbuf->el[ii].jj&XY_FLAG) {
        i++; // next matrix line
      }
    }
    if (hbuf->v0) sum= hbuf->el[i%hbuf->num_ahead_lines].rr;
#endif
  return(0);                /* old: return nHv or nH1(?), new return err */
}


/*
 * void-function for pthreads, read a complete H-block and do v1=H*v0
 * return hbuf->rx=<v0|H|v0> real
 * input: hbuf->{block,v0,v1}
 * working:  ToDo: MPIperLine hbuf.n,hbuf.el[].jj,blkj,rr=H|v0>
 * output: hbuf->rx=<v0|H|v0>, v1+=H*v0
 *  called by hamilton() for diag.c:inc1
 *
 * xx is only correct for the first block ?! 
 * h_get was put into geth_block (v2.26, keywords: h_get, hget, geth, get_h)
 * for better speed, this part should be read in ca. 1000 values from vy, 
 *   than from vx0, than store to vx1 to use the CPU cache optimal
 * ToDo: h_ropen/close h_files before calling this
 * ToDo: MPI, at the moment we send mpi_n*data, so it does not scale
 *       this is only for simplicity and will be changed in future
 *
 */
void hamilton_geth_block( void *ptr ) { /* thbuf is in prinzip not needed! */
  ULong i, in, ii, ihblk, nhblks; thbuf *hbuf=(thbuf *)ptr;
  int block=hbuf->block, by=0; /* block index for x and y */
  static volatile int itr=0;
  ULong xx, xy;
  mcplx *v0x,*v1x;   cplx r, sum_ry;
  struct thxy *tbh;                      /* H storage buffer */
  static struct mpi_Hbuff {   /* Data to be transferred between Nodes */
    mcplx  vr[HBLen];  /* results from one remote node */
    int    iy[HBLen];  /* index list for faster sorting back results */
    mcplx  vy[HBLen];  /* results for all nodes */
  } mpi_thb[2*B_NUM];  /* two Hbuffers for each thread */
#ifdef CONFIG_MPI
  static struct mpi_Hbuff1 { /* Data pairs transferred between Nodes */
    int yy;  /* index for one remote node << pt_n_bits + block no, 32bit */
  } mpi_th1[2*B_NUM][HBLen];  /* two Hbuffers for each thread */
  if ((int)sizeof(struct mpi_Hbuff1)*8<n1_bits+pt_n_bits) ERR(9999);
#endif
  if (block>=pt_n) ERR(9999);
  tbh=&(b_hbuf[block]); /* use pointer to H buffer for speedup */
  v0x=hbuf->v0->v[block];
  v1x=hbuf->v1->v[block];
  xx=0;      /* 0..b_len-1 */
  in=b_snz[block]; /* number of stored elements in HBlock */
  if (block==0) itr++; /* count iterations for debugging */
  sum_ry=0;
  nhblks=(b_snz[block]+HBLen-1)/HBLen; /* number of Hblocks written */
  if (nhblks>bsnz_max) ERR(9999);
  ii = 0; /* number hnz element of block, ii<HBLen */
  i  = 0; /* i is only the hnz counter, i<b_len[gblock] */
  for (ihblk=0; ihblk<bsnz_max; ihblk++) { /* ensure mpi balance */
// if (vvv&4 && itr<4 && (ihblk<10 || ihblk+10>bsnz_max)) /* DEBUGGING */
//   printf("\nMPI_0 itr= %3d thread %2d ihblk=%lu %lu %lu i=%lu %lu",
//          itr, mpi_i*pt_n+block, ihblk, nhblks, bsnz_max, i, in);
#if defined(STOREH) && STOREH==0  /* ToDo: replace this */
    if (ihblk<nhblks) pipe_read((char *)tbh, Hsize, block); /* reduce overhead */
#else
    if (ihblk<nhblks) h_read(block); /* ToDo: return_pointer to buffer */
#endif
    /* copy random scattered local v0y data to the mpi buffer */
    for (ii=0l; i+ii<in && ii<HBLen; ii++) {
      xy = (ULong)tbh->xy[ ii ]; /* incl. XY_FLAG */
      by =        tbh->by[ ii ]; /* block<mpi_n*pt_n */
#ifdef CONFIG_MPI
      /* skip remote vector components */
      /* if (by/pt_n!=mpi_i) continue; --> div is slow (+70%) on IA64 */
      if (by<mpi_i*pt_n || by>=(mpi_i+1)*pt_n) continue;
      by%=pt_n;  /* costs much on IA64, ToDo: replace/remove */
#endif
      mpi_thb[block].vy[ii] = hbuf->v0->v[by][xy & ~XY_FLAG];
    } /* for (i=0..hnz,ii=0..HBLen) */
    for (;ii<HBLen;ii++) mpi_thb[block].vy[ii] = 0; // not needed!???
#ifdef CONFIG_MPI
    { /* nondiagonal hnz elements, do MPI */
      /* see nhm() for multithread problem with MPI_Sendrecv */
      int len1, len2, id, i1, node2, node1; ULong yy;
      MPI_Status status; /* ToDo: static? need per thread */
      for (id=1;id<mpi_n;id++) { /* sendrecv to/from all other nodes */
        node1=(mpi_i      +id)%mpi_n;  /* dest node+id */
        node2=(mpi_i+mpi_n-id)%mpi_n;  /* src  node-id */
        /* ToDo: why node1 != node2? only good for half duplex? */
        len1=0;
        /* copy only data for remote node */
        if (ihblk<nhblks) /* not for dummy blocks */
        for (i1=0l; i+i1<in && i1<HBLen; i1++) {
           xy = (ULong)tbh->xy[ i1 ]; /* incl. XY_FLAG */
           by =        tbh->by[ i1 ]; /* block<mpi_n*pt_n */
           /* skip local and non node1 remote data */
           /* if (by/pt_n!=node1) continue; --> div is slow (+70%) on IA64 */
           if (by<node1*pt_n || by>=(node1+1)*pt_n) continue;
           mpi_thb[block].iy[len1] = i1;
           mpi_th1[block][len1].yy = ((xy & ~XY_FLAG)<<pt_n_bits)+by%pt_n;
           len1++;
        } /* for (i1=0..HBLen) */
        len2=0;
        pthread_mutex_lock(&token_ring[hbuf->block]);
// if((vvv&4) && itr<4) printf("\nMPI_1 %3d %3d S= %3d R= %3d lenS=%6d i=%8lu ihblk= %6lu %6lu",itr,mpi_i*pt_n+hbuf->block,node1,node2,len1,xx,ihblk,nhblks);
        MPI_Sendrecv(mpi_th1[block     ],len1 *sizeof(struct mpi_Hbuff1),MPI_BYTE,node1,0,
                     mpi_th1[block+pt_n],HBLen*sizeof(struct mpi_Hbuff1),MPI_BYTE,node2,0,MPI_COMM_WORLD,&status);
        MPI_Get_count(&status,MPI_BYTE,&len2); /* works for Sendrecv too? */ 
        pthread_mutex_unlock(&token_ring[hbuf->block+1]);
        len2/=sizeof(struct mpi_Hbuff1);
        for (i1=0l; i1<len2; i1++) {
          by = mpi_th1[block+pt_n][i1].yy & ((1<<pt_n_bits)-1); /* local blk */
          yy = mpi_th1[block+pt_n][i1].yy>>pt_n_bits;    /* local index num */
//          if (by<0 || by>=pt_n || yy>=b_len[mpi_i*pt_n+by%pt_n]) printf ("\nERR %d %lu",by,yy);
          mpi_thb[block+pt_n].vr[i1] = hbuf->v0->v[by][yy];
        } /* for (i1=0..HBLen) */
        pthread_mutex_lock(&token_ring[hbuf->block+pt_n]);
// if((vvv&4) && itr<4) printf("\nMPI_2 %3d %3d S= %3d R= %3d lenS=%6d",itr,mpi_i*pt_n+hbuf->block,node2,node1,len2);
        /* ToDo: we only have to send .vy */
        MPI_Sendrecv(mpi_thb[block+pt_n].vr,len2*sizeof(mcplx),MPI_BYTE,node2,1,
                     mpi_thb[block     ].vr,len1*sizeof(mcplx),MPI_BYTE,node1,1,MPI_COMM_WORLD,&status);
        pthread_mutex_unlock(&token_ring[(hbuf->block+pt_n+1)%(2*pt_n)]);
//        pthread_mutex_unlock(&token_ring[(hbuf->block+1)%(pt_n)]);
        /* sort back results from remote note */
        for (i1=0l; i1<len1; i1++)
           mpi_thb[block].vy[ mpi_thb[block].iy[i1] ] =
                              mpi_thb[block].vr[i1];
      } /* id=0..mpi_n-1 */
    } /* MPI */
    // if ((vvv&4) && itr<4) printf("\nMPI_3 %3d %2d %lu %lu", itr, mpi_i*pt_n+block, i, in); OK
#endif  /* MPI */
    for (ii=0l; xx<min_b_sn1 && i<in && ii<HBLen; i++,ii++) {
      /* ToDo: unroll loop by get quad entries */
      /* this is the innermost loop, NFLOP=5*hnz*NumIterations */
      xy=    (ULong) tbh->xy[ ii ]; /* idx 0..n1/(mpi_n*pt_n) [+XY_FLAG] */
#if HRMAX>0
      r = hr[block][ tbh->rr[ ii ] ];
#else
      r =     M_f2d( tbh->rr[ ii ] );
#endif
      /* end of h_get */
      sum_ry += r * mpi_thb[block].vy[ii];  /* <ii|H|v0> += <ii|H|yy> */
// if ((vvv&4) && itr<4 && sum_ry) printf("\nMPI_3 %3d %2d xx=%6lu yy=%6lu r=%.2f v=%.2f sum=%.2f",itr,mpi_i*pt_n+block,b_ofs[mpi_i*pt_n+block]+xx,b_ofs[mpi_i*pt_n+block]+xy&~XY_FLAG,r,mpi_thb[block].vy[ii],sum_ry);
      if (!(xy & XY_FLAG)) continue; /* EOL (xx) of sparse matrix H? */
      hbuf->rx += D(Conj(M_f2d(*v0x))*sum_ry); /* <vHv> should be real */
      (*v1x)   +=        M_d2f(       sum_ry); /* r/w memory, cache miss!? */
// if ((vvv&4) && itr<4 && sum_ry) printf("\nMPI_4 %3d %2d xx=%6lu v1x+=%.2f ", itr, mpi_i*pt_n+block, b_ofs[mpi_i*pt_n+block]+xx, sum_ry);
      xx++; v0x++; v1x++; sum_ry=0; /* next line */
    } /* for (i=0..hnz,ii=0..HBLen) */
  }  /* 0<=ihblk<hblks_max */

  /* now handle the nonstored elements of H-matrix
   * ToDo: blocking:
   *   return block+num_lines stored into, or dynamic growing buffer
   *   or call nhv with b_ofs + b_len (full block)
   *  + better call _nhv once for 1..ahead (block wise)!
   */
  if (vvv&256 && mpi_i<4 && mpi_i>=mpi_n-4) { /* v2.33 */
    printf("\nv&256 geth_block blk= %d.%d sn1= %lu b_len= %lu"
           " min_b_sn1= %lu extra_nhv: %lu",
      mpi_i, block, b_sn1[block], b_len[mpi_i*pt_n+block], min_b_sn1,
      ((max_b_len-min_b_sn1)+hbuf->num_ahead_lines-1)
                            /hbuf->num_ahead_lines);
  }
  /* old=unbalanced: 0 <= i < b_len[blk],
   * new=balanced:   min_b_sn1 <= i < max_b_len */
  for (i=min_b_sn1; i<max_b_len; i += hbuf->num_ahead_lines) {
    /* if (v0) return hbuf[i].rr = <i|H|v0> (true here), ignore i>=b_len */
    hamilton_nhv(i, hbuf->num_ahead_lines, hbuf, 0);
    for (ii=0; ii<hbuf->num_ahead_lines
          && i+ii<b_len[mpi_i*pt_n+block]; ii++) {
      r= hbuf->el[ii].rr;
      hbuf->v1->v[block][i+ii] += M_d2f(r); /* v1+= <nHv> */
      hbuf->rx += D(Conj(M_f2d(hbuf->v0->v[block][i+ii]))*r); /* real <vHv> */
    }
  }
}

/*  v1=H*v0+v1 (parallel),  return r1=<v0|H*v0>=real
 *   (a-ib)*(c+id) = (ac+bd) + i(ad-bc) 
 * <r-i|Hr+Hi|r+i> = <r|Hr r>+<r|Hi i>+<i Hr i>+<i Hi r>
 * using long double ???
 * ToDo: hamilton_geth_block + hamilton_nhv_block => hamilton_block
 * called by diag.c:lanz1()
 *           diag.c:inc1()   v0+v1 set
 *           diag.c:incv1()
 *
 *   calls hamilton_geth_block(storedH) incl. hamilton_nhv_block(nonstoredH)
 * ToDo: Move pthread-code into MPI-loop
 */
double hamilton(tvec *v0, tvec *v1){  /* v1=H*v0+v1, return <v0|H|v0> */
  // ULong i; 
  int block;
  thbuf * (hbuf)[B_NUM]; /* one buffer per compute thread, better alloc locally */
  double r2=0.0;  
  for (block=0; block<pt_n; block++) {
    h_ropen(block);             /* open I/O streams + buffer v2.26 */
    hbuf[block] = alloc_hbuf(); /* better alloc local in thread? but free? */
    if (!hbuf[block]) {
      printf("\nERROR hamilton: alloc_hbuf failed, mpi=%d thread=%d",
             mpi_i, block);
      continue;
    }
    hbuf[block]->block = block;
    hbuf[block]->v0 = v0;
    hbuf[block]->v1 = v1;
    hbuf[block]->rx = 0.0;
    parallel_start( block, hamilton_geth_block, hbuf[block] );
  }
  for (block=0; block<pt_n; block++) {
    if (hbuf[block]) parallel_wait(block); /* only for threads */
    h_rclose(block); /* v2.26 close buffer chain and finaly the file */
    if (!hbuf[block]) continue;
    r2 += hbuf[block]->rx;    /* <vHv> */
    free_hbuf(hbuf[block]);
  }
#ifdef CONFIG_MPI
  { double add=r2, sum=0;
    MPI_Allreduce(&add, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    r2=sum; }
#endif
  return r2;
}

/* void-function for pthreads
 * ToDo: return compute time to calculate load balance
 *  input: blk=0..pt_n-1 (its cheaper to add mpi_i*pt_n as dividing by pt/n)
 * must be called for all blocks per mpi-process avoiding deadlocks
 */
void storeh_block( void *ptr ){
  ULong i, iout, l1, l2;
  thbuf *hbuf=(thbuf *)ptr;
  int block=hbuf->block,            /* node block    0..pt_n-1 */
      c_blk=mpi_i*pt_n+hbuf->block; /* cluster block 0..pt_n*mpi_n-1 */
  l1=itime();
  iout=1<<12;  /* output alive msg sparingly, N=40: 2^29 8outs */
  /* one MPI transaction per loop, max(b_len[c_blk]) for balance */
  for(i=0; i<max_b_len; i+=hbuf->num_ahead_lines){
    if ((!c_blk) && iout && i>=iout) { /* give alive signals */
      l2=itime()-l1; iout<<=2; /* factor 4, iout may overflow to 0 */
      if (l2>9) printf("\n SH %6.2f%% hnz/i= %6.2f t100%%= %6.2fm",
                  i*100./b_len[0], b_hnz[0]*1./i, l2/60.*b_len[0]*1./i);
    }
    /* ToDo: return number of lines filled into the buffer? */
    /* ToDo: send/recv Hij elements to/from other nodes (cached) */
    /* storeH to disk */
    hamilton_nhv( i, hbuf->num_ahead_lines, hbuf, 2);
  }
  if ((vvv&4) && (c_blk<4 || c_blk>=mpi_n*pt_n-4)) 
    printf("\nvvv&4 SH_block.end %3d sn1= %9lu snz= %9lu hnz= %9lu",
             c_blk, b_sn1[block], b_snz[block], b_hnz[block]);
  if (err_fatal)
    printf("\nerr_fatal %d: %d.%d i=%8lu h_nz=%9lu (SH_block.end)",\
       err_fatal,mpi_i,block,i,b_hnz[block]);
  hbuf->rtime=itime()-l1;
}

/* store H elements on disk, reading disk is faster than computing!? Aug07 */
void storeh(){
  int block;
  long i1, i2;                  /* runtime in seconds */
  thbuf * (hbuf)[B_NUM]; double h_nn, h_nz;
  long       rtime_max,       rtime_min; /* balance + timing measurement */
  int  block_rtime_min, block_rtime_max;
  rtime_max=rtime_min=block_rtime_min=block_rtime_max=0; /* avoid gcc warning */
  MPI_Barrier(MPI_COMM_WORLD); /* debug */
  if (!mpi_i) {
    if (vvv) printf("\n storeh");
    /* (buffersize-overhead)/numelements=size_per_element */
    if (vvv) printf(" NZXMAX= %d  AHEAD_LINES= %d hnzSize=HSize/HBLen: %d/%d=%d",
             NZXMAX, NUM_AHEAD_LINES, (int)Hsize, HBLen, (int)Hsize/HBLen);
    if (!(vvv&2)) {  /* ToDo: measure factor! */
      printf("\n try option vvv&2 for optional checks (1..2x slower)");
    } else {
      printf("\n try option vvv&64 to output matrix elements");
      printf("\n vvv&2: optional checks on H, but may be more slow (1..2x)");
    }
  }
  statH_ini();	/* initialize H-statistics Oct99 */

  ini_thxy(n1); /* see h_file.c, initialize H-storage, print size needed */

  if (!mpi_i) printf(" sizeof(HR_TYPE)=%d\n",(int)sizeof(HR_TYPE));
  i1=itime();i2=iclock();
  MPI_Barrier(MPI_COMM_WORLD); /* debug */
  if (checkpoint<3) {
    for(block=0;block<pt_n;block++){
      h_wopen(block); /* open I/O streams + ini buffer */
      /* big local vars (sbase), relatively bad for parallelisation */
      hbuf[block]=alloc_hbuf(); /* ToDo: alloc data area within the thread */
      hbuf[block]->block=block; /* overall block number */
      /* start compute threads, hbuf->v0=v1=NULL */
      parallel_start( block, storeh_block, hbuf[block] );
    }
    for(block=0;block<pt_n;block++){
      parallel_wait(block); /* stop compute threads and let them wait */
      h_wclose(block);      /* flush H-buffers to disk */
      if (block==0) {
        rtime_max=rtime_min=hbuf[0]->rtime; /* balance + timing measurement */
        block_rtime_min=block_rtime_max=0;
      } else {
        if (rtime_min>hbuf[block]->rtime) {
            rtime_min=hbuf[block]->rtime; block_rtime_min=block+mpi_i*B_NUM;}
        if (rtime_max<hbuf[block]->rtime) {
            rtime_max=hbuf[block]->rtime; block_rtime_max=block+mpi_i*B_NUM;}
      }
      free_hbuf(hbuf[block]);
    }
    i2=iclock()-i2; i1=itime()-i1;
    MPI_Barrier(MPI_COMM_WORLD); /* block to get real end time */
    /* ToDo: output time in a form where it can be removed for comparisions
     * example: diff <(sed 's/\([0-9]{2}\)/#1/g' out1) <(sed '' out2)
     * ToDo: output trace?
     */
#if STOREH != 0  /* not for memory only, v2.32 */
    h_checkpoint(1); /* write hr_saveNNN.dat */
#endif
    ++checkpoint;
    if (!mpi_i) printf(" checkpoint=%d", checkpoint);
    if (vvv && !mpi_i) {
       /* parallel computing balance: (compare to b_hnz = length of htmp*) */
       printf("\n SH.tmin= %5.2f blk= %d", rtime_min/60., block_rtime_min);
       printf("\n SH.tmax= %5.2f blk= %d", rtime_max/60., block_rtime_max);
       printf("\n SH.t=   %s clocks=%lu ", sclock(rtime_max), i2);
    }
  } else h_restore(1); /* read hr_saveNNN.dat */
  /* storeh complete, print statistics */
  if( h_error ){  /* error during write is fatal, ignore all things written */
    E0("disk full? written data will not be used"); h_error=0;
    for(h_nn=0,block=0;block<B_NUM;block++){ b_snz[block]=0; }
    for(h_nn=0,block=0;block<B_NUM;block++){ b_sn1[block]=0; }
  }
  /* max/min/sum hnz_blocks per thread (balance?) */
  bsnz_max=bhnz_sum=bhnz_max=0;
  bhnz_min=((b_hnz[0]+HBLen-1)/HBLen)*mpi_n*pt_n;
  min_b_sn1=max_b_len;  /* pseudo minimum stored n1 of all threads */
  /* ToDo: measure balance, hnz_min*pt_n/h_nz%..hnz_max... */
  for (block=0;block<pt_n;block++) { /* ToDo: nondiag_blocks per thread */
    ULong nblks=(b_hnz[block]+HBLen-1)/HBLen;
    if  (nblks!=(b_hnz[block]+HBLen-1)/HBLen) F0("ULong overflow, enlarge HBLen");
    /* encode info about blk-src */
    if (bhnz_max<nblks*mpi_n*pt_n) bhnz_max=nblks*mpi_n*pt_n+mpi_i*pt_n+block%pt_n;
    if (bhnz_min>nblks*mpi_n*pt_n) bhnz_min=nblks*mpi_n*pt_n+mpi_i*pt_n+block%pt_n;
    nblks=(b_snz[block]+HBLen-1)/HBLen;
    if (bsnz_max<nblks) bsnz_max=nblks;
    bhnz_sum+=nblks;
    if (b_sn1[block]<b_len[mpi_i*pt_n+block]) 
      if (min_b_sn1>b_sn1[block])
          min_b_sn1=b_sn1[block];
  }
  for (h_nz=0,block=0;block<pt_n;block++) h_nz+=(double)b_hnz[block];
  for (h_nn=0,block=0;block<pt_n;block++) h_nn+=(double)b_snz[block];
#ifdef CONFIG_MPI
  { double r1[2],r2[2];
    ULong  t1[4],t2[4]; /* ULLong may be a problem for MPI */
    /* stored + overall hnz + non stored n1 */
    t1[0]=bsnz_max; t1[1]=bhnz_max; t1[2]=mpi_maxlen;
    MPI_Allreduce(t1, t2,      3, MPI_UNSIGNED_LONG, MPI_MAX, MPI_COMM_WORLD);
    bsnz_max=t2[0]; bhnz_max=t2[1]; mpi_maxlen=t1[2];
    t1[0]=min_b_sn1;
    MPI_Allreduce(t1, t2,      1, MPI_UNSIGNED_LONG, MPI_MIN, MPI_COMM_WORLD);
    min_b_sn1=t2[0];
    t1[0]=bhnz_min;
    MPI_Allreduce(t1,&bhnz_min,1, MPI_UNSIGNED_LONG, MPI_MIN, MPI_COMM_WORLD);
    t1[0]=bhnz_sum;
    MPI_Allreduce(t1,&bhnz_sum,1, MPI_UNSIGNED_LONG, MPI_SUM, MPI_COMM_WORLD);
    r1[0]=h_nz; r1[1]=h_nn;
    MPI_Allreduce(r1, r2,      2, MPI_DOUBLE,        MPI_SUM, MPI_COMM_WORLD);
    h_nz=r2[0]; h_nn=r2[1];
    t1[0]=mpi_traffic;
    MPI_Allreduce(t1,t2,       1, MPI_UNSIGNED_LONG, MPI_SUM, MPI_COMM_WORLD);
    mpi_traffic=t2[0];
  }
  if (vvv && !mpi_i)  /* t=bytes/BW+pkts*latency, maxlen in data objects */
    printf("\n MPI traffic  bytes= %.2e  pkts= %.2e  maxlen= %lu",
      mpi_traffic*1., mpi_syncs*1., mpi_maxlen);
  mpi_traffic=0;
  mpi_syncs=0;
  mpi_maxlen=0;
#endif
  if (vvv && !mpi_i) {
    printf("\n n1 = hilbert dimension,  hnz = number of nonzero matrix elements");
    /* ToDo: better output counted hnz: n1 hnz/n1 ... */
    printf("\n n1= %.2e hnz/n1= %.3f  hnz_stored= %3.0f%%  min_b_sn1= %.0f",
            1.*n1, h_nz/n1, ((h_nz)?(100*h_nn)/h_nz:0.0), 1.*min_b_sn1);
    printf("\n hnz_blks/threads= %9.2f"
           " balance= %.0f..%.0f %.1f%% min,maxblk=%lu,%lu",
             bhnz_sum*1./(pt_n*mpi_n),
             /* min,max has its blocknum encoded, decode it */
             (bhnz_min/(mpi_n*pt_n))*1.,
             (bhnz_max/(mpi_n*pt_n))*1.,
         100-(bhnz_max/(mpi_n*pt_n)
             -bhnz_min/(mpi_n*pt_n))*100./(bhnz_max/(pt_n*mpi_n)),
             (bhnz_min%(mpi_n*pt_n)),
             (bhnz_max%(mpi_n*pt_n)) );
  }
  if (mpi_i==0 && mpi_n>1 && 2*(h_nz/node_n1)<NZXMAX)
    printf("\nWarning 2hnz/n1<NZXMAX: waste of memory, see v&2");
  if (mpi_i==0 && mpi_n>1 && MaxSBase>1)
    printf("\nWarning MaxSBase>1: waste of MPI bandwidth (factor %.1f),"
           "\n please reduce MaxSBase in spins.h to 1",
           sizeof(thbuf)*1./(sizeof(thbuf)
              -(MaxSBase-1)*(sizeof(tbase)+sizeof(tphase))));
  if(!mpi_i)
  { double h_out_speed=h_nz*0.000001*(Hsize/HBLen);
    h_out_speed/=((i1)?i1:1); /* time in seconds, no div 0 */
    printf("\n H_out_speed= %8.2e MB/s  err= %.0f%%",
           h_out_speed, 100*(1.0-i1/(i1+1.0)));
  }
  if ((!mpi_i) && h_nn<h_nz) {
    printf(
     "\nWarning: H stored incompletely, need maxfile=%.2e for best speed.",
            1.*Hsize*h_nz/HBLen/(mpi_n*pt_n)+Hsize);
  }
  if (vvv&2) statH_out();
  MPI_Barrier(MPI_COMM_WORLD); /* debug */
  return;
}

/* --------------------- Jan13 2000 -------------------------------
 * new way for storing l1 and H, is it faster than the old way?
 * it should generate the symmetric states itself by
 * - starting with symmetric config
 * - using iterativ H, and symconfig and sort to generate new sym-configs
 * - until H generates no new configs (m1=0 run1 for n1, m1=1 run2 for sh)
 * algo&8:
 * ToDo: make MPI ready! sort in using ising-ergs + hashing!
 *       or mean-(diag|ising)erg of 0th and 1st excitations
 *       to have best blocking and small number hnz in of diagonal blocks
 * use kind of MC? 
 *    1st step generate mpi_n*pt*100 states and part it to mpi_n*pt blocks
 *    generate all states between start_block_i and start_block_i+1
 */
void storeh2(ULong m1){
  ULong t1,t2,i1,i2,i3,i4,i5,i6,i0; int j1,j2,block;
  struct sbase b1,b2; tbase ib; double h_nz, h_nn;
  thbuf *hbuf;
  hbuf=alloc_hbuf();  /* very big */
  hbuf->num_ahead_lines=1; /* bad, but code is not ready */
  if (vvv) printf("\n storeh2 (warning, buggy code, needs review)");
  b1.o=0; statH_ini();	/* Oct99 */
  /* das speichern des letzten hnz der Zeile macht probleme! da sich n1 aendert! */
  ini_thxy(n1); /* compute bytes needed for storing H elements, ini files */
  t1=itime();t2=iclock();
  i1=0;		/* pointer to last unchecked state */
  i2=1;		/* pointer to next free entry */ 
  if(m1==0){		/* starting state == Neelstate (lowest ising erg) */
    ib=TBASE0;i0=0l;
    for(j1=0;j1<2*nd;j1+=2) ib=puti(ib,j1%nn+j1/nn,2); /* ib=Neel */
    /* MC steps to lower Ising energy if possible, only JJ-model, ToDo */
    /* ToDo: more simple: erg=nHn + nhv to get next n for next better min? */
    {
      int *k1, *k2;   /* pair correlations == ising erg */
      tbase ib2, ib3=ib;      /* test state */
      int z1, z2, j3;
      for (j3=2;j3>0;j3--) { /* down counter */
        k1=get_dcorr(ib ,3, NULL);
        for (j1=0   ;j1<nn-1;j1++)
        for (j2=j1+1;j2<nn  ;j2++)
          if(j1!=j2 && (z1=geti(ib,j1))!=(z2=geti(ib,j2))) {
          /* now j1 and j2 are different neighbours */
          ib2=puti(puti(ib,j2,z1),j1,z2); /* j1 <=> j2 */
          k2=get_dcorr(ib2, 1, NULL); /* get NumIsingExcitations, ToDo: ising, not dcorr */
          /* prefer more movable pairs, ToDo: check all corrs */
          if (k1[1*CM(2,4,5)]< k2[1*CM(2,4,5)] ||  /* smallest j1 */
             (k1[1*CM(2,4,5)]==k2[1*CM(2,4,5)] &&
              k1[2*CM(2,4,5)]< k2[2*CM(2,4,5)])) { 
                 j3=2; ib=ib2; free(k1); k1=k2; j1=j2=nn;
                 b_smallest(ib,&b1);
                 if (b1.sign!=0) ib3=ib; /* store sign!=0 */
              }
          else free(k2);
        }
        free(k1);
      }
      ib=ib3;
    }
    printf("\n start konfig= %s",ccfg(ib));
    b_smallest(ib,&b1); ib=b1.scfg;
    printf("\n         scfg= %s  sign=%+g",ccfg(b1.scfg), D(b1.sign));
    if( b1.sign==0
#ifdef CONFIG_ANSATZ
     || 0==ansatz_test(hbuf->el[j1].bj)		/* variational ansatz */
#endif
    ) {
    printf("\n bad start konfig (sign==0 || proj==0), try another");
    for(i0=0;i0<10000;i0++){ ib=ll1(i0); b_smallest(ib,&b1); ib=b1.scfg;
      if( b1.sign!=0
#ifdef CONFIG_ANSATZ
      && 0!=ansatz_test(hbuf->el[j1].bj)		/* variational ansatz */
#endif
      )break; }
    }
    putl1(0,ib,l1);	/* first state = starting state, n1=1 ??? */
    if (l2b) l2b[0]=0;	/* sorting-idx */
    printf("\n start konfig= %s",ccfg(ib));
  } else i2=m1;   /* start with picked state */
  h_wopen(0);
  if (mpi_i==0 && mpi_n>1) E0("missing MPI code in storeh2()");
  MPI_Barrier(MPI_COMM_WORLD);
  for (;;) {
    hbuf->n=0;
    /* ToDo: i1 must match 0..(b_len[blk]-1) v2.27 */

    /* ToDo: set num_ahead_lines=1 ? */
    hamilton_nhv( i1, 1, hbuf, 4);    /* store nHm only in buf_ */
    if (err_fatal) { E1(" err_fatal: i1=%lu",i1); break;}
    if (i1<100 && vvv&64) { printf("\n H|i1> = H|%s> i1=%5lu i2=%5lu",
                       ccfg(getl1(i1)),i1,i2);
                       get_dcorr(getl1(i1),2,NULL);
    }
    b_smallest(getl1(i1),&b2);	/* need for b2.norm2 */
    if(phase_eq(b2.sign,0)) E0(" should not happen!");

    /* set scfg (compress extracted in v2.29, getbase replaced) */
    for (j1=0;j1<hbuf->n;j1++) if (hbuf->el[j1].rr!=0) /* test all H|cfg> */
    {					/* every cfg with <cfg|H|i1>!=0 */
      b_smallest(hbuf->el[j1].bj,&b1);		/* dont forget is==0!!! ??? */
      /* representator, better use w2w ??? */
      hbuf->el[j1].rr*=b1.sign;
      hbuf->el[j1].bj =b1.scfg;
      if(b1.norm2!=0 && b2.norm2!=b1.norm2)
         hbuf->el[j1].rr*=sqrt((double)b2.norm2/(double)b1.norm2);
    }		/* debugging friendly src-code ??? */

    /* compress */
    for (j1=0;j1<hbuf->n;j1++) if (hbuf->el[j1].rr!=0) /* test all H|cfg> */
    {					/* every cfg with <cfg|H|i1>!=0 */
      /* search hbuf->el[j1].bj in rest of hbuf, compression */
      for(j2=j1+1;j2<hbuf->n;j2++)
        if (hbuf->el[j2].rr==0
         || hbuf->el[j2].bj==hbuf->el[j1].bj) {
        hbuf->el[j1].rr += hbuf->el[j2].rr;
        hbuf->n--;
        hbuf->el[j2].rr = hbuf->el[hbuf->n].rr;
        hbuf->el[j2].bj = hbuf->el[hbuf->n].bj; /* compress */
        j2--;
      }
    }		/* debugging friendly src-code ??? */

    /* remove zeros */
    for (j1=0;j1<hbuf->n;j1++) if (hbuf->el[j1].rr==0) /* test all H|cfg> */
    {					/* every cfg with <cfg|H|i1>!=0 */
        hbuf->n--;
        hbuf->el[j1].rr = hbuf->el[hbuf->n].rr;
        hbuf->el[j1].bj = hbuf->el[hbuf->n].bj; /* compress */
        j1--;
    }		/* debugging friendly src-code ??? */

/* parallel ???????, braucht immense Zeit */
    i4=i2;
    for(j1=0;j1<hbuf->n;j1++)if(hbuf->el[j1].rr!=0){	/* allready stored? */
      if(hbuf->el[j1].bj==getl1(i1)){ hbuf->el[j1].jj=i1;continue; }
      i3=i2;
      if(l2b)
      {	/* suche im sortierten hilfsindex - schneller?! */
        for(i6=1;i6<=i4;i6<<=1);	/* i4 - hintere bits */
        for(i5=0;i6;i6>>=1){		/* interval-Halbierung */
          if (i5+i6>=i4) continue;      /* i+i/2 > max */
          if (getl1(l2b[i5+i6])<=hbuf->el[j1].bj) i5+=i6;
        }
        if(getl1(l2b[i5])==hbuf->el[j1].bj){
              hbuf->el[j1].jj=l2b[i5];continue; } /* found, min_j? */
        if(i4>=n1){ hbuf->el[j1].rr=0;continue; } /* 0-Projection, count? */
        /* insert new element, max_j? -> max(maxj-minj) v-size? */
        for(i6=i4;i6>i5;i6--)l2b[i6]=l2b[i6-1];
        if(i5==0 && hbuf->el[j1].bj<=getl1(l2b[i5])) l2b[0]=i4; else
        l2b[i5+1]=i4;
      } else {	/* sequential searching is very slow! */
        for(i3=0;i3<i2;i3++)if(hbuf->el[j1].bj==getl1(i3))break;
        if(i3<i2){ hbuf->el[j1].jj=i3;continue; }	/* already stored */
      }
      if(i4>=n1){ hbuf->el[j1].rr=0;continue; }
      putl1(i4,hbuf->el[j1].bj,l1); hbuf->el[j1].jj=i4;
      i4++;
      if(i4==n1) E1("n1-ovl, n1=%.0f", 1.*n1); /* report once */
    }
    if(i4==1 && hbuf->n==0){ ERR(500);break; }
    if(i4>=n1){ E0("n1-ovl");break; }
    hbuf->i=i1;
    hamilton_puth_line(hbuf);	/* write Hamiltonian!!! run2 n1=correct */
    i2=i4;i1++;if(i1>=i2)break;

  }
  h_wclose(0); 
  t2=iclock()-t2;t1=itime()-t1; h_checkpoint(1); n1=i1;
  if (vvv) printf(" SH2 nk=%lu SH.t=%s=%lu n1=%lu\n",i2, sclock((long)t1),t2,i1);
  if (vvv&2) statH_out();
  /* hnz = number of stored elements, snz = number of stored lines */
  for(h_nz=0,block=0;block<pt_n;block++) h_nz+=(double)b_hnz[block];
  for(h_nn=0,block=0;block<pt_n;block++) h_nn+=(double)b_snz[block];
  for(block=0;block<pt_n;block++) if (b_len[block]>i1) b_len[block]=i1;
  if (vvv && !mpi_i)
    printf("\n h_nz = num of matrix elements, h_nn = stored elements");
  printf("\n h_nz %.3e h_nn %.3e %3.0f%% hxy %d",
             h_nz,     h_nn,(h_nn*100.)/h_nz,(int)Hsize/HBLen);
  if(h_nn<h_nz)
  printf("\nWarning: %.2e B for maximum speed needed.", 1.*Hsize*h_nz/HBLen);
  free_hbuf(hbuf);
  return;
}

int outbasis(FILE *f){
  ULong i;
  static struct sbase b1; int i1,i2,i3,i4,i6; b1.o=0;
  if(vvv&64 && !mpi_i){ fprintf(f,"\n vvv&64: B=[");
    for(i=0l;i<node_n1;i++){    if(i!= 0l)fprintf(f,",");
      if (i<x_max) fprintf(f,"\n%4lu %s ",(ULong)i,ccfg(getl1(i)));
             else {fprintf(f," ... ");break;}
      b_smallest(getl1(i),&b1);fprintf(f,"*q(%4g)=(",(double)b1.norm2);
      /* this is ising like energy for all diff. correlations */
      /* ToDo: replace by get_dcorr */
      for(i4=0,i1=1;i1<sym.num_dcorr;i1++){ i6=0;
         for(i2=   0;i2<nn;i2++)
         for(i3=i2+1;i3<nn;i3++)
         if(sym.corr[i2][i3]==i1) {
          if(geti(b1.scfg,i2)==1 && geti(b1.scfg,i3)==1); else
          if(geti(b1.scfg,i2)==2 && geti(b1.scfg,i3)==2); else
          i6++; /* ud - symmetric */
         }
         i4+=i6;
         printf(" %2d",i6); /* printf("+%2d+%2d",i4,i6); */
      }
      fprintf(f," sum= %d)  ",i4);
    } fprintf(f,"]\n");
  }
  return(0);
}

int zi(tbase cfg,int i){   /* Siz -> -1,0,1 */
#if CM(1,0,0)
  return 1-2*(int)(1&(cfg>>i));}
#else
  int i1=geti(cfg,i); return( (i1==2) ? -1 : (i1==1) ? 1 : 0);}
#endif

int zizj(tbase cfg,int i,int j){   /* Siz*Sjz -> -1,0,1 */
#if CM(1,0,0)
  return 1-2*(int)(1&((cfg>>i)^(cfg>>j)));}
#else
  int i1=geti(cfg,i)*geti(cfg,j); return((i1==2)?-1:(i1==4 || i1==1)?1:0);}
#endif

double sq( double x){ return x*x; }

/* calculate the geometrical distance^2 between site i1 and i2 */
double distance( int i1, int i2){
  double d1,d2;
  int i,j,k;
  d1=sq(pos[i1][0]-pos[i2][0])
    +sq(pos[i1][1]-pos[i2][1])
    +sq(pos[i1][2]-pos[i2][2]);
  for(i=-1; i<2; i++)
  for(j=-1; j<2; j++)
  for(k=-1; k<2; k++) { /* brute force, test surrounding boxes */
    d2=sq(pos[i1][0]-pos[i2][0]+i*edg[0][0]+j*edg[1][0]+k*edg[2][0])
      +sq(pos[i1][1]-pos[i2][1]+i*edg[0][1]+j*edg[1][1]+k*edg[2][1])
      +sq(pos[i1][2]-pos[i2][2]+i*edg[0][2]+j*edg[1][2]+k*edg[2][2]);
    if (d2<d1) d1=d2;
  }
  return d1;
}

/*
 ***********************************************************************
 * output a table(j1)/matrix(j1,j2) of Ising weigths (picture of state?)
 * using w2w-matrix (output via vvv&8)
 */
#define Imax (2*NN)    /* maximum number uu-pairs per bond type */
extern int w2w[NN][NN];

struct s_ising_weight {
  int blk; /* thread number */
  ULong  states[Imax*Imax];  /* number of states */
  double weight[Imax*Imax];  /* weight matrix */
  int   d1a,d1b, d2a,d2b;     /* matrix dimension (maxIsingErg=max_ud_per_bond) */
  tvec *v0;
} ising_ptr[B_NUM];

void ising_weight_thread( void *ptr ) {
  /* map an site pair nisnjs to a Ising index to count up, see get_dcorr */
  const char map[16]= /* fits to L1 cache line !? */
    {  2, 3, 3, 4 ,  3, 0, 1, 3 ,  3, 1, 0, 3 ,  4, 3, 3, 2 };
   /* .. .u .d .3   u. uu ud u3   d. du dd d3   3. 3u 3d 33 => 0..4 = 3bit */
   /* index: 0=uu+dd  1=ud+du  [2=..+33  3=.u+u.+.d+d.  [4=.3+3.]] */
  int   *buf=NULL;  /* buffer for counted pairs (IsingErg) */
  struct s_ising_weight *pp=(struct s_ising_weight *)ptr;
  int blk=pp->blk;
  ULong i1;       /* state counter */
  tbase b1;       /* cfg[i] */
  int j1, j2;     /* num excitations */
  int i2, i3, i4, ib1, ib2; /* work variables */
  buf=(int *)malloc(CM(2,4,5)*(numpara+2)*sizeof(int));
  if (!buf) { E0("ising_weight: malloc failed"); return; }
  /* upper border of IsingErgs=Neel is nn, add a overflow index nn */
  for (j1=0;j1<Imax*Imax;j1++) {
    pp->states[j1]=0;
    pp->weight[j1]=0;
  }
  pp->d1b=pp->d2b=0;
  pp->d1a=pp->d2a=Imax;

  for(i1=0l; i1<b_len[mpi_i*pt_n+blk]; i1++){  /* l1[] sort for norm2[]??? */
    b1=getl1(i1+b_ofs[mpi_i*pt_n+blk]);
    // l2.num=0; l2.o=0; b_smallest(ib1,&l2); /* get norm2 = slow */
    /* reset IsingErg table */
    for (i4=0;i4<numpara*CM(2,4,5);i4++) { buf[i4]=0; }
    /* get Ising energies per bond */
    for (i2=   0;i2<nn-1;i2++)    /* N*(N-1) site pairs */
    for (i3=i2+1;i3<nn  ;i3++) {
           i4=w2w[i2][i3];    /* access N*N pair array */
           if (i4<0 || i4>=numpara || i4>2+1) continue;
           ib1=geti(b1,i2);  /* 0=. 1=u=o 2=d=x 3=u+d=Q */
           ib2=geti(b1,i3);
           if (CM(2,4,5)*i4+map[4*ib1+ib2]>CM(2,4,5)*(numpara)) ERR(9999);
           buf[CM(2,4,5)*i4+map[4*ib1+ib2]]++;
    }
    j1=buf[CM(2,4, 5)]/2;
    if (numpara<3) j2=0; else
    j2=buf[CM(4,8,10)]/2; /* segfault if numpara<3 ? */
    if (j1>=Imax) j1=Imax-1; /* overflow */
    if (j2>=Imax) j2=Imax-1; /* overflow */
    if (j1>pp->d1b) pp->d1b=j1; /* max matrix idx, minimum output matrix */
    if (j2>pp->d2b) pp->d2b=j2;
    if (j1<pp->d1a) pp->d1a=j1; /* min matrix idx, minimum output matrix */
    if (j2<pp->d2a) pp->d2a=j2;
    pp->states[j2*Imax+j1]++;
    pp->weight[j2*Imax+j1]+=Norm(pp->v0->v[blk][i1]); // *l2.norm2;
  }
  free(buf);
}

int out_ising_weight(tvec *v0) {
  long  t1=itime();
  int j1, j2, blk;
  double sum;
  ULong i1;
  if (!mpi_i)
    printf("\nout_ising_weight: numpara= %d", numpara);
  MPI_Barrier(MPI_COMM_WORLD);
  if (mpi_n>0 && !mpi_i)
    printf("\nWarning: mpi_n>1 can produce wrong results (ToDo)");

  for(blk=0; blk<pt_n; blk++) {
    ising_ptr[blk].blk=blk;
    ising_ptr[blk].v0=v0;
    parallel_start(blk, ising_weight_thread, &ising_ptr[blk]);
  }
  for(blk=0; blk<pt_n; blk++) {
    parallel_wait(blk);
  }
  if (!mpi_i)
    printf("\nout_ising_weight: parallel work done.");
  MPI_Barrier(MPI_COMM_WORLD);
  /* reduce results from threads */
  for(blk=1; blk<pt_n; blk++) {
    for (j1=0;j1<Imax*Imax;j1++) {
      ising_ptr[0].states[j1] += ising_ptr[blk].states[j1];
      ising_ptr[0].weight[j1] += ising_ptr[blk].weight[j1];
    }
    if (ising_ptr[0].d1b<ising_ptr[blk].d1b)
        ising_ptr[0].d1b=ising_ptr[blk].d1b; /* max matrix idx, min output */
    if (ising_ptr[0].d2b<ising_ptr[blk].d2b)
        ising_ptr[0].d2b=ising_ptr[blk].d2b;
    if (ising_ptr[0].d1a>ising_ptr[blk].d1a)
        ising_ptr[0].d1a=ising_ptr[blk].d1a; /* min matrix idx, min output */
    if (ising_ptr[0].d2a>ising_ptr[blk].d2a)
        ising_ptr[0].d2a=ising_ptr[blk].d2a;
  }
  if (!mpi_i)
    printf("\nout_ising_weight: threads summed.");
  MPI_Barrier(MPI_COMM_WORLD);

#ifdef CONFIG_MPI
  {
    ULong  tstates[Imax*Imax];
    double tweight[Imax*Imax];  /* weight matrix */
    int   td1a,td1b, td2a,td2b;     /* matrix dimension (maxIsingErg=max_ud_per_bond) */
    MPI_Allreduce(&ising_ptr[0].states, &tstates, Imax*Imax, MPI_UNSIGNED_LONG, MPI_SUM, MPI_COMM_WORLD);
    MPI_Allreduce(&ising_ptr[0].weight, &tweight, Imax*Imax, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    for (j1=0;j1<Imax*Imax;j1++) {
      ising_ptr[0].states[j1]=tstates[j1];
      ising_ptr[0].weight[j1]=tweight[j1];
    }
    MPI_Allreduce(&ising_ptr[0].d1a, &td1a, 1, MPI_INT, MPI_MIN, MPI_COMM_WORLD);
    MPI_Allreduce(&ising_ptr[0].d2a, &td2a, 1, MPI_INT, MPI_MIN, MPI_COMM_WORLD);
    MPI_Allreduce(&ising_ptr[0].d1b, &td1b, 1, MPI_INT, MPI_MAX, MPI_COMM_WORLD);
    MPI_Allreduce(&ising_ptr[0].d2b, &td2b, 1, MPI_INT, MPI_MAX, MPI_COMM_WORLD);
    ising_ptr[0].d1a=td1a;
    ising_ptr[0].d2a=td2a;
    ising_ptr[0].d1b=td1b;
    ising_ptr[0].d2b=td2b;     
  }
  if (!mpi_i)
    printf("\nout_ising_weight: processes summed.");
  MPI_Barrier(MPI_COMM_WORLD);
#endif

  if (!mpi_i) { /* ToDo: mass center, mean deviation, momentums? */
    double center1=0, center2=0; /* mass center of Ising weight */
    double dev1=0, dev2=0,       /* deviation of Ising weight */
           cstate1=0, cstate2=0, /* center of states */
           r1, r2;
    for (j2=ising_ptr[0].d2a; j2<=ising_ptr[0].d2b; j2++)
    for (j1=ising_ptr[0].d1a; j1<=ising_ptr[0].d1b; j1++) {
      center1+=j1*ising_ptr[0].weight[j2*Imax+j1];
      center2+=j2*ising_ptr[0].weight[j2*Imax+j1];
      cstate1+=j1*ising_ptr[0].states[j2*Imax+j1];
      cstate2+=j2*ising_ptr[0].states[j2*Imax+j1];
    }
    for (j2=ising_ptr[0].d2a; j2<=ising_ptr[0].d2b; j2++)
    for (j1=ising_ptr[0].d1a; j1<=ising_ptr[0].d1b; j1++) {
      r1=(center1-j1)*ising_ptr[0].weight[j2*Imax+j1]; dev1+=r1*r1;
      r2=(center2-j2)*ising_ptr[0].weight[j2*Imax+j1]; dev2+=r2*r2;
    }
    printf("\n iw.t= %.2f min (Ising weight)"
           "\n entropy_center1,2= %8.4f %8.4f "
           "\n mass_center1,2=    %8.4f %8.4f  deviation1,2= %8.4f %8.4f"
           "\n weight matrix (%d-%d,%d-%d):",
            (itime()-t1)/60., cstate1/n1, cstate2/n1,
             center1, center2, sqrt(dev1), sqrt(dev2),
             ising_ptr[0].d1a, ising_ptr[0].d1b,
             ising_ptr[0].d2a, ising_ptr[0].d2b);

    for (j2=ising_ptr[0].d2a; j2<=ising_ptr[0].d2b; j2++)
    for (j1=ising_ptr[0].d1a; j1<=ising_ptr[0].d1b; j1++) {
      if (j1==ising_ptr[0].d1a) printf("\n ");
      if (ising_ptr[0].states[j2*Imax+j1])
            printf(" %11.9f", ising_ptr[0].weight[j2*Imax+j1]);
      else  printf(" %11g",   ising_ptr[0].weight[j2*Imax+j1]);
    }
    printf("\n#");
    for (j1=ising_ptr[0].d1a; j1<=ising_ptr[0].d1b; j1++) {
      sum=0;
      for (j2=ising_ptr[0].d2a; j2<=ising_ptr[0].d2b; j2++)
        sum+= ising_ptr[0].weight[j2*Imax+j1];
      printf(" %11.9f", sum);
    }

    printf("\n states matrix (%d-%d,%d-%d):",
      ising_ptr[0].d1a, ising_ptr[0].d1b,
      ising_ptr[0].d2a, ising_ptr[0].d2b);
    for (j2=ising_ptr[0].d2a; j2<=ising_ptr[0].d2b; j2++)
    for (j1=ising_ptr[0].d1a; j1<=ising_ptr[0].d1b; j1++) {
      if (j1==ising_ptr[0].d1a) printf("\n ");
      printf(" %11lu", ising_ptr[0].states[j2*Imax+j1]);
    }
    printf("\n#");
    for (j1=ising_ptr[0].d1a; j1<=ising_ptr[0].d1b; j1++) {
      for (i1=0,j2=ising_ptr[0].d2a; j2<=ising_ptr[0].d2b; j2++)
        i1+= ising_ptr[0].states[j2*Imax+j1];
      printf(" %11lu", i1);
    }
  }

  return 0;
}

/*  ************ check of the eigenvector and eigenvalue ************ */
/* Inp: v0  Out: v1  H*v0, rh=<v0*H*v0>, rq=norm(H*v0-h*v0) 
   Ausgaben werden mit XX= beschrieben, diese koennen mit awk bearbeitet
   werden  */
int chk3(tvec *v0,tvec *v1,int iv){
    int i1,i2,i3,j1,j2,j3,deg; long l1; double d1,d2,rr;
    static double sisj[NN][NN],zizj[NN][NN];
    deg=0; /* degenerate ? */

    for(i1=0;i1<nn;i1++)
    for(i2=0;i2<nn;i2++)zizj[i1][i2]=sisj[i1][i2]=0.0;

  if (mpi_n>0 && !mpi_i)
    printf("\nWarning: mpi_n>1 can produce wrong results (ToDo)");

#if !defined(Dimer)	/* better: only for generators <T>!!! Nov98 */

/* write a general diff_pair_corr(fkt{-NR},array[][],v0)
 *  and call with sisj,zizj or fkt=1,2,3... */
    l1=itime(); /* measure zizj-time */
#if Korrelations
    /* 3*<ZiZj>=<SiSj> nur, wenn <SS>=0 (singlet), auch bei Entartung (?) */
    /* could be problematicaly od degenerate states! not proofed */
    for(i3=0;i3<sym.num_dcorr;i3++){
      int sites[2]; long l2=itime();
      sites[0]=i1=sym.dcorr[i3][0];
      sites[1]=i2=sym.dcorr[i3][1];
      zizj[i1][i2]= wop(&op_zizj,NULL,2,sites,v0,v0); /* ZiZj */
      for(j3=j1=0;j1<nn;j1++)for(j2=0;j2<nn;j2++)
      if(sym.corr[j1][j2]==i3){
        zizj[j1][j2]=zizj[i1][i2]; /* copy same bonds */
        j3++;              /* count num of same bonds */
      }
      if (!mpi_i) {
        printf("\n i-j %2d %2d ZiZj %9.6f",i1,i2,zizj[i1][i2]); /* kurz+knapp */
        printf("  num= %4d  r^2= %7.3f", j3, distance(i1,i2));
        /* ToDo: no or only once? if fast, easier for diff-check */
        if (n1>>16) printf(" ij.t= %5.1f m", (itime()-l2)/60.); /* for diff-check */
      }
    }
    if (!mpi_i){
      printf("\n %dzij.t= %s", sym.num_dcorr, sclock(itime()-l1));

      for(i3=-1,i1=0;i1<nn;i1++)
      if(sym.corr[i1][i1]>i3){ i3=sym.corr[i1][i1];
        printf("\n i= %2d ZiZj=",i1);	/* 3ZiZj using sh/awk (simpler!) */
        for(i2=0;i2<nn;i2++) printf(" %9.6f",zizj[i1][i2]);
      }
      printf("\n Warning: SiSj=3ZiZj, ZiZj only correct for EA=1, sum over Sz too!");/* Jun99 */
    }

    for(i1=0;i1<nn;i1++)for(i2=0;i2<nn;i2++){
      sisj[i1][i2]=3*zizj[i1][i2]; }		/* 3*ZiZj isotropic */
    if (cfg_sisj&1)	/* set in daten.i */
    { 
      int sites[2]; long l2;
      l1=itime();
      if (vvv&1 && !mpi_i) printf(" t+=%s",sclock(~0l));
      if (!mpi_i) printf("\n cfg_sisj&1 is set, sisj will be calculated ");
      /* could be problematicaly on degenerate states! not proofed */
      for(i3=0;i3<sym.num_dcorr;i3++){
        l2=itime();
        sites[0]=i1=sym.dcorr[i3][0];
        sites[1]=i2=sym.dcorr[i3][1];
        sisj[i1][i2]=wop(&op_sisj,NULL,2,sites,v0,v0);  /* SiSj */
        if (!mpi_i)
          printf("\n i-j %2d %2d SiSj %9.6f",i1,i2,sisj[i1][i2]);
        for(j3=j1=0;j1<nn;j1++)for(j2=0;j2<nn;j2++)
        if(sym.corr[j1][j2]==i3){
          sisj[j1][j2]=sisj[i1][i2]; /* copy same bonds */
          j3++;              /* count num of same bonds */
        }
        d1=sisj[i1][i2]-3*zizj[i1][i2];
        if (!mpi_i) {
          printf("  Sij-3Zij= %9.6f %6.0e num= %4d", d1, d1, j3);
          if (n1>>16) printf(" ij.t= %5.1f m", (itime()-l2)/60.); /* for diff-check */
        }
      }
      if (!mpi_i) {
        printf("\n %dsij.t= %s", sym.num_dcorr, sclock(itime()-l1));
        for(i3=-1,i1=0;i1<nn;i1++)
        if(sym.corr[i1][i1]>i3){ i3=sym.corr[i1][i1];
          printf("\n i= %2d SiSj=",i1);				/* SiSj */
          for(i2=0;i2<nn;i2++) printf(" %9.6f",sisj[i1][i2]); }
      }
    }
    if ((!(cfg_sisj&1)) && !mpi_i)
      printf("\n cfg_sisj&1 is not set, sisj will not be calculated ");

    for(d1=i1=0;i1<nn;i1++)for(i2=0;i2<nn;i2++)
      d1+=sisj[i1][i2]; d2=fabs(d1); d2=fabs(-1.0+sqrt(1+4*d2));    /* 2S */
    if (!mpi_i) {
      printf("\n SS= %5.3f 2S= %5.3f % 5.0e", d1, d2, d2-(int)(d2+.5));
      if (cfg_sisj&1) printf(" (as sum sisj)");
      else            printf(" (as sum 3zizj)");
    }

    if (vvv&1 && !mpi_i) printf(" t+=%s",sclock(~0l));
    for(i3=-1,i1=0;i1<nn;i1++) if(sym.corr[i1][i1]>i3){	/* oct98 */
      if (!mpi_i) printf("\n i= %2d Zi=",i1);		/* Zi */
      i3=sym.corr[i1][i1];
      d1=wop(&op_zi,NULL,1,&i1,v0,v0); /* Zi */
      if (!mpi_i)
        printf(" %9.6f Zi-(nu-nd)/(2N)= %9.6f",d1,d1-(nu-nd)/(2.*nn));
    }
    if (vvv&1 && !mpi_i) printf(" t+=%s",sclock(~0l));
    for(d1=i1=0;i1<nn;i1++)for(i2=0;i2<nn;i2++)
    d1+=fabs(zizj[i1][i2]); d1=fabs(d1/(nn*nn)); /* ZMag=1/N^2 Sum |ZiZj| */
    if (!mpi_i) printf("\n ZMag= %10.8f",d1); d2=d1;

    for(d1=i1=0;i1<nn;i1++)for(i2=0;i2<nn;i2++)
    d1+=fabs(sisj[i1][i2]-zizj[i1][i2]); d1=fabs(d1/(nn*nn));
    if (!mpi_i) printf(  " XYMag= %10.8f",d1);		/* XYMag=1/N^2 Sum |XYiXYj| */

    for(d1=i1=0;i1<nn;i1++)for(i2=0;i2<nn;i2++) /* SMag=1/N^2 Sum |SiSj| */
    d1+=fabs(sisj[i1][i2]); d1=fabs(d1/(nn*nn));
    if (!mpi_i) printf(  " SMag= %10.8f SMag-3ZMag= % 10.8f",d1,d1-3*d2);

    /* this type of magnetization is for test if SMag(E) in same order? */
    for(d1=i1=0;i1<nn;i1++)for(i2=0;i2<nn;i2++)
    d1+=zizj[i1][i2]*zizj[i1][i2]; d1=sqrt(d1/(nn*nn));
    if (!mpi_i) printf("\n Z2Mag= %10.8f",d1); d2=d1; /* Z2Mag=(1/N^2 Sum ZiZj^2)^.5 */

    for(d1=i1=0;i1<nn;i1++)for(i2=0;i2<nn;i2++)
    d1+=sisj[i1][i2]*sisj[i1][i2]; d1=sqrt(d1/(nn*nn));
    if (!mpi_i) printf(  " S2Mag= %10.8f ",d1);  /* S2Mag=(1/N^2 Sum SiSj^2)^.5 */

#endif /* Korrelations */

#if 1 /* check again energy and its error: 0=yes 1=no */
    if (!mpi_i)
    printf("\nChange L%d in spins.c to"
           " check the energy again (slow!)\n",__LINE__-8);
#else /* nonparallel and slow version, makes no sence */
    r=wh(v0);	 if (!mpi_i) printf(" E%d/N= %12.8f",iv,D(r)/nn);
    		 if (!mpi_i) printf(  " E%d= %12.8f",iv,D(r));
    r=whq(v0,r); if (!mpi_i) printf(" %5.0e" ,D(r));   /* (H*v-<vHv>*v)^2 */
#endif
    /* if((fabs(r)>1.e-5) && (v1!=NULL)) return deg; */ 

    if(sym.numsym<60 || (double)n1<1e6) /* takes time, change ??? */
    /* wss is slow, if we have ss from sisj we dont need it again */
    if ((!mpi_i) && !(cfg_sisj&2))
      printf("\n cfg_sisj&2 is not set, 2S will not be calculated ");
    if(cfg_sisj&2) {
      if (!mpi_i) printf("\n cfg_sisj&2 is set, 2S will be calculated,");
      l1=itime();
      rr = wss(v0,v0);
      if (!mpi_i) printf(" 2S= %5.3f", rr);		/* 2S */
      if (!mpi_i) printf(" 2S.t=%s", sclock(itime()-l1));
    }
#if  CM(0,0,1)
    rr = D( wjj(v0,v0) );
    if (!mpi_i) printf(" 2J= %5.3f", rr);		/* 2J */
#endif

    MPI_Barrier(MPI_COMM_WORLD);
#if defined(SITE_DIST)
    /* --- staggered Magnetization --- */
    for(rr=i2=0;i2<nn;i2++)rr+=((site_dist(0,i2)&1)?-1:1)*sisj[0][i2];
    if (!mpi_i) printf(" StaggMag= %9.6f", rr/nn);
#endif
# ifdef Spin1chain
    stringorder2(stdout,v0);
# endif    
#else /* defined(Dimer) */
				/* SiSj */
    stringorder(stdout,v0);
    for(i1=0;i1<nn;i1++)sisj[0][i1]=3*zizj[0][i1];/* sowieso schon Dez99 */
    for(i1=0;i1<nn;i1++)sisj[2][i1]=3*zizj[2][i1];
    /* ##### C_x=Sum_(i!=j,(i,j aus X)){|<SiSj>|} ##### */
    d1=.0;							/* C_tot */
    for(i1=0;i1<nn;i1++)if(i1!=0)  d1+=fabs(sisj[0][i1]);
    for(i1=0;i1<nn;i1++)if(i1!=2)  d1+=fabs(sisj[2][i1]);
    if (!mpi_i) printf(" C_tot= %8.6f",(double)d1/(2*(nn-1)));
    d1=.0;							/* C_d */
    for(i1=0;i1<nn;i1++)if(i1%4<2)if(i1!=0) d1+=fabs(sisj[0][i1]);	
    if (!mpi_i) printf(  "C_d= %8.6f",(double)d1/(nn/2-1));
    d1=.0;							/* C_p */
    for(i1=0;i1<nn;i1++)if(i1%4>1)if(i1!=2) d1+=fabs(sisj[2][i1]);	
    if (!mpi_i) printf(  "C_p= %8.6f",(double)d1/(nn/2-1));
    d1=.0;							/* C_pd */
    for(i1=0;i1<nn;i1++)if(i1%4<2)if(i1!=2) d1+=fabs(sisj[2][i1]);	
    if (!mpi_i) printf(  "C_pd %8.6f",(double)d1/(nn/2));
#endif /* Dimer */
#if  CM(0,1,1)
    if (!mpi_i) printf("\n niu.=");  /* occupation value <v0|n_s|v1> */
    for(i1=0;i1<nn;i1++) {
      rr = wop(&op_nis,(void *)"1",1,&i1,v0,v0);
      if (!mpi_i) printf(" %8.6f", rr); 			/* niu */
    }
    if (!mpi_i) printf("\n niud=");  /* double occupation value (nu*nd)(i) */
    for(i1=0;i1<nn;i1++) {
      rr = wop(&op_nis,(void *)"3",1,&i1,v0,v0);
      if (!mpi_i) printf(" %8.6f", rr); /* niud */
    }
    if ((!mpi_i) && !(cfg_sisj&1))
      printf("\n cfg_sisj&1 is not set, slow cicj will not be calculated ");
    for(i3=0;i3<sym.num_dcorr;i3++){
      double uu, dd, ud, du;
      int iw[2];
      iw[0]=i1=sym.dcorr[i3][0];
      iw[1]=i2=sym.dcorr[i3][1];
      rr =  wop(&op_ninj,NULL,2,iw,v0,v0); /* ninj */
      if (!mpi_i) printf("\n i-j %2d %2d nn % 9.6f", i1, i2, rr); /* ninj */
      /* for (sym_ud) nuij is wrong, better take (nuij+ndij)/2 */
      uu=wop(&op_nisnjs,(void *)"1111",2,iw,v0,v0);
      dd=wop(&op_nisnjs,(void *)"2222",2,iw,v0,v0);
      ud=wop(&op_nisnjs,(void *)"1122",2,iw,v0,v0);
      du=wop(&op_nisnjs,(void *)"2211",2,iw,v0,v0);
      rr=wop(&op_nisnjs,(void *)"3030",2,iw,v0,v0);
      if (!mpi_i) printf(  " uu % 9.6f",uu);
      if (!mpi_i) printf(  " dd % 9.6f",dd);
      if (!mpi_i) printf(  " ud % 9.6f",ud);
      if (!mpi_i) printf(  " du % 9.6f",du);
      if (!mpi_i) printf(  " hh % 9.6f",rr);
      /* we need symmetrized operators if we use spin-inversion */
      if (nu==nd && sym.wud) if (!mpi_i) printf(  " uudd % 9.6f", 0.5*(uu+dd));
      if (nu==nd && sym.wud) if (!mpi_i) printf(  " uddu % 9.6f", 0.5*(ud+du));
      /* hopping between site iw[0] and iw[1]    *
       *    sum_s c(+,i,s)c(-,j,s)+h.c.          */
      if (cfg_sisj&1) { /* slow */
        rr = wop(&op_cicj,NULL,2,iw,v0,v0);
        if (!mpi_i) printf(  " cc % 9.6f", rr);
      }
    }
    if (nu==nd && sym.wud)
      printf("\nWarning: use uudd and uddu instead of uu, dd, ud and du");
#if 0
    /* example for F. Ouchni, makes sence for chains N>4 */
    { int iw[4]={2,4,0,2};
      if (nn>4) {
        rr = wop(&op_sisjsksl,NULL,4,iw,v0,v0);
        if (!mpi_i) printf("\n <s(i)s(i+2)-s(i-2)s(i)> i= 2 % 9.6f", rr);
      }
    }
#endif
#if 0 && CM(0,1,1)
/*  Definition by F. Ouchni (do you know about citations?)
 *  "bond-order wave" (BOW) (product of four fermionic operators)
 *  reference:PRB vol_67 page_205109
 *     BOW(r) = (-1)^r ( (\sum_{l} <B_{l,l+1} B_{l+r,l+r+1}>/N) - B^2)
 *          B = \sum_{l} <B_{l,l+1}>/N
 *  B_{l,l+1} = \sum_{spin} (C^{\dag}_{l s} C_{l+1 s} + h.c)
 *  sample: 
 *  correct? only if using symmetry? see m_tU.c
 *   m_tU 10 >daten.def  # v.0.22
 *      param=10,0,1,0,0,1,0.5,0.3333333,0.25,0.1 nud=2,2
 *   E0= 0.4921456 (ok)
 *        sum_k= -2 0 -1000 : <B(l,l+1)> l= 0  -0.642523 (see m_tU.c)
 *               <B(l,l+1) B(l+r,l+r+1)> r= 1   0.299550 OK
 *               BOW(r)                  r= 1   0.113287 OK (F.Ouchni)
 *               
 */
    /* only for chains */
    {
      int iw[4]={0,1,1,2}, r=1;
      double b2=0, b4;
      if (nn>1) printf("\n <B(l,l+1)>              l= 0 % 9.6f",
        b2= wop(&op_cicj,NULL,2,iw,v0,v0));
      if (nn>2)
      for (r=1; r<=nn/2; r++) {
        iw[2]=iw[0]+r;
        iw[3]=iw[1]+r;
        b4= wop(&op_cicjckcl,NULL,4,iw,v0,v0);
        if (mpi_i) continue;
        printf("\n <B(l,l+1) B(l+r,l+r+1)> r= %d % 9.6f",r,b4);
        printf("\n BOW(r)                  r= %d % 9.6f",r,(1-((r&1)<<1))*(b4-b2*b2));
      }
    }
#endif

#else /* now only JJ */
    if ((!mpi_i) && !(cfg_sisj&4))
      printf("\n cfg_sisj&4 is not set, mrule will not be calculated ");
    if (  cfg_sisj&4 ) {
      if (!mpi_i) printf("\n cfg_sisj&4 is set, mrule will be calculated ");
      for (i1=0;i1<sym.subl;i1++){
        for(i3=i2=0;i2<nn;i2++) if (sym.sl[i1][i2]+1>i3) i3=sym.sl[i1][i2]+1;
        mrule(v0,sym.sl[i1],i3); /* Marshall+output */
      }
    }
#endif /* JJ */

    MPI_Barrier(MPI_COMM_WORLD);
    /* calculate structure factor S(q)/N^2 v2.27 */
    for (i1=0;i1<sym.subl;i1++){
      int *sl, nsl;
      sl=sym.sl[i1];
      /* get the maximum SL index */
      for(nsl=i2=0;i2<nn;i2++) if (sl[i2]+1>nsl) nsl=sl[i2]+1;
      d1=0;
      if (nsl==2) {
        /* sublattices A,B => S(q=(pi,pi)), A-A: cos=1, A-B: cos=-1 */
        for (i2=0;i2<nn;i2++) for (i3=0;i3<nn;i3++)
          d1+=sisj[i2][i3]*((sl[i2]==sl[i3])?1:-1);
        if (!mpi_i) printf("\n SL%d S(pi)/N^2= % .8f",i1,d1/(nn*nn));
      }
      if (nsl==3) {
        /* sublattices A,B,C => S(q=(4pi/3,0)), A-A: cos=1, A-B: cos=-0.5 */
        for (i2=0;i2<nn;i2++) for (i3=0;i3<nn;i3++)
          d1+=sisj[i2][i3]*((sl[i2]==sl[i3])?1:-0.5);
        if (!mpi_i) printf("\n SL%d S(q3)/N^2= % .8f",i1,d1/(nn*nn));
      }
      if ((!mpi_i) && 2<=nsl && nsl<=3) {
        printf(" SL= ");
        for (i2=0;i2<nn;i2++) printf("%c",sl[i2]+'A');
      }
    }
    MPI_Barrier(MPI_COMM_WORLD); /* sync output */

#if CM(1,0,0)
#if defined(N1) && defined(N2)
#if N1>1 /* ToDo: adapt to symmetries */
    printf("spins.c L%d: old version, please adapt to SBase\n", __LINE__);
   /* local sublattice magnetization Msi^2=<1/N*sum_{j=1}^N{ti*tj*Siz*Sjz}>
    * sublattice magnetization Ms^2=1/N*sum_{i=1}^N{Msi^2}
    * ti=-1 for i odd (A), ti=+1 for i=even (B)      */
    printf(" Z= %8.5f",    D(wz(v0)));
    wtatb(&r1,&r2,v0);
    printf(" TA= %8.5f",   D(r1));
    printf(" TB= %8.5f",   D(r2));
    printf(" Dx= %8.5f",   D(wdx(v0)));
    printf(" Dy= %8.5f",   D(wdy(v0)));
    printf( " C= %8.5f",    D(wc(v0)));
#endif /* N1>1 */
#endif /* defined betts */
#endif

#ifdef CONFIG_MPI
#warning Missing MPI Code
#else
    if (vvv&1) out_ising_weight(v0);
    MPI_Barrier(MPI_COMM_WORLD);
    if (vvv&1) x_out(stdout,v0,0);       /* x */
    if (vvv&1) x_out(stdout,v0,1);
    MPI_Barrier(MPI_COMM_WORLD);
    if (vvv&1 && !(n1>>20)) x_sort(stdout,v0,v1); /* to slow */
#endif
    if (vvv&1 && !mpi_i) printf(" t+=%s",sclock(~0l));
    MPI_Barrier(MPI_COMM_WORLD);
    return(deg);
}
/* ====================== end of chk3 ======================= */    
    
    
/* sorting indizes using shellsort, R.Sedgewick: "Algorithmen in C" */
/* input: a,n;  output: ia; Nov98 JS +error in 1st Auflage */
void shellsort(double a[],int ia[],ULong n){  /* ULong statt int Mar99 JS */
  int v; ULong i,j,h;
  if(vvv) printf(" shellsort");
  for(i=0;i<n;i++) ia[i]=(int)i;	/* set indizes, (int) ??? Mar99 */
  for(h=1;h<=n/9;h=3*h+1);		/* using int or ULong ??? confusing */
  for(;h>0;h/=3)
   for(i=h-1;i<n;i++){
     v=ia[i]; j=i;
     while(j>=h && a[ia[j-h]]>a[v]){
       ia[j]=ia[j-h]; j-=h; }
     ia[j]=v;
   }
}

/* ================== USING LAPACK (linear algebra package) ================
 DSYEV - compute all eigenvalues and, optionally, eigenvectors of a real
         symmetric matrix A (LAPACK)
 ZHEEV - compute all eigenvalues and, optionally, eigenvectors of a complex
         Hermitian matrix A
 Makefile: add DFLAGS= -DLAPACK and system specific options LIB=...
 You can test LAPACK before using it with spinpack:
   Tru64: cc -migrate lapack_test.c -ldxml; PARALLEL=4; KMP_STACKSIZE=262144
   IRIX:  cc lapack_test.c -lscs[_i8][_mp] # [integer8][multi-processor]
   linux: gcc lapack_test.c -llapack -lf2c # lapack 3.0 + update + blas 3.0
   linux: gcc lapack_test.c -lacml -lg2c   # acml + f2c-Nachfolger=g2c gcc-3.4
 time[s] = 4e-8 * n1^3 (wf: GS160-alpha-731MHz dsyev nev=1 double ) 
 time[s] = 8e-8 * n1^3 (wf: GS160-alpha-731MHz zheev nev=1 complex)
 example:
   4x4-square-cplx sym=q_0_0_0+ n1=365 e0=-11.22848321
   jacobi=21s  lapack=1s  (alpha731MHz, real)
   jacobi=     lapack=1s  (alpha731MHz, complex)
   i686-2.6GHz n1=1617 zheev nev=1 139s  44MB # lapack-3.0 SuSE-9.2
   i686-2.6GHz n1=1617 zheev nev=0  24s  44MB
   i686-2.6GHz n1=1617 dsyev nev=1  44s  23MB # lapack-3.0 SuSE-9.2
   i686-2.6GHz n1=1617 dsyev nev=0   8s  23MB
   i686-2.6GHz n1=3583 dsyev nev=0  85s 103MB
   i686-2.6GHz n1=4704 dsyev nev=0 189s 176MB v2.22
   AOpt-2.2GHz n1=5481 dsyev nev=0 281s 240MB v2.23 4m..5m14s (259M/231M) acml=5m11s
                 10584 dsyev nev=0  ??  880MB v2.23 (vsz reduced)
                 16852 zheev nev=0 314m       v2.24 loki-Opteron-2.2GHz
                 18018 zheev nev=0 13h=48600s v2.22 8e-9s*n1^3 marvel
                 25000 zheevd     1e5s (ca.)
                  
 * * * WARNING * * *
 LAPACK is a Fortran-Library. There is no garanty that
 this C interface will work well on every system!
 SuSE-9.2 + lapack-3.0-920 = OK (n1>5400 wrong Es, spur ok)
  valgrind: Warning: set address range perms: large range 240330888, a 0, v 1
   5481^2*8 = 240330888

 ToDo: extract DSYEV and ZHEEV and convert to C routines (paralize!)
       this would also allow to use long (8byte) instead of int (4byte) for
       bufsize
*/
#ifdef LAPACK
#if (VecType & 4) == 0
#ifdef __cplusplus
extern "C"    /* C++ shall not create special object names */
#endif
void dsyev_(
        const char *jobz,               /* (in) "N"=NoVectors, "V"=Vectors */
        const char *uplo,               /* (in) "U"=UpperTriangleA, "L"=LowerTA */
        const int *n,                   /* (in) order of matrix */
        double *a,                      /* a[n][lda] (in/out) */
        const int *lda,                 /* (inp) leading dimension >=max(1,n) */
        double *w,                      /* w[n] (out) eigenvalues */
        double *work,                   /* work[lwork] (workspace/out) */
        const int *lwork,               /* (in)   >= max(1,3*n-1) better (nb+2)*n */
        int *info                       /* (out) ==0 on success */
        );
#else
extern "C"  /* let the name of the object link unchanged */
void zheev_(
        const char *jobz,               /* (in) "N"=NoVectors, "V"=Vectors */
        const char *uplo,               /* (in) "U"=UpperTriangleA, "L"=LowerTA */
        const int *n,                   /* (in) order of matrix */
        double *a,                      /* a[n][lda][2] (in/out) */
        const int *lda,                 /* (inp) leading dimension >=max(1,n) */
        double *w,                      /* w[n] (out) eigenvalues */
        double *work,                   /* work[lwork][2] (workspace/out) */
        const int *lwork,               /* (in)   >= max(1,2*n-1) better (nb+2) */
        double *rwork,                  /* (in)   >= max(1,3*n-2) better (nb+2) */
        int *info                       /* (out) ==0 on success */
        );
#endif
#endif

/* ??? use hbuf without ev! a4
 * for complex (hermitian) matrizes of type: A+iB
 *   simply diagonalyze [[ A,-B], [ B, A]], but all eigenvals twice.
 *   EVs: (u,v) -> u+iv
 * old driver for routine jacobi() for symmetric reel matrizes
 * jacobi sparc-Ultra1-143MHz(boson): time/s = 1.5e-6 dim ^ 3.2
 * ToDo: output matrix_asc.pbm for vizualization (or using gnuplot?)
 * --------------------------  Matrizen ----------------------------------- */
void fulldiag(tvec ev[],int nev){
    int nrot,*ix,blk,ii; ULong k,i,j,iv=1,jv=1;
    double *xm,*xv,*xd,xr,xi,trace1,trace2,h12,i12; cplx hh;   /* Matrizen 1 & 2 */
    thbuf *hbuf;
    minfo_t m_xm = {NULL,0,0,"tmp/xm.map",NULL};
    minfo_t m_xv = {NULL,0,0,"tmp/xv.map",NULL};
    minfo_t m_xd = {NULL,0,0,"tmp/xd.map",NULL};
    minfo_t m_ix = {NULL,0,0,"tmp/ix.map",NULL};
    if(vvv)printf("\nfulldiag nev=%d",nev); xi=0;
    if (mpi_n>1) E0("fulldiag dont work with MPI (ToDo)");
#if VecType & 4    		/* Okt99 JS */
    iv=2; /* trick: put real and imag-part in one real matrix */
    jv=2; /* trick to use jacobi for complex H via [[Hr,0],[0,Hi]] */
#ifdef LAPACK
    jv=1;
#endif
#endif
    /* standart jacobi and LAPACK expects simple vectors, convert block form */
    /* add 1 for the case n1=1, where Lapack fails */
    xm=xv=xd=NULL; trace1=trace2=0; h12=i12=0;
    /* int overflow ? */
#ifndef LAPACK
    xm=(double *)mymalloc(jv*n1*iv*n1*S_double,&m_xm); /* Matrix_H-in =  */
#endif
    xv=(double *)mymalloc(jv*n1*iv*n1*S_double,&m_xv); /* eigenvects-out */
    xd=(double *)mymalloc(      iv*n1*S_double,&m_xd); /* eigenvalues-out */
    ix=   (int *)mymalloc(      iv*n1*S_int,&m_ix);    /* sort-idx-out */
    for(i=0;i<iv*n1;i++) ix[i]=(int)i; /* initialize sort-index (LAPACK) */
    hbuf=alloc_hbuf();
    // hbuf->num_ahead_lines=1; /* ToDo: check! v2.32 */
    if (mpi_n>1) F0("missing MPI code");
    if((xm!=NULL || xv!=NULL) && xd!=NULL && ix!=NULL){
      int xflag=0;
      if (vvv) printf("\n store matrix H ...");
      /* generate a simple matrix, store all elements to xm[] */
      if (xm) for(i=0;i<iv*n1*jv*n1;i++) { xm[i]=0.0; }
      if (xv) for(i=0;i<iv*n1*jv*n1;i++) { xv[i]=0.0; }
      for(blk=0;blk<pt_n;blk++) for(i=0;i<b_len[blk];/* i++ */){
        hbuf->n=0; /* reset H-buffer, ToDo: num_ahead_lines=1 ??? */
        hbuf->block=blk;
        /* get hbuf->el.jj+r fulldiag, single thread only */
        hamilton_nhv( i, hbuf->num_ahead_lines, hbuf, 8);
        /* printf("\n n=%d ",hbuf->n[0]); */
        for(ii=0;ii<hbuf->n;ii++){
          j =hbuf->el[ii].jj; xflag=((j&XY_FLAG)?1:0); j&=~XY_FLAG;
          hh=hbuf->el[ii].rr;
          /* if (i+b_ofs[blk]==j) trace1+=Re(hh); */ /* just as a test */
          if (vvv&32) printf("\n H blk= %2d xy= %3ld %3ld r= %6g ",
             blk, i+b_ofs[blk], j, D(hh));
#ifdef LAPACK
          xv[((i+b_ofs[blk])*n1+j)*iv  ]+=Re(hh);
#if VecType & 4
          xv[((i+b_ofs[blk])*n1+j)*iv+1]+=Im(hh);
          i12+=Im(hh)*Im(hh); /* test for a reel matrix */
#endif
#else /* jacobi, construct [[ A,-B], [ B, A]] from [A+iB] */
          xm[(   i+b_ofs[blk])*iv*n1+(   j)]+=Re(hh);
#if VecType & 4
          xm[(n1+i+b_ofs[blk])*iv*n1+(n1+j)]+=Re(hh);
          xm[(n1+i+b_ofs[blk])*iv*n1+(   j)]+=Im(hh);
          xm[(   i+b_ofs[blk])*iv*n1+(n1+j)]-=Im(hh);
          i12+=Im(hh)*Im(hh); /* test for a reel matrix */
#endif
#endif
          if (xflag) i++; /* ToDo: v2.32 ??? check correctness  */
        }
      }
      if (vvv) printf(" t+=%s",sclock(~0l));
#if VecType & 4
      if (i12==0) printf("\n Warning: You are wasting more than 50%% memory"
                " and cpu time when using real matrix H in complex mode!\n");
#endif
      if (vvv) printf("\n use vvv&256 to output matrix.pbm (n1<=1024)");
      if ((vvv&256)&& n1<=1024) { /* PBMTXT max. 70 chars per line, v2.33 */
        FILE *f1; unsigned int bb, ib;
        int sorti[1024]; double diagH[1024];
#ifdef LAPACK
# define Hfull(ii,jj) xv[(ii*n1+jj)*iv]
#else
# define Hfull(ii,jj) xm[(ii*n1*iv+jj)]
#endif
        printf("\n use vvv&768 for sorted matrix.pbm (n1<=1024)");
        for (i=0;i<n1;i++) {
          sorti[i]=i; diagH[i]=Hfull(i,i); }
        if (vvv&512) shellsort(diagH, sorti, n1);
        /* output of relevant Hii ranges */
        if (vvv&512)
        for (i=0;i<n1-1;i++)
          if (diagH[sorti[i]]!=diagH[sorti[i+1]])
            printf("\n %4lu  ii= %4d  Hii= %9.4f",i,sorti[i],diagH[sorti[i]]);
        f1=fopen("matrix.pbm","w+"); /* PBMBIN 1=black 0=white HL */
        if (f1) {
          fprintf(f1,"P4\n# SpinPack matrix H\n%lu %lu\n", n1, n1);
          for(i=0;i<n1;i++)
          for(j=0;j<n1;j+=8) { /* arch specific */
            for(bb=ib=0;ib<8 && ib<n1-j;ib++)
              if (Hfull(sorti[i],sorti[j+ib])) bb|=1<<(7-ib); /* MSB..LSB */
            fprintf(f1,"%c",bb);
          }
          fclose(f1);
        }
      }
#ifdef LAPACK
      for(i=0;i<n1;i++) { trace1+=xv[(i*n1+i)*iv]; } /* build the trace */
      for(i=  0;i<n1;i++)         /* as a check sum upper left elements */
      for(j=i+1;j<n1;j++) {  h12+=xv[(i*n1+j)*iv]; }
      ii=(int)n1;              /* buffer for n1 */
      nrot=(int)(3*n1); /* buffer for lwork >= max(1,3*n-1) better (nb+2) */
      /* ToDo: problem if nrot is 4 byte and n1^2 is larger than 2e9 */
      if (  (3 *n1*iv)*S_double
          !=(3.*n1*iv)*S_double ) E0("integer overflow");
      blk=-1;                  /* buffer for info, set by lapack */
      /* use  ev as matrix  and  xm as workbuffer */
      myfree((void *)xm,&m_xm);
      xm=(double *)mymalloc(nrot*iv*S_double,&m_xm); /* work-space */
#if (VecType & 4) == 0   /* reel symmetric matrix, iv=1 */
      printf("\n LAPACK dsyev");
      dsyev_(((nev)?"V":"N"),"U",&ii,xv,&ii,xd,xm,&nrot,&blk);
      /* there is probably a bug in LAPACK-1.1,
       * I got sometimes segfaults freeing xm (small n1 and nrot=n1*n1+1)
       * ToDo: paralize dsyev?
       */
      printf("_info=%d (0=ok)",blk);
#else                 /* hermitian matrix, iv=2 */
      /* nrot/=2; ** use half of mx as work-buffer, other half as rwork-buffer */
      minfo_t m_rw = {NULL,0,0,"tmp/rw.map",NULL};
      double *rwork=(double *)mymalloc(nrot*sizeof(double),&m_rw);
      if (!rwork) { E0(" malloc *rwork failed - skip zheev"); }
      printf("\n LAPACK zheev");
      /* valgrind shows that zhemv_ of blas-3.0 accesses WORK+len=xm+nrot...
       * so we malloc more elements than nrot=LWORK counts 
       */
      if (rwork)
        zheev_(((nev)?"V":"N"),"U",&ii,xv,&ii,xd,xm,&nrot,rwork,&blk);
      printf("_info=%d (0=ok)",blk);
      if (rwork) myfree(rwork,&m_rw); 
  /* if not freed, zheev causes segfault after about 7 runs on Tru64-cxx */
#endif
#else
      for(i=0;i<n1;i++) { trace1+=xm[(i*n1*iv+i)]; } /* build the trace */
      for(i=  0;i<n1;i++)         /* as a check sum upper left elements */
      for(j=i+1;j<n1;j++) {  h12+=xm[(i*n1*iv+j)]; }
      printf(" JACOBI");
      jacobi2(&xm[0],(int)(iv*n1),xd,xv,&nrot);
      printf(" rotations: %3d",nrot);
      shellsort(xd,ix,iv*n1);		/* sort xd[ix[i]] */
      if (iv>1) for (j=0;j<n1;j++){   /* remove every 2nd */
        ix[j]=ix[j*2]; /* does not work? */
      }
#endif
      printf(" t+=%s\n eigenvals=",sclock(~0l));
      for (j=0;j<n1;j++){ trace2+=xd[ix[j]]; printf(" %+12.8f",xd[ix[j]]); }
      printf("\n  traceH= %e sumE= %e traceH-sumE= %e\n sum_h12= %e",
             trace1, trace2, trace2-trace1, h12);
      if(vvv&4)printf("\n number eigenvals eigenvects (sorted):");
      /* convert simple vectors to block form for calling <O>-functions */
      if (nev)
      for (j=0;j<n1;j++){ /* every eigenvect */
        if(j<NEW)ew[j]=xd[ix[j]];
        if(vvv&1)printf("\n%3ld: %+12.8f",j,xd[ix[j]]);
        for(blk=0;blk<pt_n;blk++) for(k=0;k<b_len[blk];k++){
#ifdef LAPACK   /* this is hopefully right, normalization? */
          xr=           xv[((b_ofs[blk]+k)+n1*j)*iv  ];
          if (iv>1) xi=-xv[((b_ofs[blk]+k)+n1*j)*iv+1];
          /* the minus sign was found empirically (N=6=4+2, k=-2 1) */
#else
          xr=          xv[(b_ofs[blk]+k   )*iv*n1+ix[j]];
          if (iv>1) xi=xv[(b_ofs[blk]+k+n1)*iv*n1+ix[j]];
#endif
          if(vvv&4){ /* output all koefficients of the eigenstate */
             printf(" ");
             if(iv>1)  printf("("); outmpl(stdout,xr);
             if(iv>1){ printf(","); outmpl(stdout,xi); printf(")"); }
          }
          /* copy ev in blk-structure */
          if(!A_ptr(ev)) continue;
          ev[0].v[blk][k] = M_d2f(C(xr,xi)); /* ignore xi if not complex */
        }

        /* call <ev|O|ev>-functions here, SiSj-example: */
        if(cfg_sisj&5)
        { /* new block necessary to allow definition of int sites */
          int i3,sites[2];   /* must follow the left bracket */
          /* could be problematicaly on degenerate states! not proofed */
          for(i3=0;i3<sym.num_dcorr;i3++){
            sites[0]=sym.dcorr[i3][0];
            sites[1]=sym.dcorr[i3][1];
            printf(" Z%dZ%d= %9.6f", sites[0], sites[1],
                    wop(&op_zizj,NULL,2,sites,ev,ev));
            if(cfg_sisj&1)
            printf(" S%dS%d= %9.6f", sites[0], sites[1],
                    wop(&op_sisj,NULL,2,sites,ev,ev));
          }
        }

        /* normvec(ev);        LAPACK not normalized? */
        if(cfg_sisj&2) wss(ev,ev); /* output S^2 */
      }
      if (vvv&4 ) printf("\n");
      if (A_ptr(ev))
      for (j=0;j<(ULong)nev;j++){    /* copy eigenvects */
        for(blk=0;blk<pt_n;blk++) for(k=0;k<b_len[blk];k++){ 
#ifdef LAPACK   /* this is hopefully right */
          xr=           xv[(b_ofs[blk]+k+n1*j)*iv  ];
          if (iv>1) xi=-xv[(b_ofs[blk]+k+n1*j)*iv+1];
#else
          xr=          xv[(b_ofs[blk]+k   )*iv*n1+ix[j]];
          if (iv>1) xi=xv[(b_ofs[blk]+k+n1)*iv*n1+ix[j]];
#endif
          ev[j].v[blk][k] = M_d2f(C(xr,xi));
        }
      }
    }
    myfree((void *)ix,&m_ix);
    myfree((void *)xm,&m_xm);
    myfree((void *)xv,&m_xv);
    myfree((void *)xd,&m_xd);
    free_hbuf(hbuf);
    return;}

/* ================== read/write parameters  ====================== */

char *strsub(char *sd,const char *ss,int i1,int i2){/* copy ss[i1..i1+i3-1] */
 int i3;for(i3=0;i3<i2;i3++)sd[i3]=ss[i1+i3];sd[i2]=0;return(sd);}
char *fk(const char *ss,int i2){ /* get i2-tes element from komma-list */
 static char s1[512];int i1,i3,i4,i5;char s2;
 s1[0]=0;
 i1=0;  /* pos letztes komma +1 */
 i3=0;  /* klammerebene */
 i4=0;  /* komma-nr */
 for(i5=0;i5<(int)strlen(ss)+1;i5++){
  s2=ss[i5];if(s2==0)s2=',';
  if(s2=='(') i3++;else
  if(s2==')') i3--;else
  if((s2==',') && (i3==0)){ 
    i4++; if(i4<i2) i1=i5+1; else break;  } }
 if((i4>=i2) && (i2>0))strsub(s1,ss,i1,i5-i1);
 if((s1[0]=='(')&&(s1[strlen(s1)-1]==')'))strsub(s1,s1,1,(int)strlen(s1)-2);
 return s1; }


cplx atoc(char *s){	/* (r,w) => r*exp(I*w*Pi) */
  zahl x=Z(0),y=Z(0);int i=0;
  while(s[i]!=0 && s[i]!=',')i++;
  x=Z(atof(s));if(s[i]==',')y=Z(atof(&s[i+1]));
#if VecType & 4     
  return polar(x,y*M_PI);}
#else
  if(y!=0)E0("imag(r)!=0"); return C(x,0);}
#endif     

void outpara(FILE *in){
  int i;
  fprintf(in," param=");
  for(i=1;i<numpara;i++){
    fprintf(in," %.9g",(double)para[i]);
    if(i<numpara-1)fprintf(in,",");
  } 
  fprintf(in,"\n# p1= %g\n# p2= %g  ... can also be used\n",
          (double)para[1], (double)para[2]);
  fflush(in);
  return;
}

int set_ww(){
  int i;
  for(i=0;i<nw;i++){
    ww.co[i]=cos(M_PI*para[ww.w[i]]*ww.ex[i]);
    ww.si[i]=sin(M_PI*para[ww.w[i]]*ww.ex[i]);
  }
  return 0;
}

int scan_ud(char *s1,int sum){
  if (sum) {  /* add values, used for loops */
    nu+=atoi(fk(s1,1));
    nd+=atoi(fk(s1,2));
  } else {
    nu=atoi(fk(s1,1));
    nd=atoi(fk(s1,2));
  }
  if(nu>nn)nu=nn;
  if(nd>nn)nd=nn;
#if CM(1,0,0)
  if(nu+nd!=nn){ nu=nn/2;nd=nn-nu;
    if (!mpi_i) printf("\nwarning: nu+nd!=N set nu=%d nd=%d",nu,nd); }
#endif  
#if CM(0,1,0)
  if(nu+nd>nn){ nu=nn/2;nd=nn-nu-1;
    if (!mpi_i) printf("\nwarning: nu+nd> N set nu=%d nd=%d",nu,nd); }
#endif  
  return 0;
}

int scan_para(char *s1, int sum){
  int i;
  if (sum)  for(i=1;i<numpara;i++){para[i]+=atof(fk(s1,i));}
  else      for(i=1;i<numpara;i++){para[i]= atof(fk(s1,i));}
  set_ww();
  return 0;
}

void inppara(char *s1){
  int i;
  if (vvv) printf("\nwarning: do not use :-command, use new param=,nud= instead!");
  scan_ud(s1,0);
  for(i=1;i<numpara;i++){para[i]=atof(fk(s1,i+2));}
  set_ww();
  return;
}

/* =========================== useful functions (handling strings) ====== */

void testsym(){
}

/*
 * search end of line
 */
char *scan_eol(char *s){
  char *s2;
  s2=strchr(s,'#');if(s2) return s2; /* comment-line */
  s2=strchr(s,'%');if(s2) return s2; /* comment-line */
  s2=s+strlen(s);
  return s2;
}
/*
 * search "vnam=" in s, get position of next char, ret 0=not found
 *  ToDo: search expression "p[0-9]*=" or p%d=%f ?
 */
char *scan_var(char *s,const char *vnam){
  char *s1,*s2;
  if((s1=strstr(s,vnam))==NULL)return NULL;s1+=strlen(vnam);
  s2=scan_eol(s);if(s2)if(s2<s1)return NULL; /* comment */
  if(s1[0]!='='){if (vvv) printf(" %s=???",vnam); return NULL;} 
  s1++; while(s1[0]==' ' || s1[0]=='\t') s1++;
  return s1;
}

int scan_int(char *s,const char *vnam,int *var){
  char *s1;
  if( (s1=scan_var(s,vnam))==NULL ) return -1;
  if(var){ *var=atoi(s1); if(vvv && !mpi_i) printf(" %s=%d",vnam,*var);}
  return 0;
}

int scan_inta(char *s,const char *vnam,int *var,int len){
  char *s1,*s2; int i=0;
  if( (s1=scan_var(s,vnam))==NULL ) return -1;
  s2=scan_eol(s); len=(int)(s2-s1);
  if(var){
   if (vvv && !mpi_i) printf(" %s=",vnam);
   for(i=0;i<len;i++){
    var[i]=(int)strtol(s1,&s2,10);
    if(s1==s2 || s2[0]==0) break;
    if (vvv && !mpi_i) printf(" %d",var[i]);
    s1=s2;
   }
   if (vvv && !mpi_i) printf("\n");
  }
  return i;
}

/* search "vnam=" in s, set value to var, ret 0=found,1=not found,2=err
   last modify 1998-139 */
int scan_dbl(char *s,const char *vnam,double *var){
  char *s1;
  if( (s1=scan_var(s,vnam))==NULL ) return -1;
  if(var){ *var=strtod(s1,NULL);
    if (vvv && !mpi_i) printf(" %s=%g",vnam,*var);
  }
  return 0;}

int clr_status(){ FILE *f1;f1=fopen("tmp/status","w");fclose(f1);return 0;}
int wl_status(ULong n1){ FILE *f1;
  if(vvv)printf(" write tmp/status");
  f1=fopen("tmp/status","w");fprintf(f1," %lu",n1);
  fclose(f1);return 0;}
int rl_status(ULong *n1){ FILE *f1;
  if(vvv)printf(" load tmp/status");
  f1=fopen("tmp/status","r");
  if(!f1) { printf("ERROR open tmp/status\n"); n1=0; return 1; }
  fscanf(f1," %lu",n1);
  fclose(f1);return 0;}

int p_mask_ini(int nn){
  int i,j;
  /* --- initialize bitmasks for faster parity() --- */
  for(i=0;i<nn;i++)for(j=0;j<nn;j++)
    if(i<j) p_mask[i][j]=(~(~(tbase1)0 << j << j)) & (~(tbase1)0 << i << i << 2);
    else    p_mask[i][j]=(~(~(tbase1)0 << i << i)) & (~(tbase1)0 << j << j << 2);
  return 0;
}

void ns_thread( void *ptr); /* see hilbert.c */
void wopij_block( void *ptr ); /* see xval.c */
void wop_block( void *ptr ); /* see xval.c */
void diag2_thread( void *ptr ); /* see diag.c */

/* ToDo: write a test_task which only outputs "(mpi.thread): I am here" */
void test_pthread( void *ptr ){
  main_pts *pp=(main_pts *)ptr;
  if (mpi_i*pt_n+pp->num<4 || mpi_i*pt_n+pp->num>mpi_n*pt_n-1-4)
    printf(" node.thread %d.%d is running as block %d\n",
           mpi_i, pp->num, mpi_i*pt_n+pp->num);
  else
  if (mpi_i*pt_n+pp->num==4)
    printf(" skip output of threads %d.%d..%d.%d\n",
      mpi_i, pp->num, (mpi_n*pt_n-1-4)/pt_n, (mpi_n*pt_n-1-4)%pt_n);
}

/* void-function for pthreads */
void main_pthread( void *ptr ){
#ifdef CONFIG_PTHREAD
  /* volatile */ main_pts *pp=(/* volatile */ main_pts *)ptr;
  /* test_pthread(ptr); */
  for(;;) {  /* endless loop */
    pthread_mutex_lock(&pt_mutex1);
    while (pp->task==0)
      pthread_cond_wait(&pt_cond1,&pt_mutex1); /* wait for broadcast */
      /* ToDo: better have one(two) condition variable(s) per thread */
    pthread_mutex_unlock(&pt_mutex1);

    if (pp->task<0) break;
    pp->worker( (void *)pp->data ); /* call worker function for this thread */

    pthread_mutex_lock(&pt_mutex1); /* ToDo: check for race condition? */
    pp->task=0; /* ready */
    pthread_cond_signal(&pt_cond2); /* wake up master */
    pthread_mutex_unlock(&pt_mutex1);
  }
  pthread_exit(0);
#endif
}

/* ---------------------------- MAIN --------------------------------- */
int    main(int argc,char *argv[]){
  static char nam1[12],s1[500],*ss; ULong n1max=1000;
  int i,ii,i1,j,itr=0,iv,nev,ne0,blk,nruns=0;
  FILE *in1=NULL, *in2=NULL;     /* in2 is inline-backup */
  double r1;
  tvec v0,v1,ev[NEV]; /* large vectors */
  minfo_t m_v0[B_NUM], m_v1[B_NUM], m_ev[NEV][B_NUM],
          m_l2 = {NULL,0,0,"tmp/l2.map",NULL};
#ifdef CONFIG_PTHREAD
  /* stack memory is local, memory + filesystem is shared */
  pthread_t pt[B_NUM];
#endif
#ifdef CONFIG_MPI /* accept both (MPI+MPP) together */
  /* parallel but share everything exxcept local variables */
  /* memory is local, filesystem is divided in shared and local parts */
  int mpi_thread=0;
#if defined(CONFIG_PTHREAD) && defined(MPI_THREAD_SERIALIZED)
  /* try setenv LAM_MPI_THREAD_LEVEL=3
   0: Corresponds to MPI_THREAD_SINGLE
   1: Corresponds to MPI_THREAD_FUNNELED
   2: Corresponds to MPI_THREAD_SERIALIZED
   3: Corresponds to MPI_THREAD_MULTIPLE
  */
  if (MPI_Init_thread(&argc, &argv, MPI_THREAD_SERIALIZED, &mpi_thread) != MPI_SUCCESS) exit(1);
  if (mpi_thread!=MPI_THREAD_SERIALIZED) printf("mpi_thread=%d\n",mpi_thread);
#else
  if (MPI_Init(&argc, &argv) != MPI_SUCCESS) exit(1); i=0;
#endif
  MPI_Comm_size(MPI_COMM_WORLD, &mpi_n);
  MPI_Comm_rank(MPI_COMM_WORLD, &mpi_i);
  for (mpi_n_bits=0;mpi_n_bits<32;mpi_n_bits++)
     if (1<<mpi_n_bits > mpi_n-1) break;
#endif
  sprintf(nam1,"daten.i");
  i=0;while(++i<argc){ ss=argv[i];
    if(ss[0]=='-'){
      if(ss[1]=='v'){ vvv=atoi(ss+2);             continue; } /* verbose */
      if(ss[1]=='i'){ sprintf(nam1,"%.11s",ss+2); continue; } /* daten.i, ToDo: stdin */
      if(ss[1]=='o'){ freopen(ss+2,"a",stdout);   continue; } /* redirect stdout  */
      if(ss[1]=='t'){ pt_n=atoi(ss+2);            continue; } /* threads per node */
    }
    if (!mpi_i)
      printf("options: [-v<verbose>] [-i<daten.i>]"
                     " [-t<threads_per_node>] [-o<outfile>]\n");
    return 1;
  }
  setvbuf(stdout, (char *)NULL, _IONBF, 0); /* no buffer */

  if (pt_n>B_NUM) {
      pt_n=B_NUM; /* set to maximum */
    if (!mpi_i)
      printf("WARNING: number of threads is set back to maximum (B_NUM)\n");
  }
  for (pt_n_bits=0;pt_n_bits<32;pt_n_bits++)
     if (1<<pt_n_bits > pt_n-1) break;

  for(j=0;j<pt_n;j++){
    /* give vectors blocks different names for shared filesystems */
    m_v0[j].p=NULL; m_v0[j].l=0; m_v0[j].mode=0; m_v0[j].ff=NULL;
    sprintf(m_v0[j].name,"tmp/v0_%03d.map", pt_n*mpi_i+j);
    m_v1[j].p=NULL; m_v1[j].l=0; m_v1[j].mode=0; m_v1[j].ff=NULL;
    sprintf(m_v1[j].name,"tmp/v1_%03d.map", pt_n*mpi_i+j);

    v0.v[j]=v1.v[j]=(mcplx *)NULL;	/* NULL = not allocated */
    v0.id=0; v1.id=1;		/* identification ??? */
    for(i=0;i<NEV;i++){
      m_ev[i][j].p=NULL; m_ev[i][j].l=0; m_ev[i][j].mode=0; m_ev[i][j].ff=NULL;
      sprintf(m_ev[i][j].name,"tmp/ev%d_%03d.map", i, pt_n*mpi_i+j);
      ev[i].id=8+i; ev[i].v[j]=(mcplx *)NULL;
    }
  } 
  l2b=NULL; l1=NULL; /* initialization */
  iv=0; algo=0; x_p=3;n1=0l; nn=NN; nw=0*Nw; ne0=0;
  sym.subl=0; /* if 0 auto detection will be used */
  hfmax=1000000l/(Hsize/HBLen);		/* max. 1MB Filesize hxy.tmp */
  x_max=5; if(MAXx<x_max) x_max=MAXx;   /* max. num. of output-configs */
  /* bug: x_max=min(5,MAXx); - this gives warnings on SGI-CC, JS Mar99 */
  pew=2; max_ea=1; nev=NEV; cfg_sisj=1;
  usemem=0l; maxmem=MaxMem; itime();	/* init itime */
  /* output some info for possible bug reports */
  if (!mpi_i) { /* output only once per MPI cluster */
  printf("\nSPINPACK version " version_string " " version_date
           " compiled at "__DATE__" "__TIME__"\n"
           " Copyright (C) 2001-2008 Joerg Schulenburg at ovgu.de\n"
           " released under the GNU General Public License\n"
           " base coding for %s-model\n", CM("Heisenberg","t-J","Hubbard"));
#ifdef __GNUC__
  printf(" Compiler: GNUC %d.%d\n",__GNUC__,__GNUC_MINOR__);
#endif
#ifdef __DECCXX_VER
  printf(" Compiler: DECCXX %d\n",__DECCXX_VER);
#endif
#ifdef __SUNPRO_CC
  printf(" Compiler: SUNPRO_CC 0x%x\n",__SUNPRO_CC);
#endif
#if defined(__mips) && defined(_COMPILER_VERSION)
  printf(" Compiler: MIPS_CC %d\n",_COMPILER_VERSION);
#endif
#ifdef HAVE_ZLIB
  printf(" using ZLIB\n");
#endif
#ifdef _Vlint_h
  printf(" using class Vlint for %d-bit emulation (slow)\n",sizeof(vlint)*8);
#endif
  printf(" use symmetric base configurations (mandatory)\n");
#ifdef CONFIG_S1SYM
  /* ToDo: replace by set of commutating symmetries */
  printf(" use special S1-symmetric base configurations (default)\n");
#else
  printf(" no special S1-symmetries (slow, only for debugging)\n");
#endif
#ifdef AddSS
  printf(" AddSS is on (full matrix, makes the program much more slow)\n");
#endif
#ifdef DEBUG
  printf(" DEBUG is on (more checks, makes the program slower)\n");
#endif
  }

  ini_handler(mpi_i, vvv); /* signalhandler (sighand.h) after pthread_create? */

#ifdef CONFIG_MPI
  { int version, subversion;
  MPI_Get_version(&version,&subversion);
  if (!mpi_i)
   printf(" using MPI %d.%d with %d nodes (mpi_n_bits=%d)\n"
          "   HNZ_MPI_BUFSIZE/NUM_AHEAD_LINES= %.3f\n",
           version, subversion, mpi_n, mpi_n_bits,
              HNZ_MPI_BUFSIZE*1./NUM_AHEAD_LINES);
  }
#endif
#ifdef CONFIG_PTHREAD
  /* parallel but share everything exxcept local variables */
  if (!mpi_i)
   printf(" using PTHREAD with %d threads (per node, max=%d, bits=%d)"
          ", use option -t\n", pt_n, B_NUM, pt_n_bits);
  MPI_Barrier(MPI_COMM_WORLD); /* sync "I am here" output */
  /* IRIX, True64 and Linux works with NULL=default, but SunOS does not */
  pthread_attr_init (&pt_attr);
  pthread_attr_setscope (&pt_attr, PTHREAD_SCOPE_SYSTEM);
  pthread_mutex_init(&pt_mutex1,NULL);
  pthread_mutex_init(&pt_mutex2,NULL);
  pthread_cond_init(&pt_cond1,NULL); /* correct init? */
  pthread_cond_init(&pt_cond2,NULL); /* correct init? */
  /* create main threads */
  for (i=0;i<pt_n;i++) {
    pthread_mutex_init(&token_ring[i       ],NULL);
    pthread_mutex_init(&token_ring[i+  pt_n],NULL);
    pthread_mutex_init(&token_ring[i+2*pt_n],NULL);
    pthread_mutex_init(&token_ring[i+3*pt_n],NULL);
    pthread_mutex_lock(&token_ring[i       ]);
    pthread_mutex_lock(&token_ring[i+  pt_n]);
    pthread_mutex_lock(&token_ring[i+2*pt_n]);
    pthread_mutex_lock(&token_ring[i+3*pt_n]);
    main_pt[i].num=i;     /* put thread number to child */
    main_pt[i].data=NULL; /* for first do nothing */
    main_pt[i].task=0;    /* initialize */
    if( pthread_create((pthread_t *)&(pt[i]),
                       (const pthread_attr_t *)&pt_attr,
                       (void * (*)(void *)) &main_pthread,
                       (void *)  &(main_pt[i]) ) !=0 )  F0("pthread_create");
  }
  { /* coordinate the output of "I am here!" messages */
    int ii;
    for (ii=0; ii<pt_n; ii++)
      parallel_start(ii, test_pthread, (void *)&main_pt[ii]);
    for (ii=0; ii<pt_n; ii++)
      parallel_wait(ii);
  }
  pthread_mutex_unlock(&token_ring[0     ]); /* put token into the ring1 */
  pthread_mutex_unlock(&token_ring[2*pt_n]); /* put token into the ring2 */
#endif
  MPI_Barrier(MPI_COMM_WORLD); /* sync "I am here" messages */
  if (!mpi_i)
    printf(" using Ham1=%d Ham2=%d Ham3=%d VecType=%d B_NUM=%d\n",
           Ham1, Ham2, Ham3, VecType, B_NUM);
  if (!mpi_i)
  printf(" h_file.h: TYPE_SIZE XY=%d BY=%d HR=%d in Bytes, Hsize=%.0f kB\n",
   (int)sizeof(XY_TYPE),(int)sizeof(BY_TYPE),(int)sizeof(HR_TYPE),Hsize/1024.);
  if (((BY_TYPE)(mpi_n*pt_n))!=mpi_n*pt_n) {
    printf(" num_of_nodes * num_of_threads dont fit to BY_TYPE\n");
    ERR(9999);
  }

  /* create the child early to save memory on pipe or fork */
  MPI_Barrier(MPI_COMM_WORLD);

  sub_ini();		/* v1.6.1 start child for gzip */
  initvars(mpi_i);	/* define the model from daten.def ToDo: later *in */
  initmasks(nn);	/* initialize bitmasks for lintab/sym/etc. */
  p_mask_ini(nn);       /* initialize bitmasks for parity() */

  MPI_Barrier(MPI_COMM_WORLD);

  /* initialize ud-symmetry data, HFIELD or TBC? */
  sym.wud=+1;           /* initialize symmetry data */
#ifdef Sud
# error "macro Sud removed, switch off sym_ud by setting sym.wud=sym_ud=0"
#endif
#if defined(HFIELD) || defined(CONFIG_TBC) /* reset sym.wud */
  sym.wud=0;		/* do not use ud-sym */
#endif     /* reset sym.wud */
  
  for (i=0;i<MaxSym;i++){ 
    sym.k[i]=0; sym.w[i]=1; sym.l[i]=2;/* sym.d[i][0]=0;sym.m[i][0]=M0F; */
    for (j=0;j<nn;j++) sym.i[i][j]=j;
  }

  /* output helpful configuration for support */
  if ((vvv&1) && !mpi_i)
    printf(" llong=%d long=%d int=%d zahl=%d mzahl=%d cplx=%d"
           " mcplx=%d hxy=%d tbase1=%d tbase=%d\n",
    (int)sizeof(ULLong),(int)Lsize,(int)sizeof(int),
    (int)sizeof(zahl),(int)sizeof(mzahl),(int)sizeof(cplx),(int)sizeof(mcplx),
    (int)Hsize/HBLen,(int)sizeof(tbase1),(int)sizeof(tbase));

  /* check bond data for errors */
  for (i=0;i<2*nw;i++) if(iw[i>>1][i&1]>=nn) F1("site>=NN iw[%u]",i>>1);
  /* initialize phase parameters */
  for (i=0;i<nw;i++){  ww.ex[i]=1;
    if (iw[i][0]>iw[i][1]){		/* lowest index first */ 
      j=iw[i][0]; iw[i][0]=iw[i][1]; iw[i][1]=j;
      ww.si[i]=-ww.si[i]; ww.ex[i]=-1;
    }
  }

  MPI_Barrier(MPI_COMM_WORLD);

  if((in1=fopen(nam1,"rt"))==NULL) if (!mpi_i) {
    printf("\n ERROR: open daten.i, see example doc/daten.i\n\n");
    return 0;
  }

  paini();  /* parameter initialization */
  sprintf(FILE_l1, "tmp/l1_%03d.dat", mpi_i); /* initialize the name */
  
  for(;;){  /* load and execute input file */
    if (feof(in1)){ if (in2==NULL) break; in1=in2; in2=NULL; } /* restore */
    s1[0]=0;fgets(s1,500,in1);  /* without \n ?? , .begin ... .end */
    if(s1[0]=='#' || s1[0]=='%' || s1[0]==0 || s1[0]=='\n'){ /* nop */ } else
    if(s1[0]=='!'){
      /* the tmp/pid check dont work on MPI + network_filesystems */
      if (!mpi_i) { printf(s1);system((char *)&s1[1]);/* exec arg */}
    } else
    if(in2==NULL && strncmp(s1,"include",7)==0) { /* max include depth = 1 */
        for (ss=s1+8;*ss!=0 && *ss>='0';ss++); *ss=0;
        in2=in1; in1=fopen(s1+8,"rt");
        printf("\ninclude %s",s1+8);
        if(!in1) { in1=in2; printf("failed"); }
        continue;
      } else
    if(!scan_int(s1,"max_ea",&max_ea)){max_ea=max(1,max_ea);} else
    if(!scan_int(s1,"save_mode",&save_mode)){ } else
    if(!scan_int(s1,"last_point",&checkpoint)){ } else
    if(!scan_int(s1,"nev",&nev)){nev=min(max(0,nev),NEV);} else
    if(!scan_int(s1,"ne0",&ne0)){ne0=min(max(0,ne0),NEW-1);} else
    if(!scan_int(s1,"pew",&pew)){pew=min(max(1,pew),NEW);} else
    if(!scan_int(s1,"startvec",&startvec)){} else
    if(!scan_int(s1,"sym_ud",&sym.wud)){ if(abs(sym.wud)>1)ERR(9999);
#if defined(HFIELD) || defined(CONFIG_TBC)
      printf("\nError: sym_ud can not be used for fields and TBC\n\n");
      sym.wud=0;
#endif
    } else
    if(!scan_int(s1,"sym_k",NULL)){ /* in daten.k schreiben? */
      for(i=0;i<MaxSym;i++)sym.kgen[i]=0;
      scan_inta(s1,"sym_k",sym.kgen,MaxSym); } else
    if(scan_var(s1,"SL")){ /* set sublattice 0=A 1=B etc. */
      char *s2=scan_var(s1,"SL");
      printf(" set sublattice %d: ", sym.subl);
      for (i=0;i<nn;i++) {
        sym.sl[sym.subl][i]=((s2[i]&~0x20)-'A')%nn;
        printf("%c",'A'+sym.sl[sym.subl][i]);
      } printf("\n");
      /* old: scan_inta(s1,"SL",sym.sl[sym.subl],nn); */
      sym.subl++; sym.subl%=MaxNumSL; /* handle overflow as reset */
    } else
    if(!scan_dbl(s1,"maxmem" ,&r1)){ maxmem=(ULong)r1;} else  
    if(!scan_dbl(s1,"maxfile",&r1)){ hfmax=(ULong)(r1*HBLen/Hsize);
      if( (double)hfmax+1<r1*HBLen/Hsize ) ERR(612);
      } else
    if(!scan_dbl(s1,"xy_anisotropy+",&r1)){ h_xypara+=r1;} else
    if(!scan_dbl(s1,"z_anisotropy+",&r1)){ h_zpara+=r1;} else
    if(!scan_dbl(s1,"xy_anisotropy",&r1)){ h_xypara=r1;} else
    if(!scan_dbl(s1,"z_anisotropy",&r1)){ h_zpara=r1;} else
    if(!scan_int(s1,"verbose",&vvv)){} else
    if(!scan_int(s1,"sisj",&cfg_sisj)){} else
    if( (ss=scan_var(s1,"nud+"  ))!=NULL ){ scan_ud(ss,1); } else
    if( (ss=scan_var(s1,"param+"))!=NULL ){ scan_para(ss,1); } else
    if( (ss=scan_var(s1,"nud"  ))!=NULL ){ scan_ud(ss,0); } else
    if( (ss=scan_var(s1,"param"))!=NULL ){ scan_para(ss,0); } else
    /* allow "p1=-1.0", "p1= -1.0" or " p1= -1.0" as input per line */
    if (sscanf(s1," p%d= %lf",&i1,&r1)==2) {
        para[i1]=r1; printf("set parameter p%d= %f\n",i1,r1); } else
    if(s1[0]=='x'){sscanf(s1,"%*c%d%lu",&x_p,&x_max);} else /* ToDo: replace by function */
    if(s1[0]=='l'){ loadvec(&(ev[0]),s1+1,-1,-1);
    } else
    if(s1[0]==':'){inppara((char *)&s1[1]);} else
    if(strncmp(s1,"newtime",7)==0){starttime=0l;itime();} else	/* reset clock */
    if(strncmp(s1,".out",4)==0){if(!mpi_i)outpara(stdout);} else
    if(strncmp(s1,"pout",4)==0){printf("%s",&s1[4]);} else
    if(s1[0]=='a'){ sscanf(s1,"%*c%d%lu",&algo,&n1max);
      if (!mpi_i) outpara(stdout);
      MPI_Barrier(MPI_COMM_WORLD); /* debug */
      /* --- test and if neccessary create Sym symmetries P --- */
      symcreate(mpi_i);

      MPI_Barrier(MPI_COMM_WORLD); /* debug */

      if (vvv&1) testsym();

      MPI_Barrier(MPI_COMM_WORLD);

      if (!mpi_i) { /* output on node 0 */
        printf("\n nn,nu,nd %d %d %d   nw= %d"
               "  (number of sites, up/down-spins and bounds)\n",nn,nu,nd,nw);
        /* wud=(+1,-1, 0) to kud=(0,1,-1), wud=exp(kud*I*Pi) */
        printf(" k_ud %01d k_T ", ((sym.wud) ? 1-(sym.wud+1)/2 : -1));
        for (i=0;i<sym.numsym;i++) if (sym.g[i]) printf("_%01d",sym.k[i]);
        /* ToDo: move output of parameters to a function? */
        printf("\n parameters ");
        //  for (j=0;j<nw;j++){ printf("t[%d]=%d ",j,ww.wt[j]); }
        for(i=1;i<numpara;i++){ int j,k=0;
          /* output human readable names, ex: J1=J2= 0.3 t1= -0.7 ... */
          for (j=0;j<nw;j++) if (ww.wj[j]==i) {printf("J%d=",(k=i)); break;}
          for (j=0;j<nw;j++) if (ww.wt[j]==i) {printf("t%d=",(k=i)); break;}
          for (j=0;j<nw;j++) if (ww.w[j]==i)  {printf("w%d=",(k=i)); break;}
          for (j=0;j<nw;j++) if (ww.wv[j]==i) {printf("V%d=",(k=i)); break;}
          for (j=0;j<nw;j++) if (ww.wh[j]==i) {printf("h%d=",(k=i)); break;}
          for (j=0;j<nn;j++) if (ww.we[j]==i) {printf("e%d=",(k=i)); break;}
          for (j=0;j<nn;j++) if (ww.wu[j]==i) {printf("U%d=",(k=i)); break;}
          for (j=0;j<nn;j++) if (ww.wf[j]==i) {printf("f%d=",(k=i)); break;}
          for (j=0;j<nn;j++) if (ww.wd[j]==i) {printf("D%d=",(k=i)); break;}
          if (!k) printf("p%d=",i); /* unused or unknown parameter */
          printf(" %g  ",(double)para[i]);
        } printf("\n");
        printf(" Bsize=%d t+=%s\n",(int)Bsize,sclock(~0l));
      }

      MPI_Barrier(MPI_COMM_WORLD);
      
      /* ===================== new-method l1+H storage ================ */
      /* generate hilbert space+matrix by applying hilbert Op
         on single configs iterative, ToDo: check speed */	
      if (algo&8) {
        if (mpi_n!=1 || pt_n!=1) { 
          printf("WARNING: (mpi_n!=1 || pt_n!=1) is untested, please report\n");
        } /* not tested, ToDo */
        n1=(n1max)?n1max:100000;	/* n1max<=n1 variiere Basis! */
        if (((XY_TYPE)(n1/(mpi_n*pt_n)))!=n1/(mpi_n*pt_n)) {
          printf(" block_n1 dont fit to XY_TYPE, see h_file.h\n");
          ERR(9999);
        }
        for (n1_bits=0;n1_bits<(int)sizeof(ULong)*8;n1_bits++) 
           if (((ULong)1)<<n1_bits > n1-1) break;
        b_ini(n1,vvv);          /* init b_len[] */
        if((algo&7)<2){
          int vvv2=vvv; if (pt_n*mpi_n>4) vvv&=~2; /* avoid to much output */
          /* ToDo: alloc within threads or remove thread-blocks */
          if ((!mpi_i) && (!v0.v[0])) printf(" malloc v0,v1");
          for(blk=0;blk<pt_n;blk++)
            v0.v[blk]=(mcplx *)mymalloc(b_len[mpi_i*pt_n+blk]*Csize,&m_v0[blk]);
          for(blk=0;blk<pt_n;blk++)
            v1.v[blk]=(mcplx *)mymalloc(b_len[mpi_i*pt_n+blk]*Csize,&m_v1[blk]);
          vvv=vvv2;
        }
        {
          if (1.*(node_n1*(off_t)Bsize)!=node_n1*1.*Bsize)
            printf("\nERROR mymap(n1*Bsize): off_t overflow");
          fl1=fopen(FILE_l1,"r+b");
          if(!fl1)fl1=fopen(FILE_l1,"w+b"); if(!fl1) F1("fopen %s",FILE_l1);
          l1 = (unsigned char *)
                mymap(FILE_l1, fl1, node_n1 *(off_t)Bsize,
                                    node_ofs*(off_t)Bsize);
          if (!l1) { E0("mymap failed, skipping"); continue; }
        }
        for(i=0;i<nev;i++){
          int vvv2=vvv; if (pt_n*mpi_n>4) vvv&=~2; /* avoid to much output */
          if ((!mpi_i) && (!ev[i].v[0])) printf(" malloc ev%d", i);
          for(blk=0;blk<pt_n;blk++)
            ev[i].v[blk]=(mcplx *)mymalloc(b_len[mpi_i*pt_n+blk]*Csize,&m_ev[i][blk]);
          vvv=vvv2;
        }
// sprintf(m_n2.name, "tmp/n2_%03d.map", mpi_i);
// norm2=(ATnorm2 *)mymalloc(node_n1*sizeof(ATnorm2),&m_n2);base2ok=0;
        l2b  =  (ULong *)mymalloc(node_n1*Lsize,&m_l2); /* hilfsarray2 */
        n1max=node_n1;       /* save memory block len before n1 changed */
        storeh2(0);	/* ??? provisorium - testphase */
        for (n1_bits=0;n1_bits<(int)sizeof(ULong)*8;n1_bits++) 
           if (((ULong)1)<<n1_bits > n1-1) break;
        /* l2b=(ULong *)myfree((void *) l2b,&m_l2); / * zur Zeit noch!!! */
        randvec(&v1,0);itr=inc1(nev,ne0,&v0,&v1);   /* EW(T),EV(T)  */
        randvec(&v1,0);incv1(nev,&v0,&v1,ev,itr);   /* EV(H)   */
        /* repeat computation with new startvector */
        if(n1==n1max && nev>0)F0("Fatal, should not break here");
        if (nev) for(j=0;j<nev;j++)chk3(&ev[j],&ev[j],0);
      } else 
      {	/* =================== old-method =================== */
        /* generating Hilbert space, and Hilbert matrix in two steps */

      MPI_Barrier(MPI_COMM_WORLD);
        if( checkpoint<1 ) {
          long l1=itime();
          n1=(ULong)numsymconf();          /* slow, calculate n1 */

          if (save_mode & 1) wl_status(n1); /* write n1 to disk */
          checkpoint++;
          if (!mpi_i)
            printf("\n ns.t=%s checkpoint=%d",sclock(itime()-l1),checkpoint);
        } else rl_status(&n1);
        for (n1_bits=0;n1_bits<(int)sizeof(ULong)*8;n1_bits++) 
           if (((ULong)1)<<n1_bits > n1-1) break;
        
        if (!n1) { if (!mpi_i) E0("n1==0"); checkpoint=0; continue; }
        if (!mpi_i) {
          printf("\n N,nu,nd= %u  %u %u  n1= %lu  log2(n1)= %.1f",
                  nn, nu, nd, n1, log(n1)/log(2.0)); /* n1_bits */
          /* Tru64@alpha: log2(0) segfault */
          /* (long+mzahl*2)*n1/nodes, Dsize=sizeof(mzahl) */
          printf("\n MinMem %.2e", (Bsize+Dsize*(2+nev))*1.0*n1/mpi_n);
          printf(" userMaxMem %.2e userMaxHfile %.2e (per node)\n",
                 (double)maxmem, (double)hfmax*(double)Hsize/HBLen);
        }
        /* we need a bit for the XY_FLAG, ToDo: change for MPI */
        if (((n1<<1)>>1)!=n1) { if (!mpi_i) E0("n1>max"); continue; }
        if (((XY_TYPE)(n1/(mpi_n*pt_n)))!=n1/(mpi_n*pt_n)) {
          printf(" block_n1 dont fit to XY_TYPE, see h_file.h\n");
          ERR(9999);
        }

      MPI_Barrier(MPI_COMM_WORLD);
        b_ini(n1,vvv);	/* block initialization (b_len[], b_ofs[] etc.) */
        MPI_Barrier(MPI_COMM_WORLD);

        /* --- l1 is now initialized by numsymconf --- */
        if (!l1) {
          fl1=fopen(FILE_l10,"r+b");
          if (!fl1) fl1=fopen(FILE_l10,"w+b"); /* create */
          if (!fl1) F1("fopen %s",FILE_l10);
          /* Oct07: every node maps only its part of l1[]!
           * may be this is not necessary because only accessed 
           * parts loaded into mem? but if mmap failes, its loaded!
           */
          if (1.*(node_n1*(off_t)Bsize)!=node_n1*1.*Bsize)
            printf("\nERROR mymap(n1*Bsize): off_t overflow");
          l1 = (unsigned char *)
               mymap(FILE_l10, fl1, node_n1 *(off_t)Bsize,
                                    node_ofs*(off_t)Bsize);
          if (!l1) { E0("mymap failed, skipping"); continue; }
        }

        MPI_Barrier(MPI_COMM_WORLD);
        /* ToDo: add vec_malloc[_block](tvec,m_v0) to vector.h */
        if((algo&7)<2){ /* we need this 2 vectors for Lanczos algorithm */
          int vvv2=vvv; if (pt_n*mpi_n>4) vvv&=~2; /* avoid outputs */
          /* ToDo: alloc within threads or remove thread-blocks */
          if ((!mpi_i) && (!v0.v[0])) printf(" malloc v0,v1");
          for(blk=0;blk<pt_n;blk++)
            v0.v[blk]=(mcplx *)mymalloc(b_len[mpi_i*pt_n+blk]*Csize,&m_v0[blk]);
          for(blk=0;blk<pt_n;blk++)
            v1.v[blk]=(mcplx *)mymalloc(b_len[mpi_i*pt_n+blk]*Csize,&m_v1[blk]);
          vvv=vvv2;
        }
        MPI_Barrier(MPI_COMM_WORLD);
        if(((algo&7)==2) && nev<1){ nev=1; E0("nev is 0, set 1"); }
        for(i=0;i<nev;i++){
          int vvv2=vvv; if (mpi_i>3) vvv&=~2; /* avoid to much output */
          if ((!mpi_i) && (!ev[i].v[0])) printf(" malloc ev%d", i);
          for(blk=0;blk<pt_n;blk++)
            ev[i].v[blk]=(mcplx *)mymalloc(b_len[mpi_i*pt_n+blk]*Csize,&m_ev[i][blk]);
          vvv=vvv2;
        }
// sprintf(m_n2.name, "tmp/n2_%03d.map", mpi_i);
// norm2=(ATnorm2 *)mymalloc(node_n1*sizeof(ATnorm2),&m_n2);base2ok=0;
        /* if( checkpoint < 2 ) */ 
        { /* norm2 is not mapped! change it ??? */
          long l1=itime();
          ini_norm2(); if (checkpoint<2) checkpoint++;
          if (!mpi_i)
             printf("\n n2.t=%s checkpoint=%d",sclock(itime()-l1),checkpoint);
        } /* else load_l1(l1,n1); --- mmap used */
        MPI_Barrier(MPI_COMM_WORLD);
        outbasis(stdout);

        MPI_Barrier(MPI_COMM_WORLD);
        if ((algo&7)<3) { storeh(); }         /* write H to disk */
        ERR(-1);	/* reset error + print old errors */

        if (vvv && !mpi_i) printf(" t+=%s",sclock(~0l));
        MPI_Barrier(MPI_COMM_WORLD);
        test_all_sym();	/* only array-ov test Mai99 */
        if((algo&7)<2){ /* ToDo: remove! its twice!? */
          int vvv2=vvv; if (pt_n*mpi_n>4) vvv&=~2; /* avoid outputs */
          if ((!mpi_i) && (!v0.v[0])) printf(" malloc v0,v1");
          /* ToDo: alloc within threads or remove thread-blocks */
          for(blk=0;blk<pt_n;blk++)if(!v0.v[blk])
          v0.v[blk]=(mcplx *)mymalloc(b_len[mpi_i*pt_n+blk]*Csize,&m_v0[blk]);
          for(blk=0;blk<pt_n;blk++)if(!v1.v[blk])
          v1.v[blk]=(mcplx *)mymalloc(b_len[mpi_i*pt_n+blk]*Csize,&m_v1[blk]);
          vvv=vvv2;
        }

        MPI_Barrier(MPI_COMM_WORLD);

        if(!err_fatal)
        for(iv=1;iv<=max_ea;iv++){	/* Nummer des Startvektors   */
          if (iv>1) checkpoint=3;	/* reset checkpoint */
          if((algo&7)==0){	/* pseudo random startvector depends only on iv! */
           /* here is the right place to choose iv2=random instead of iv */
           clrvec(&v1); /* to get local blocks for threads on ccNUMA */
           randvec(&v1,iv+(startvec&4)*2*nruns);
           itr=inc1(nev,ne0,&v0,&v1);                   /* EW(T),EV(T)  */
           if( checkpoint<5 || iv>1 ) {			/* v1.7.14 */
             if(nev){ randvec(&v1,iv+(startvec&4)*2*nruns);
             incv1(nev,&v0,&v1,ev,itr); }               /* EV(H) */
             if ( iv<=1 ) { checkpoint++; 
                            if (!mpi_i) printf(" checkpoint=%d",checkpoint);
                          }
           } else { if(nev>0)loadvec(&ev[0],FileGS,j,-1); }; /* reload */
          } else if((algo&7)==1){
           randvec(&v0,iv+(startvec&4)*2*nruns);
           itr=lanz1(&v0,&v1,&ev[0]);        /* 1-step-lanzcos EW,EV(H) */
          } else if((algo&7)==2 && nev>0){
           randvec(&ev[0],iv+(startvec&4)*2*nruns);
           itr=lanz2(&ev[0]);                /* 2x2-Matrix     EW,EV(H) */
          } else if((algo&7)==3 && nev>0){
            printf("\n algo=3: load vectors from disk");
            if (loadvec(&ev[0],FileGS,iv,-1)) continue; /* v1.7.13+ */
            if (nev>1)
            if (loadvec(&ev[1],FileEX,iv,-1)) continue; /* v2.28 */
          } /* <O> zu H' */
          else if((algo&7)==4){ fulldiag(ev,nev); } /* full diag = jacobi */
          if (!mpi_i) {
            printf("\n erg= "); outew(stdout);
            if (vvv) fprintf(stdout," t+=%s",sclock(~0l));
            printf("\n E0/N= %12.8f E0/nw= %12.8f",D(ew[0]/nn),D(ew[0]/nw)); /* Feb99 */
            if (ne0)
            printf("\n E%d/N= %12.8f E%d/nw= %12.8f E%d= %12.8f",ne0,D(ew[ne0]/nn),ne0,D(ew[ne0]/nw),ne0,ew[ne0]); /* Feb99 */
          }
          MPI_Barrier(MPI_COMM_WORLD);

          if((algo&7)==4 || (!nev))break;	
          for(ii=0,j=1;j<iv;j++){ /* orthogonalize ev against saved vectors */
            loadvec(&v0,FileGS,j,-1);
            if((ii=iortho(&ev[0],&v0))==1)break;
          } /* stop if we have found all orthogonal ev's */
          if(ii!=0)break; savevec(&ev[0],FileGS,iv,-1);
          if(nev>1){      savevec(&ev[1],FileEX,iv,-1); }
          if((algo&7) <3)for(j=0;j<nev;j++)chk3(&ev[j],&v1,iv);
          if((algo&7)==3)chk3(&ev[0],&ev[0],iv);/* only Expection Values */
          if((algo&7)==4)for(j=0;j<nev;j++)chk3(&ev[j],&ev[j],iv);
#ifdef CONFIG_DIMER_CHECK
          if (nev>2) { /* Richter extension to check for dimers */
            /* we break the symmetry by adding lowest two states */
            printf("\n dimer-check+ (ev0+ev1)/2:");
            clrvec(&ev[2]);
            add_av(&ev[2],+sqrt(0.5),&ev[0]);
            add_av(&ev[2],+sqrt(0.5),&ev[1]);
            normvec(&ev[2]); // not needed?  
            chk3(&ev[2],&v1,1);
            printf("\n dimer-check- (ev0-ev1)/2:");
            clrvec(&ev[2]);
            add_av(&ev[2],+sqrt(0.5),&ev[0]);
            add_av(&ev[2],-sqrt(0.5),&ev[1]);
            normvec(&ev[2]);
            chk3(&ev[2],&v1,1);
          }
#endif
        } 
        if (!mpi_i) printf("\n degree_of_degeneracy=%d<%d\n",iv-1,max_ea);
      }
 /* label_end: */

      MPI_Barrier(MPI_COMM_WORLD);

// norm2=(ATnorm2 *)myfree((void *) norm2, &m_n2);base2ok=0;
      l2b=    (ULong *)myfree((void *) l2b, &m_l2);
      if(algo&8) myunmap(l1,fl1,n1max*(off_t)Bsize,0);
      else       myunmap(l1,fl1,n1   *(off_t)Bsize,0);
      fclose(fl1); l1=NULL; fl1=NULL;
      {
        int vvv2=vvv; if (pt_n*mpi_n>4) vvv&=~2; /* avoid to much output */
        if ((!mpi_i) && (!ev[0].v[0])) printf(" free ev");
        for(i=0;i<nev;i++) {
          for(blk=0;blk<pt_n;blk++)
            ev[i].v[blk]=(mcplx *)myfree((void *)ev[i].v[blk],&m_ev[i][blk]);
        }
        if ((!mpi_i) && (!v0.v[0])) printf(" free v0,v1");
        for(blk=0;blk<pt_n;blk++) {
             v1.v[blk]=(mcplx *)myfree((void *)v1.v[blk],&m_v1[blk]);
             v0.v[blk]=(mcplx *)myfree((void *)v0.v[blk],&m_v0[blk]);
        }
        vvv=vvv2;
      }
      n1=0l; /* reset n1 */
      checkpoint=0; nruns++;

      MPI_Barrier(MPI_COMM_WORLD);
      if (!mpi_i) printf(" --- next run\n");

    } /* end a#-algo-command */
    if (err_fatal){ err_fatal=0; printf("\n reset err_fatal"); }

  } /* for(;;) scan input lines from daten.i */ 
  fclose(in1); if(!mpi_i) printf("\n");
  b_free(); /* release dynamic arrays b_ofs[], b_len[] */

#ifdef CONFIG_PTHREAD
  /* send no further tasks to threads */
  pthread_mutex_lock(&pt_mutex1);
  for (i=0;i<pt_n;i++) main_pt[i].task=-1;    /* finalize */
  pthread_cond_broadcast(&pt_cond1);
  pthread_mutex_unlock(&pt_mutex1);
  for (i=0;i<pt_n;i++) {          /* join main threads */
    pthread_join(pt[i],NULL);
  }
#endif
  sub_end();	/* kill child */
#if defined(CONFIG_MPI)
  MPI_Barrier(MPI_COMM_WORLD);
  MPI_Finalize();
#endif
  return(0);
}

#ifdef F2C
/* lapack-3.0 for linux needs f2c and f2c needs MAIN__() */
/* if you know a better solution, please tell me */
#if (VecType & 4) || defined(__cplusplus)
extern "C"
#endif
int MAIN__(){ return main(0,NULL); }
#endif
